/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/
window.CodeCheckers = (function() {
    const CodeCheckers = this;

/** Create code checker function returning only first return value if return value of no function is falsey */
CodeCheckers.or = function(...fkts) {
    fkts.push(fkts.shift());
    return (function(obj) {
        return fkts.reduce((prev, curr) => (prev && curr.call(this, obj)), true);
    });
};

/** Create a code checker which simply accepts everything */
CodeCheckers.accept = function(
    unlock, // Optional: Unlock level step?
    fine,   // Optional: Number of fine points to give if main procedure does not pass this test
    message // Optional: Message to display
) {
    if (!Number.isFinite(fine))
        fine = 0;
    return (function() {
        const result = { accept: true };
        if (fine)
            result.fine = fine;
        if (message)
            result.msg = message;
        if (unlock !== undefined)
            result.unlock = unlock;
        return result;
    });
};

/** Create a code checker which simply rejects everything */
CodeCheckers.reject = function(
    fine,   // Optional: Number of fine points to give if main procedure does not pass this test
    message // Optional: Message to display
) {
    if (!Number.isFinite(fine))
        fine = 0;
    return (function() {
        const result = { accept: false };
        if (fine)
            result.fine = fine;
        result.msg = message ? message : 'In dieser Aufgabe darf der Algorithmus nicht ausprobiert werden.';
        return result;
    });
}

/** Create a function that reject any end of execution but a successfull end */
CodeCheckers.allowOnlyDone = function(
    fine,       // Optional: Number of fine points to give if main procedure does not pass this test
    message     // Optional: Message to display if execution did not end successfully
) {
    if (!Number.isFinite(fine))
        fine = 0;
    return (function ({ result }) {
        const obj = { accept: false, fine };
        if (message)
            obj.message = message;
        if (result != RobotSimulator.RESULT_DONE)
            return obj;
    });
};

/** Create a function that accepts manual execution termination by user */
CodeCheckers.acceptUserAbort = function(
    fine,   // Optional: Number of fine points to give if main procedure does not pass this test
    unlock, // Optional: Unlock level step?
    message // Optional: Message to display
) {
    if (!Number.isFinite(fine))
        fine = 0;
    return (function({ result }) {
        if (result !== undefined)
            return null;
        const r = { accept: true };
        if (fine)
            r.fine = fine;
        if (message)
            r.msg = message;
        if (unlock !== undefined)
            r.unlock = unlock;
        return r;
    });
};

/** Functions creating a function disallowing a single type of token (a number of fine points and a message can be specified or not) */
(() => {
    function createDisallowCreator(tokenclass, stdMessage) {
        return (function(fine, message) {
            if (!Number.isFinite(fine))
                fine = 0;
            const msg = message ? message : stdMessage;
            return (function({ tokens }) {
                const wrongTokens = tokens.filter(token => (token instanceof tokenclass));
                return wrongTokens.length ? { accept: false, line: wrongTokens[0].line, fine, msg } : null;
            });
        });
    }

    CodeCheckers.disallowIf = createDisallowCreator(TokenIf, "FALLS darf nicht verwendet werden");
    CodeCheckers.disallowThen = createDisallowCreator(TokenThen, "DANN darf nicht verwendet werden");
    CodeCheckers.disallowElse = createDisallowCreator(TokenElse, "SONST darf nicht verwendet werden");
    CodeCheckers.disallowDefinition = createDisallowCreator(TokenDefinition, "LERNE darf nicht verwendet werden");
    CodeCheckers.disallowRunFast = createDisallowCreator(TokenRunFast, "VERTIEFT darf nicht verwendet werden");
    CodeCheckers.disallowRepeat = createDisallowCreator(TokenRepeat, "WIEDERHOLE darf nicht verwendet werden");
    CodeCheckers.disallowTimes = createDisallowCreator(TokenTimes, "MAL darf nicht verwendet werden");
    CodeCheckers.disallowWhile = createDisallowCreator(TokenWhile, "SOLANGE darf nicht verwendet werden");
    CodeCheckers.disallowBracketOpen = createDisallowCreator(TokenBracketOpen, "Eine öffnende Klammer darf nicht verwendet werden");
    CodeCheckers.disallowBracketClose = createDisallowCreator(TokenBracketClose, "Eine schließende Klammer darf nicht verwendet werden");
    CodeCheckers.disallowBlockComment = createDisallowCreator(TokenBlockComment, "Ein Blockkommentar darf nicht verwendet werden");
    CodeCheckers.disallowLineComment = createDisallowCreator(TokenLineComment, "Ein Zeilenkommentar darf nicht verwendet werden");
    CodeCheckers.disallowCommand = createDisallowCreator(TokenBlockComment, "Eine Anweisung darf nicht verwendet werden");
    CodeCheckers.disallowWhitespace = createDisallowCreator(TokenWhitespace, "Leerzeichen usw. dürfen nicht verwendet werden");
})();

/** Functions creating a function disallowing typical requested groups of token types (a number of fine points and a message can be specified or not) */
(() => {
    function createDisallowCreator(arrDisallowed) {
        return (function(fine, message) {
            const checkers = arrDisallowed.map(fkt => fkt(fine, message));
            return (function(obj) {
                return checkers.reduce((prev, checker) => (prev || checker.call(this, obj)), null);
            });
        });
    }

    CodeCheckers.allowOnlyCommands = createDisallowCreator([CodeCheckers.disallowDefinition, CodeCheckers.disallowIf, CodeCheckers.disallowRepeat, CodeCheckers.disallowWhile]);
    CodeCheckers.allowOnlyCommandsAndSlowDefs = createDisallowCreator([CodeCheckers.disallowRunFast, CodeCheckers.disallowIf, CodeCheckers.disallowRepeat, CodeCheckers.disallowWhile]);
    CodeCheckers.allowOnlyCommandsAndDefs = createDisallowCreator([CodeCheckers.disallowIf, CodeCheckers.disallowRepeat, CodeCheckers.disallowWhile]);
})();

/** Create a code checking function allowing only a specified number of occurrences of a command */
CodeCheckers.countCommand = function(
    fine,           // Number of fine points to give if main procedure does not pass this test
    command,        // Name of command to count
    min,            // Minimal allowed number of occurrences
    max,            // Maximal allowed number of occurrences (can be Infinity)
    message,        // Optional: Message to display if number of occurrences is too low
    messageTooMuch  // Optional: Message to display if number of occurrences is too hight (defaults to message)
) {
    if (!Number.isFinite(fine))
        fine = 0;
    if (!Number.isFinite(min))
        min = 0;
    if (!Number.isFinite(max))
        max = Infinity;
    return (function({ tokens }) {
        const occurrences = tokens.filter(token => ((token instanceof TokenCommand) && (token.name == command)));
        const result = { accept: false, fine };
        if (occurrences.length)
            result.line = occurrences[occurrences.length - 1].line;
        if (occurrences.length < min) {
            result.msg = message ? message : ('Es darf nicht so viele Aufrufe von „' + command + '” geben');
            return result;
        }
        if (occurrences.length > max) {
            result.msg = messageTooMuch ? messageTooMuch : (message ? message : ('Es muss mehr Aufrufe von „' + command + '” geben'));
            return result;
        }
        return null;
    });
};

/** Function creating a checking function to ensure content of main procedure (without definitions) */
CodeCheckers.ensureMain = function(
    fine,               // Number of fine points to give if main procedure does not pass this test
    compareWithCode,    // Code to compare code of main procedure with
    compareStrict       // Set true if subsequences should not be collapsed and repeats not be expanded
) {
    let target = SourceCodePart.tokensToSequenz(Token.tokenize(compareWithCode), true);
    if (!compareStrict)
        target = target.createCanonical();
    if (!Number.isFinite(fine))
        fine = 0;
    return (function({ compiled }) {
        const main = compareStrict ? compiled.copyWithoutDefinitions() : compiled.copyWithoutDefinitions().createCanonical();
        const diff = target.compare(main);
        if (!diff)
            return null;
        const result = { fine, accept: false };
        if (diff.actualValue instanceof SourceCodePart)
            result.line = diff.actualValue.line;
        else if (main.content.length)
            result.line = main.content[main.content.length - 1].line;
        if (diff.type != 'item') {
            result.msg = 'In der Hauptsequenz stimmt etwas nicht…';
            return result;
        }
        if (diff.actualValue instanceof SourceCodePart)
            result.msg = 'In der hervorgehobenen Zeile stimmt deine Lösung nicht';
        else
            result.msg = 'In der Hauptsequenz fehlt etwas';
        return result;
    });
};

/** Functions creating a function checking user defined definitions */
(() => {
    function getUserDefinedNames(compiled) {
        return compiled.getContentByClass(SourceCodeDefinition).map(def => def.name);
    }

    function getUserDefinedByName(compiled, fktName) {
        const def = compiled.getContentByClass(SourceCodeDefinition).filter(def => (def.name == fktName));
        return def.length ? def[0] : null;
    }

    function cmpUserDefined(compiled, fktName, planed, compareStrict, forceRunFast) {
        let def = getUserDefinedByName(compiled, fktName);
        if (!def)
            return { index: -1, actualValue: undefined, plannedValue: fktName, type: 'notDefined' };
        if ((forceRunFast === true) && def.isSlow)
            return { index: -1, type: 'runFast', actualValue: false, plannedValue: true };
        if ((forceRunFast === false) && !def.isSlow)
            return { index: -1, type: 'runFast', actualValue: true, plannedValue: false };
        if (compareStrict)
            return def.body.compare(planed);
        return def.body.createCanonical().compare(planed);
    }

    /** Create a code checking function allowing only a specified number of user defined functions */
    CodeCheckers.countUserDefined = function(
        fine,           // Number of fine points to give if main procedure does not pass this test
        min,            // Optional: Minimal allowed number of occurrences
        max,            // Optional: Maximal allowed number of occurrences (can be Infinity)
        message,        // Optional: Message to display if number of occurrences is too low
        messageTooMuch  // Optional: Message to display if number of occurrences is too hight (defaults to message)
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        if (!Number.isFinite(min))
            min = 0;
        if (!Number.isFinite(max))
            max = Infinity;
        return (function({ tokens, compiled }) {
            const occurrences = getUserDefinedNames(compiled);
            const result = { accept: false, fine };
            if (occurrences.length < min) {
                result.msg = message ? message : ('Es muss ' + (occurrences.length ? 'mehr ' : '') + 'selbstdefinierte Anweisungen geben');
                return result;
            }
            if (occurrences.length > max) {
                result.msg = messageTooMuch ? messageTooMuch : (message ? message : ('Es darf nicht so viele selbstdefinierte Anweisungen geben'));
                return result;
            }
            return null;
        });
    };

    /** Create a code checking function allowing only a specified number of calls of user defined functions */
    CodeCheckers.countUserDefinedCalls = function(
        fine,           // Number of fine points to give if main procedure does not pass this test
        min,            // Minimal allowed number of occurrences
        max,            // Maximal allowed number of occurrences (can be Infinity)
        message,        // Optional: Message to display if number of occurrences is too low
        messageTooMuch  // Optional: Message to display if number of occurrences is too hight (defaults to message)
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        if (!Number.isFinite(min))
            min = 0;
        if (!Number.isFinite(max))
            max = Infinity;
        return (function({ tokens, compiled }) {
            const userDefinedNames = getUserDefinedNames(compiled);
            const occurrences = tokens.filter(token => ((token instanceof TokenCommand) && (userDefinedNames.indexOf(token.name) != -1)));
            const result = { accept: false, fine };
            if (occurrences.length)
                result.line = occurrences[occurrences.length - 1].line;
            if (occurrences.length < min) {
                result.msg = message ? message : ('Es muss mehr Aufrufe der selbstdefinierten Anweisungen geben');
                return result;
            }
            if (occurrences.length > max) {
                result.msg = messageTooMuch ? messageTooMuch : (message ? message : ('Es darf nicht so viele Aufrufe der selbstdefinierten Anweisungen geben'));
                return result;
            }
            return null;
        });
    };

    /** Create a code checking function allowing only a specified number of not used user defined functions */
    CodeCheckers.countUserDefinedUnused = function(
        fine,           // Number of fine points to give if main procedure does not pass this test
        min,            // Minimal allowed number of occurrences
        max,            // Maximal allowed number of occurrences (can be Infinity)
        message,        // Optional: Message to display if number of occurrences is too low
        messageTooMuch  // Optional: Message to display if number of occurrences is too hight (defaults to message)
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        if (!Number.isFinite(min))
            min = 0;
        if (!Number.isFinite(max))
            max = Infinity;
        return (function({ tokens, compiled }) {
            const userDefined = compiled.getContentByClass(SourceCodeDefinition);
            const userDefinedNames = userDefined.map(def => def.name);
            tokens.filter(token => {
                if (!(token instanceof TokenCommand))
                    return;
                const index = userDefinedNames.indexOf(token.name);
                if (index < 0)
                    return;
                userDefined.splice(index, 1);
                userDefinedNames.splice(index, 1);
            });
            const result = { accept: false, fine };
            if (userDefined.length)
                result.line = userDefined[Math.min(userDefined.length - 1, Math.max(0, max - 1))].line;
            if (userDefined.length < min) {
                result.msg = message;
                if (!result.msg && (min < 2))
                    result.msg = 'Es darf ' + (min ? 'nur eine' : 'keine') + ' selbstdefinierte Anweisung geben, die nicht verwendet wird';
                if (!result.msg && (min >= 2))
                    result.msg = 'Es darf nicht so viele selbstdefinierte Anweisungen geben, die nicht verwendet werden';
                return result;
            }
            if (userDefined.length > max) {
                result.msg = messageTooMuch ? messageTooMuch : (message ? message : ('Es muss mehr selbstdefinierte Anweisungen geben, die nicht verwendet werden'));
                return result;
            }
            return null;
        });
    };

    /** React added user defined function not fullfilling requirements */
    CodeCheckers.ensureUserDefinedAdd = function(
        fine,               // Number of fine points to give if main procedure does not pass this test
        fktName,            // Name of user defined function to check
        compareWithCode,    // Code to compare code of user defined function with
        compareStrict,      // Optional: Set true if subsequences should not be collapsed and repeats not be expanded
        forceRunFast        // Optional: If set true or false, enforce usage of  keyword VERTIEFT or disallow it
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        let planed = SourceCodePart.tokensToSequenz(Token.tokenize(compareWithCode), true);
        if (!compareStrict)
            planed = planed.createCanonical();
        return (function({ compiled }) {
            const cmp = cmpUserDefined(compiled, fktName, planed, compareStrict, forceRunFast);
            if (!cmp)
                return null;
            const result = { accept: false, fine };
            switch (cmp.type) {
                case 'notDefined':
                    const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
                    if (noDef.length)
                        result.line = noDef[0].line;
                    result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' fehlt noch immer…';
                    return result;
                case 'runFast':
                    const def = getUserDefinedByName(compiled, fktName);
                    if (def instanceof SourceCodePart)
                        result.line = def.line;
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
                    result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
                    return result;
                case 'item':
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    result.msg = 'Gut, nun gibt es eine selbstdefinierte Anweisung ' + fktName + '. Aber was sie tut, passt noch nicht…';
                    return result;
                default:
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    return result;
            }
        });
    }

    /** Reject changed user defined function */
    CodeCheckers.ensureUserDefinedUnchanged = function(
        fine,               // Number of fine points to give if main procedure does not pass this test
        fktName,            // Name of user defined function to check
        compareWithCode,    // Code to compare code of user defined function with
        compareStrict,      // Optional: Set true if subsequences should not be collapsed and repeats not be expanded
        forceRunFast        // Optional: If set true or false, enforce usage of  keyword VERTIEFT or disallow it
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        let planed = SourceCodePart.tokensToSequenz(Token.tokenize(compareWithCode), true);
        if (!compareStrict)
            planed = planed.createCanonical();
        return (function({ compiled }) {
            const cmp = cmpUserDefined(compiled, fktName, planed, compareStrict, forceRunFast);
            if (!cmp)
                return null;
            const result = { accept: false, fine };
            switch (cmp.type) {
                case 'notDefined':
                    const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
                    if (noDef.length)
                        result.line = noDef[0].line;
                    result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' soll es weiterhin geben…';
                    return result;
                case 'runFast':
                    const def = getUserDefinedByName(compiled, fktName);
                    if (def instanceof SourceCodePart)
                        result.line = def.line;
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
                    result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
                    return result;
                case 'item':
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    result.msg = 'Eine Veränderung der selbstdefinierten Anweisung ' + fktName + ' ist hier falsch';
                    return result;
                default:
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    return result;
            }
        });
    }

    /** Ensure user defined function was changed in a required way */
    CodeCheckers.ensureUserDefinedChanged = function(
        fine,               // Number of fine points to give if main procedure does not pass this test
        fktName,            // Name of user defined function to check
        codeUnchanged,      // Code of unchanged user defined function
        codeChanged,        // Code of user defined function after required changes
        compareWithCode,    // Code to compare code of user defined function with
        compareStrict,      // Optional: Set true if subsequences should not be collapsed and repeats not be expanded
        forceRunFast        // Optional: If set true or false, enforce usage of  keyword VERTIEFT or disallow it
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        let planed = SourceCodePart.tokensToSequenz(Token.tokenize(codeChanged), true);
        let unchanged = SourceCodePart.tokensToSequenz(Token.tokenize(codeUnchanged), true);
        if (!compareStrict) {
            planed = planed.createCanonical();
            unchanged = unchanged.createCanonical();
        }
        return (function({ compiled }) {
            const cmp = cmpUserDefined(compiled, fktName, planed, compareStrict, forceRunFast);
            if (!cmp)
                return null;
            const result = { accept: false, fine };
            switch (cmp.type) {
                case 'notDefined':
                    const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
                    if (noDef.length)
                        result.line = noDef[0].line;
                    result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' soll es weiterhin geben…';
                    return result;
                case 'runFast':
                    const def = getUserDefinedByName(compiled, fktName);
                    if (def instanceof SourceCodePart)
                        result.line = def.line;
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
                    result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
                    return result;
                case 'item':
                    const unchangedCmp = cmpUserDefined(compiled, fktName, unchanged, compareStrict, forceRunFast) ? false : true;
                    if (!unchangedCmp) {
                        const def = getUserDefinedByName(compiled, fktName);
                        if (def instanceof SourceCodePart)
                            result.line = def.line;
                        result.msg = 'Um die Aufgabe zu lösen, musst du die selbstdefinierte Anweisung ' + fktName + ' verändern';
                        return result;
                    }
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    else {
                        const def = getUserDefinedByName(compiled, fktName);
                        if (def instanceof SourceCodePart)
                            result.line = def.line;
                    }
                    result.msg = 'Richtig, die selbstdefinierte Anweisung ' + fktName + ' muss verändert werden. Aber nicht so…';
                    return result;
                default:
                    result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
                    if (result.actualValue instanceof SourceCodePart)
                        result.line = result.actualValue.line;
                    return result;
            }
        });
    }

    /** Reject code without required user defined function */
    CodeCheckers.rejectMissingUserDefined = function(
        fine,       // Number of fine points to give if there is not the required user defined function
        fktName,    // Name of required user defined function
        message     // Optional: Message if there are additional user defined functions
    ) {
        return (function({ compiled }) {
            if (getUserDefinedByName(compiled, fktName))
                return null;
            return { accept: false, fine, msg: message ? message : ('Es muss eine selbstdefinierte Anweisung namens „' + fktName + '“ geben') };
        });
    }

    /** Reject additional user defined functions */
    CodeCheckers.rejectAdditionalUserDefined = function(
        fine,       // Number of fine points to give if there are additional user defined functions
        allowed,    // Array of allowed function names
        message     // Optional: Message if there are additional user defined functions
    ) {
        if (!Number.isFinite(fine))
            fine = 0;
        return (function({ compiled }) {
            const defs = compiled.getContentByClass(SourceCodeDefinition);
            const additional = defs.filter(def => (allowed.indexOf(def.name) == -1));
            if (!additional.length)
                return null;
            return { accept: false, fine, line: additional[0].line, msg: message ? message : ('Es darf keine zusätzliche Anweisung namens „' + additional[0].name + '“ geben') };
        });
    }

})();

/** Function creating a checking function to ensure useage of loop */
CodeCheckers.ensureRepeat = function(
    fine,               // Number of fine points to give if there is no loop or loop does not passe test
    times,              // Optional: Required number of repetitions
    compareWithCode,    // Optional: Code to compare loop body with
    compareStrict,      // Optional: Set true if subsequences should not be collapsed and repeats not be expanded for comparison
    message             // Optional: Message on fail
) {
    let target = ((typeof compareWithCode == 'string') || (compareWithCode instanceof String)) ? SourceCodePart.tokensToSequenz(Token.tokenize(compareWithCode), true) : null;
    if (!compareStrict && target)
        target = target.createCanonical();
    if (!Number.isFinite(fine))
        fine = 0;
    return (function({ compiled }) {
        const repeats = compiled.getContentByClass(SourceCodeIteration).reverse();
        return repeats.map(iteration => {
            const result = { fine, accept: false, line: iteration.line };
            if (Number.isFinite(times) && (times != iteration.times)) {
                result.msg = message ? message : ('Statt ' + iteration.times + ' müssen es ' + times + ' Wiederholungsdurchläufe sein');
                return result;
            }
            if (!target)
                return null;
            const diff = target.compare(iteration.body);
            if (!diff)
                return null;
            if (diff.actualValue instanceof SourceCodePart)
                result.line = diff.actualValue.line;
            else if (main.content.length)
                result.line = main.content[main.content.length - 1].line;
            if (diff.type != 'item') {
                result.msg = 'In der Wiederholung mit fester Anzal stimmt etwas nicht…';
                return result;
            }
            if (diff.actualValue instanceof SourceCodePart)
                result.msg = 'In der hervorgehobenen Zeile stimmt deine Lösung nicht';
            else
                result.msg = 'In der inneren Sequenz der Wiederholung mit fester Anzahl fehlt etwas';
            return result;
        }).reduce((prev, curr) => (prev && curr), { fine, accept: false, msg: message ? message : 'Eine Wiederholung mit fester Anzahl fehlt' });
    });
};

/** Creates a function checking all numbers in world after execution */
CodeCheckers.worldCheck = function(
    fine,           // Number of fine points to give if main procedure does not pass this test
    arrCounts,      // Flat array with required field numbers (Standardwert 0)
    isNoChange      // Optional: Describes arrCounts also state of world after execution if there is no change to code?
) {
    function checkWorld(worldStat) {
        if (!(arrCounts instanceof Array))
            arrCounts = [];
        let wrong = 0;
        let index = 0;
        for (let y = 0; y < worldStat.height; ++y) {
            for (let x = 0; x < worldStat.width; ++x) {
                if (worldStat.read(x, y) != ((index < arrCounts.length) ? arrCounts[index] : 0))
                    ++wrong;
                ++index;
            }
        }
        return wrong;
    }

    return function({ worldStat }) {
        const wrong = checkWorld(worldStat);
        const nameWrong = ['', 'einem Feld', 'zwei Feldern', 'drei Feldern', 'vier Feldern', 'fünf Feldern', 'sechs Feldern', 'sieben Feldern', 'acht Feldern', 'neun Feldern', 'zehn Feldern', 'elf Feldern', 'zwölf Feldern'];
        if (wrong) {
            const strWrong = ((wrong < nameWrong.length) ? nameWrong[wrong] : (wrong + ' Feldern'));
            const result = { accept: false, fine, msg: 'Bei ' + strWrong + ' passt die Plättchenanzahl nicht' };
            if (isNoChange)
                result.msg = 'Du hast zu viel verändert: ' + result.msg + ' mehr.';
            else
                result.msg += '.';
            return result;
        }
        return null;
    }
};

/** Function creating a checking function to ensure robot's working direction */
CodeCheckers.ensureRobotDirection = function(
    fine,               // Number of fine points to give if there is no loop or loop does not passe test
    direction,          // Required working direction of robot (0 meaning north, 1 meaning west, 2 meaning south and 3 meaning east)
    message             // Optional: Message on fail
) {
    if (!Number.isFinite(fine))
        fine = 0;
    direction = Math.abs(Math.round(direction) % 4);
    return (function({ robotDirection }) {
        if (robotDirection == direction)
            return null;
        const result = { fine, accept: false, msg: message };
        if (!message) {
            const directionStrings = ['Norden', 'Westen', 'Süden', 'Osten'];
            result.msg = 'Statt nach ' + directionStrings[robotDirection] + ' müsste der Roboter nach ' + directionStrings[direction] + ' schauen';
        }
        return result;
    });
};
    return this;
}).call({});
