/**
Copyright © 2023 Sebastian Roith <https://www.roith-unterrichtet.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

/***[ parseSourceCode/Token.js ] *************************/
/**
* A token of a source code
*/
function Token(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars in present line
    str         // String representation of content
) {
    this.line = line;
    this.char = char;
    this.str = str;
}

/** Number of line in source code (starting at 0) */
Token.prototype.line = 0;

/** Number of prevoiuse chars in present line */
Token.prototype.char = 0;

/** String representation of content */
Token.prototype.str = '';

/**
* Convert token to human readable text
* @return String    Human readable text
*/
Token.prototype.toString = function() {
    return this.str;
}

/**
* Find out if this token represents a key word
* @return Boolean       Does token represent a key word?
*/
Token.prototype.isKeyword = function() { return false; }

/**
* Is this token relevant for parsing source code
* @return Boolean       Relevant token?
*/
Token.prototype.isParsingRelevant = function() { return true }

/** Typical node name in syntax highlighted HTML representation of token */
Token.prototype.nodeName = 'span';

/**
* Add a new line to DocumentFragment instance df
* @return undefined
*/
Token.prototype._ensureLineHtml = function(
    df,         // DocumentFragment instance used to build syntax highlighted text representation
    allwaysAdd  // Optional: If true, add line even if there already is a line
) {
    if (!allwaysAdd && df.lastChild && (df.lastChild.nodeName == 'DIV') && df.lastChild.lastChild && (df.lastChild.lastChild.nodeName == 'SPAN'))
        return;
    const document = df.ownerDocument;
    const nodeLine = document.createElement('div');
    const nodeExtra = document.createElement('span');
    nodeLine.appendChild(nodeExtra);
    df.appendChild(nodeLine);
}

/**
* Add a syntax highlighted text representation of this token to DocumentFragment instance df
* @return undefined
*/
Token.prototype.toHtml = function(
    df          // DocumentFragment instance used to build syntax highlighted text representation
) {
    const document = df.ownerDocument;
    this._ensureLineHtml(df);
    if (this.nodeName) {
        const node = document.createElement(this.nodeName);
        node.appendChild(document.createTextNode(this.str));
        df.lastChild.lastChild.appendChild(node);
    } else
        df.lastChild.lastChild.appendChild(document.createTextNode(this.str));
}

/** Token classes ending other tokens */
Token.arrEndingTokenClasses = [];

/** Functions to convert string to token */
Token.arrTokenizers = [];

/**
* Convert whole string to tokens
* @return Array     Array of tokens or null on fail
*/
Token.tokenize = function(
    s           // String to convert to tokens
) {
    let env = { in: '\n' + s, line: 0, char: 0, out: [] }
    while (env.in.length) {
        if (!Token.arrTokenizers.reduce((prev, tokenizer) => (prev || tokenizer(env)), false))
            return null;
    }
    return env.out;
}

/**
* Get relevant indentation levels
* @return Array     All indentation levels (number of whitespace aquivalents) relevant for the next line sorted by indentation level ascending
*/
Token.getIndentLevels = function(
    s           // String containing all previous lines
) {
    const allTokens = Token.tokenize(s);
    const relevantTokens = allTokens.filter(token => token.isParsingRelevant()).filter((token, index, arr) => !((token instanceof TokenIndent) && (arr[index + 1] instanceof TokenIndent)) );
    const indents = relevantTokens.filter((token, index, arr) => ((token instanceof TokenIndent) && arr[index + 1] && !(arr[index + 1] instanceof TokenIndent))).map(token => token.indent).reverse();
    const result = [];
    indents.forEach((indent, index) => { if (!index || (indent < result[0])) result.unshift(indent); });
    return result;
}

/***[ parseSourceCode/keywordsTokens.js ] *************************/
/**
* Create a new class for keyword token
* @return Class     Created class
*/
function makeClass(
    keyword,        // Text representation of keyword
    isKeyword       // Is this a real keyword?
) {
    const keywordUppercase = ('' + keyword).toUpperCase();
    const keywordLength = keyword.length;
    const newClass = function(line, char, str) {
        Token.call(this, line, char, str);
    };
    newClass.prototype = Object.create(Token.prototype);
    newClass.prototype.constructor = newClass;
    newClass.prototype.nodeName = isKeyword ? 'em' : 'span';
    if (isKeyword)
        newClass.prototype.isKeyword = (() => true);
    newClass.tokenizer = ((env) => {
        const next = env.in.substring(0, keywordLength);
        if (next.toUpperCase() != keywordUppercase)
            return false;
        const remainingString = env.in.substring(keywordLength);
        if (Token.arrEndingTokenClasses.reduce((prev, t) => (prev && !t.tokenizer({ in: remainingString, out: [] })), remainingString.length))
            return false;
        env.out.push(new newClass(env.line, env.char, keyword));
        env.char += [...next].length;
        env.in = remainingString;
        return true;
    });
    return newClass;
}

// Create token classes for all keywords
window.TokenIf = makeClass('FALLS', true);
window.TokenThen = makeClass('DANN', true);
window.TokenElse = makeClass('SONST', true);
window.TokenDefinition = makeClass('LERNE', true);
window.TokenRunFast = makeClass('VERTIEFT', true);
window.TokenRepeat = makeClass('WIEDERHOLE', true);
window.TokenTimes = makeClass('MAL', true);
window.TokenWhile = makeClass('SOLANGE', true);

// Create token classes for brackets
window.TokenBracketOpen = makeClass('(', false);
window.TokenBracketClose = makeClass(')', false);

/***[ parseSourceCode/TokenIndent.js ] *************************/
/**
* Token for line start includeing indentation
*/
function TokenIndent(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars
    str,        // String representation of content
    indent      // Number of space equivalents
) {
    Token.call(this, line, char, str);
    this.indent = indent;
}
TokenIndent.prototype = Object.create(Token.prototype);
TokenIndent.prototype.constructor = TokenIndent;

// Number of space equivalents
TokenIndent.prototype.indent = 0;

/** Typical node name in syntax highlighted HTML representation of token */
TokenIndent.prototype.nodeName = 'span';

/**
* Add a syntax highlighted text representation of this token to DocumentFragment instance df
* @return undefined
*/
TokenIndent.prototype.toHtml = function(
    df          // DocumentFragment instance used to build syntax highlighted text representation
) {
    const document = df.ownerDocument;
    const lines = this.str.split('\n');
    lines.forEach((whitespaces, index) => {
        if (!index)
            return;
        this._ensureLineHtml(df, true);
        if (whitespaces.length) {
            const span = document.createElement(this.nodeName);
            span.appendChild(document.createTextNode(whitespaces));
            df.lastChild.lastChild.appendChild(span);
        }
    });
}

// Default Number of whitespace equivalents of a tab (following another tab)
TokenIndent.TAB_WIDTH = 4;

/**
* Count number of whitespace equivalents of whitespace only string whitespaces not having line breaks assumeing a tab has a maximum of tabWidth whitespace equivalents
* @return Number        Number of whitespace equivalents
*/
TokenIndent.countSpaceEquivalents = function(
    whitespaces,    // String only consisting of whitespaces without '\n'
    tabWidth        // Optional: Number of whitespace equivalents of a tab (following another tab); defaults to TokenIndent.TAB_WIDTH
) {
    if (!Number.isFinite(tabWidth))
        tabWidth = TokenIndent.TAB_WIDTH;
    return whitespaces.split('\t').reduce((prev, curr, index) => ((index ? (tabWidth * Math.floor(prev / tabWidth + 1)) : 0) + curr.length), 0);
}

/**
* Convert first chars of input string env.in to TokenIndent (if possible)
* @return Boolean       Created a token and reduced input string?
*/
TokenIndent.tokenizer = function(
    env         // Environment for converting to tokens: Object { in: 'Remaining input string', line: 'Number of completed lines', char: 'Number of completed chars in present line', out: 'Array of tokens' }
) {
    if (env.in.charAt(0) != '\n')
        return false;
    const match = (new RegExp('^\\s+')).exec(env.in);
    const lines = match[0].split('\n');
    const indent = lines[lines.length - 1];
    const remainingString = env.in.substring(match[0].length);
//    if (remainingString.length) {
        env.out.push(new TokenIndent(env.line, env.char, match[0], TokenIndent.countSpaceEquivalents(indent)));
        env.line += lines.length - 1;
        env.char = indent.length;
//    } else
//        env.out.push(new TokenWhitespace(env.line, env.char, match[0]));
    env.in = remainingString;
    return true;
}

/***[ parseSourceCode/TokenWhitespace.js ] *************************/
/**
* Token for whitespace not beeing new line or indentation
*/
function TokenWhitespace(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars in present line
    str         // String representation of content
) {
    Token.call(this, line, char, str);
}
TokenWhitespace.prototype = Object.create(Token.prototype);
TokenWhitespace.prototype.constructor = TokenWhitespace;

/** Typical node name in syntax highlighted HTML representation of token */
TokenWhitespace.prototype.nodeName = 'span';

/**
* Convert token to human readable text
* @return String    Human readable text
*/
TokenWhitespace.prototype.toString = function() {
    return ' ';
}

/**
* Is this token relevant for parsing source code
* @return Boolean       Relevant token?
*/
TokenWhitespace.prototype.isParsingRelevant = function() { return false; }

/**
* Convert first chars of input string env.in to TokenWhitespace (if possible)
* @return Boolean       Created a token and reduced input string?
*/
TokenWhitespace.tokenizer = function(
    env         // Environment for converting to tokens: Object { in: 'Remaining input string', line: 'Number of completed lines', char: 'Number of completed chars in present line', out: 'Array of tokens' }
) {
    const match = (new RegExp('^\\s+')).exec(env.in);
    if (!match)
        return false;
    const lines = match[0].split('\n');
    const whitespaces = lines[0].length;
    if (!whitespaces)
        return false;
    env.out.push(new TokenWhitespace(env.line, env.char, lines[0]));
    env.char += whitespaces;
    env.in = env.in.substring(whitespaces);
    return true;
}

/***[ parseSourceCode/TokenLineComment.js ] *************************/
/**
* Token for a comment terminated by end of line
*/
function TokenLineComment(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars in present line
    str         // String representation of content
) {
    Token.call(this, line, char, str);
}
TokenLineComment.prototype = Object.create(Token.prototype);
TokenLineComment.prototype.constructor = TokenLineComment;

/** Typical node name in syntax highlighted HTML representation of token */
TokenLineComment.prototype.nodeName = 'small';

/**
* Is this token relevant for parsing source code
* @return Boolean       Relevant token?
*/
TokenLineComment.prototype.isParsingRelevant = function() { return false; }

/** Possibilities to start comment */
TokenLineComment.start = ['#', '//', ';', '--'];

/**
* Convert first chars of input string env.in to TokenLineComment (if possible)
* @return Boolean       Created a token and reduced input string?
*/
TokenLineComment.tokenizer = function(
    env         // Environment for converting to tokens: Object { in: 'Remaining input string', line: 'Number of completed lines', char: 'Number of completed chars in present line', out: 'Array of tokens' }
) {
    const match = TokenLineComment.start.filter(start => (start == env.in.substring(0, start.length)));
    if (!match.length)
        return false;
    const posNewline = env.in.indexOf('\n');
    env.out.push(new TokenLineComment(env.line, env.char, (posNewline > 0) ? env.in.substring(0, posNewline) : env.in));
    env.char += posNewline;
    env.in = (posNewline > 0) ? env.in.substring(posNewline) : '';
    return true;
}

/**
* Convert token to human readable text
* @return String    Human readable text
*/
TokenLineComment.prototype.toString = function() {
    return TokenLineComment.start[0] + ' Zeilenkommentar';
}

/***[ parseSourceCode/TokenBlockComment.js ] *************************/
/**
* Token for a comment terminated by a special string
*/
function TokenBlockComment(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars in present line
    str         // String representation of content
) {
    Token.call(this, line, char, str);
}
TokenBlockComment.prototype = Object.create(Token.prototype);
TokenBlockComment.prototype.constructor = TokenBlockComment;

/** Typical node name in syntax highlighted HTML representation of token */
TokenBlockComment.prototype.nodeName = 'small';

/**
* Is this token relevant for parsing source code
* @return Boolean       Relevant token?
*/
TokenBlockComment.prototype.isParsingRelevant = function() { return false; }

/**
* Add a syntax highlighted text representation of this token to DocumentFragment instance df
* @return undefined
*/
TokenBlockComment.prototype.toHtml = function(
    df          // DocumentFragment instance used to build syntax highlighted text representation
) {
    const document = df.ownerDocument;
    this._ensureLineHtml(df);
    this.str.split('\n').forEach((s, index) => {
        if (index)
            this._ensureLineHtml(df, true);
        const node = document.createElement(this.nodeName);
        node.appendChild(document.createTextNode(s));
        df.lastElementChild.lastChild.appendChild(node);
    });
}

/** Possibilities delimiters of comment */
TokenBlockComment.pairs = [
    { start: '/*', end: '*/' },
    { start: '{', end: '}' }
];

/**
* Convert first chars of input string env.in to TokenBlockComment (if possible)
* @return Boolean       Created a token and reduced input string?
*/
TokenBlockComment.tokenizer = function(
    env         // Environment for converting to tokens: Object { in: 'Remaining input string', line: 'Number of completed lines', char: 'Number of completed chars in present line', out: 'Array of tokens' }
) {
    const match = TokenBlockComment.pairs.filter(({ start }) => (start == env.in.substring(0, start.length)));
    if (!match.length)
        return false;
    const lenStart = match[0].start.length;
    const strEnd = match[0].end;
    const posEnd = env.in.indexOf(strEnd, lenStart);
    const comment = env.in.substring(0, (posEnd > 0) ? (posEnd + strEnd.length) : undefined);
    const lines = comment.split('\n');
    const lenLastLine = lines[lines.length - 1].length;
    env.out.push(new TokenBlockComment(env.line, env.char, comment));
    if (lines.length > 1) {
        env.line += lines.length - 1;
        env.char = lenLastLine;
    } else
        env.char += lenLastLine;
    env.in = (posEnd > 0) ? env.in.substring(posEnd + strEnd.length) : '';
    return true;
}

/***[ parseSourceCode/TokenCommand.js ] *************************/
/**
* Token for a name of an command (or a number)
*/
function   TokenCommand(
    line,       // Number of line in source code (starting at 0)
    char,       // Number of prevoiuse chars
    name        // Name of command or number
) {
    Token.call(this, line, char, name);
    this.name = name;
}
TokenCommand.prototype = Object.create(Token.prototype);
TokenCommand.prototype.constructor =   TokenCommand;

// Name of command or number
Object.defineProperty(TokenCommand.prototype, 'name', {
    enumerable: true,
    get() { return this.str; }
});

/**
* Convert first chars of input string env.in to TokenBlockComment (if possible)
* @return Boolean       Created a token and reduced input string?
*/
TokenCommand.tokenizer = function(
    env         // Environment for converting to tokens: Object { in: 'Remaining input string', line: 'Number of completed lines', char: 'Number of completed chars in present line', out: 'Array of tokens' }
) {
    let len = 1;
    for (len; len < env.in.length; ++len) {
        const remainingString = env.in.substring(len);
        if (Token.arrEndingTokenClasses.reduce((prev, { tokenizer }) => (prev || tokenizer({ in: remainingString, out: [] })), false))
            break;
    }
    const next = env.in.substring(0, len);
    env.out.push(new TokenCommand(env.line, env.char, next));
    env.char += [...next].length;
    env.in = env.in.substring(len);
    return true;
}

/***[ parseSourceCode/registerClasses.js ] *************************/
/** Token classes ending other tokens */
Token.arrEndingTokenClasses = [TokenIndent, TokenWhitespace, TokenLineComment, TokenBlockComment, TokenBracketOpen, TokenBracketClose];

/** Functions to convert string to token */
Token.arrTokenizers = Token.arrEndingTokenClasses.slice().concat([TokenIf, TokenThen, TokenElse, TokenDefinition, TokenRunFast, TokenRepeat, TokenTimes, TokenWhile, TokenCommand]).map(({ tokenizer }) => tokenizer);

/***[ parseSourceCode/CompilingToken.js ] *************************/
/**
* Base class for tokens used for compiling
*/
function CompilingToken() {
}

/**
* Get representation in half bytes
* @return String    Hexadecimal representation of half bytes
*/
CompilingToken.prototype.toHex = function(
    env     // Information on converting: { numberLength: 'Number of half bytes for each numberic value', offset: 'Previously insterted additional half bytes' }
) {
    return '';
}

/**
* Get the biggest number represented by this token
* @return Number        Biggest number represent by this token
*/
CompilingToken.prototype.getMaxNumber = function() {
    return 0;
}

/** Does number of output 4 bit words depend on number length? */
CompilingToken.prototype.relevantLength = false;

/***[ parseSourceCode/CompilingTokenHalfbyte.js ] *************************/
/**
* Class for tokens while compiling representing a 4 bit value
*/
function CompilingTokenHalfbyte(
    value       // Value to represent
) {
    this.value = Math.max(Math.min(Math.floor(value), 15), 0);
}
CompilingTokenHalfbyte.prototype = Object.create(CompilingToken.prototype);
CompilingTokenHalfbyte.prototype.constructor = CompilingTokenHalfbyte;

/** Value to represent */
CompilingTokenHalfbyte.prototype.value = 0;

/**
* Get representation in half bytes
* @return String    Hexadecimal representation of half bytes
*/
CompilingTokenHalfbyte.prototype.toHex = function(
    env     // Information on converting: { numberLength: 'Number of half bytes for each numberic value', offset: 'Previously insterted additional half bytes' }
) {
    return this.value.toString(16);
}

/***[ parseSourceCode/CompilingTokenNumber.js ] *************************/
/**
* Class for tokens while compiling representing a line number or the number of iterations or an address
*/
function CompilingTokenNumber(
    value       // Optional: Number to represent (defaults to 0)
) {
    if (Number.isFinite(value))
        this.value = Math.max(Math.floor(value), 0);
}
CompilingTokenNumber.prototype = Object.create(CompilingToken.prototype);
CompilingTokenNumber.prototype.constructor = CompilingTokenNumber;

/** Value to represent */
CompilingTokenNumber.prototype.value = 0;

/**
* Get representation in half bytes
* @return String    Hexadecimal representation of half bytes
*/
CompilingTokenNumber.prototype.toHex = function(
    numberLength    // Number of half bytes for each numberic value
) {
    const hex = this.value.toString(16);
    return String('0').repeat(numberLength - hex.length) + hex;
}

/** Does number of output 4 bit words depend on number length? */
CompilingTokenNumber.prototype.relevantLength = true;

/**
* Get the biggest number represented by this token
* @return Number        Biggest number represent by this token
*/
CompilingToken.prototype.getMaxNumber = function() {
    return this.value;
}

/***[ parseSourceCode/CompilingEnvironment.js ] *************************/
/**
* Class for objects used to compile a source code to a string of hex characters
*/
function CompilingEnvironment() {
    this.tokens = [];
    this._addressTokens = [];
    this._userDefined = [];
}

/** CompilingToken instances already added */
CompilingEnvironment.prototype.tokens = [];

/** Number of half bytes for each numberic value (0 means unset) */
CompilingEnvironment.prototype.numberLength = 0;

/**
* Get index of token to add next
* @return Number    Index of token to add next
*/
CompilingEnvironment.prototype.getNextIndex = function() {
    return this.tokens.length;
}

/**
* Add a halfbyte
* @return Number    Index of added token
*/
CompilingEnvironment.prototype.addHalfbyte = function(
    value       // Value to add
) {
    if (this.numberLength)
        return this.getNextIndex() - 1;
    this.tokens.push(new CompilingTokenHalfbyte(value));
    return this.tokens.length - 1;
}

/**
* Add a CompilingToken containing a line number or the number of iterations
* @return CompilingTokenNumber      Added token
*/
CompilingEnvironment.prototype.addNumber = function(
    value       // Number to add
) {
    if (this.numberLength)
        return null;
    const token = new CompilingTokenNumber(value);
    this.tokens.push(token);
    return token;
}

/**
* Determine numberLength, if this has not already been done
* @return Boolean       Determined this.numberLength?
*/
CompilingEnvironment.prototype.determineNumberLength = function() {
    if (this.numberLength)
        return false;
    // Calculate additional number of halfbytes neccessary to represent all present numbers
    const maxNumber = this.tokens.reduce((prev, curr) => Math.max(prev, curr.getMaxNumber()), 0);
    const countNumbers = this.tokens.filter(token => (token instanceof CompilingTokenNumber)).length;
    let maxPossible = 15;
    while (maxNumber + this.numberLength * countNumbers > maxPossible) {
        ++this.numberLength;
        maxPossible = maxPossible * 16 + 15;
    }
    ++this.numberLength;
    return true;
}

/**
* Get string of hex representing code
* @return String    String of hex
*/
CompilingEnvironment.prototype.getExecuteable = function() {
    this.determineNumberLength();
    return this.tokens.reduce((prev, token) => (prev + token.toHex(this.numberLength)), (this.numberLength - 1).toString(16));
}

/***[ parseSourceCode/CompilingCodeEnvironment.js ] *************************/
/**
* Class for objects used to compile a source code to a string of hex characters
*/
function CompilingCodeEnvironment() {
    CompilingEnvironment.call(this);
    this._addressTokens = [];
    this._userDefined = [];
}
CompilingCodeEnvironment.prototype = Object.create(CompilingEnvironment.prototype);
CompilingCodeEnvironment.prototype.constructor = CompilingCodeEnvironment;

/** Address tokens thoose value depends on length of number tokens */
CompilingCodeEnvironment.prototype._addressTokens = [];

/** Information on user defined functions */
CompilingCodeEnvironment.prototype._userDefined = [];

/** Possible command for virtual processor */
CompilingCodeEnvironment.commands = ['returnTrue', 'returnFalse', 'canForward', 'isNot', 'isNorth', 'stillIterating', 'forward', 'inc', 'dec', 'turnLeft', 'jump', 'call', 'jumpIfFalse', 'startIteration', 'reportLine', 'reportStepDone'];

/**
* Add a command for the virtual processor
* @return Number    Index of added token
*/
CompilingCodeEnvironment.prototype.addCommand = function(
    cmd     // Name of command (see CompilingCodeEnvironment.commands)
) {
    return this.addHalfbyte(CompilingCodeEnvironment.commands.indexOf(cmd));
}

/**
* Add a CompilingToken containing an address of another CompilingToken instance
* @return CompilingTokenNumber      Added token
*/
CompilingCodeEnvironment.prototype.addTokenAddress = function(
    index       // Optional: Token index to represent (can be set later by changing result's value attribute)
) {
    if (this.numberLength)
        return null;
    const token = new CompilingTokenNumber(index);
    this._addressTokens.push(token);
    this.tokens.push(token);
    return token;
}

/**
* Add address of user defined function to tokens
* @return undefined
*/
CompilingCodeEnvironment.prototype.addAddressFunction = function(
    fktName         // Name of user defined function
) {
    if (this.numberLength)
        return;
    let infos = this._userDefined.reduce((prev, item) => (prev || ((item.fktName == fktName) ? item : null)), null);
    if (!infos) {
        infos = { fktName, token: new CompilingTokenNumber(), index: -1 };
        this._addressTokens.push(infos.token);
        this._userDefined.push(infos);
    }
    this.tokens.push(infos.token);
}

/**
* Set token index of a user defined function's start to next token to insert
* @return undefined
*/
CompilingCodeEnvironment.prototype.registerStartOfFunction = function(
    fktName         // Name of user defined function
) {
    if (this.numberLength)
        return;
    let infos = this._userDefined.reduce((prev, item) => (prev || ((item.fktName == fktName) ? item : null)), null);
    if (infos)
        infos.index = this.getNextIndex();
    else {
        infos = { fktName, token: new CompilingTokenNumber(), index: this.getNextIndex() };
        this._addressTokens.push(infos.token);
        this._userDefined.push(infos);
    }
    infos.token.value = infos.index;
}

/**
* Get string of hex representing code
* @return String    String of hex
*/
CompilingCodeEnvironment.prototype.getExecuteable = function() {
    if (this.determineNumberLength()) {
        // Convert indizes to halfbyte numbers
        const countAdditional = this.numberLength - 1;
        this._addressTokens.forEach(token => (token.value += countAdditional * this.tokens.slice(0, token.value).filter(t => (t instanceof CompilingTokenNumber)).length));
    }
    return this.tokens.reduce((prev, token) => (prev + token.toHex(this.numberLength)), (this.numberLength - 1).toString(16));
}

/***[ parseSourceCode/SourceCodePart.js ] *************************/
/**
* Base class for logical parts of source code
*/
function SourceCodePart(
    line        // Associated line number
) {
    this.line = line;
}

/** Associated line number */
SourceCodePart.prototype.line = 0;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodePart.prototype.getFirstError = function() {
    return null;
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodePart.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodePart.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    return this;
}

/**
* Replace attribute value of attribute attrName by result of its method ensureDefinedCalls called if it is an SourceCodePart instance
* @return undefined
*/
SourceCodePart.prototype._ensureDefinedCalls = function(
    attrName,   // Name of attribute to check
    defined     // Names of defined functions (includeing build in commands)
) {
    if (this[attrName] instanceof SourceCodePart)
        this[attrName] = this[attrName].ensureDefinedCalls(defined);
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodePart.prototype.getLineNumbers = function() {
    return [ this.line ];
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodePart.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    return [];
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodePart.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof this.prototype))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: this.prototype, type: 'class' };
    return null;
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodePart.prototype.clone = function() {
    return new SourceCodePart(this.line);
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodePart.prototype.createCanonical = function() {
    return this.clone();
}

/** Derived classes beeing possible parts of source code (excluding SourceCodeSequence) */
SourceCodePart.standalones = [];

/**
* Convert stream of tokens to semantik tree
* @return SourceCodeSequence|SourceCodeError    …
*/
SourceCodePart.tokensToSequenz = function(
    tokens,         // Array of all tokens (possibly includeing tokens for whitespace, comments etc.)
    allowUndefined  // Optional: Discard check for undefined commands? (Default: false)
) {
    const relevantTokens = tokens.filter(token => token.isParsingRelevant()).filter((token, index, arr) => !((token instanceof TokenIndent) && (arr[index + 1] instanceof TokenIndent)) );
    if (relevantTokens[relevantTokens.length - 1] instanceof TokenIndent)
        relevantTokens.pop();
    const defined = SourceCodeCommand.getBuildIn();
    // Create main sequence
    const sequence = SourceCodeSequence.fromTokens(-1, relevantTokens, defined);
    // Make sure no sequence beginns after main sequence
    if ((sequence instanceof SourceCodeSequence) && relevantTokens.length)
        sequence.content.push(new SourceCodeErrorMisplaced(relevantTokens[0]));
    // Ensure all command calls are defined
    return allowUndefined ? sequence : sequence.ensureDefinedCalls(defined);
}

/**
* If possible create a SourceCodePart instance using shifting tokens from array tokens
* @return SourceCodePart|SourceCodeError|null   Created instance, error message or null
*/
SourceCodePart.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined,    // Names of already defined functions
    allowDef    // Is SourceCodeDefinition instance allowed?
) {
    const part = SourceCodePart.standalones.reduce((prev, curr) => (prev || curr.fromTokens(indent, tokens, defined)), null);
    if (!allowDef && (part instanceof SourceCodeDefinition))
        return new SourceCodeErrorMisplacedDefinition(part.line);
    if (part)
        return part;
    if (tokens[0] instanceof TokenIndent)
        return new SourceCodeErrorMissing(null, tokens[0].line);
    if (tokens[0])
        return new SourceCodeErrorMisplaced(tokens.shift());
    return null;
}

/***[ parseSourceCode/SourceCodeError.js ] *************************/
/**
* Base class for syntax errors
*/
function SourceCodeError(
    line        // Associated line number in source code
) {
    SourceCodePart.call(this, line);
}
SourceCodeError.prototype = Object.create(SourceCodePart.prototype);
SourceCodeError.prototype.constructor = SourceCodeError;

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeError.prototype.clone = function() {
    return new SourceCodeError(this.line);
}

/***[ parseSourceCode/SourceCodeErrorMissing.js ] *************************/
/**
* Class for errors resulting of missing tokens
*/
function SourceCodeErrorMissing(
    what,       // Class (Token* or Number or null for any) indicating type of missing token
    line        // Number of line in which token is missing
) {
    SourceCodeError.call(this, line);
    this.what = what;
}
SourceCodeErrorMissing.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorMissing.prototype.constructor = SourceCodeErrorMissing;

/** Class (Token* or Number or null for any) indicating type of missing token */
SourceCodeErrorMissing.prototype.what = null;

/**
* Menschenlesbare Fehlermeldung generieren
* @return String        Menschenlesbare Fehlermeldung
*/
SourceCodeErrorMissing.prototype.toString = function() {
    let msg = 'Hier fehlt doch ';
    switch(this.what) {
        case Number:
            msg += 'direkt nach WIEDERHOLE die Anzahl der Wiederholungen';
            break;
        case TokenCommand:
            msg += 'direkt nach LERNE der Name der neuen Anweisung';
            break;
        case TokenBracketClose:
            msg += 'direkt nach der öffnenden die schließende Klammer';
            break;
        case TokenTimes:
            msg += 'direkt der Anzahl an Wiederholungen das Wort MAL';
            break;
        case TokenIndent:
            msg = 'Hier steht zu viel in einer Zeile';
            break;
        default:
            msg += 'etwas';
            break;
    }
    return msg + ', oder?';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorMissing.prototype.clone = function() {
    return new SourceCodeErrorMissing(this.what, this.line);
}

/***[ parseSourceCode/SourceCodeErrorMisplaced.js ] *************************/
/**
* Class for errors resulting of misplaced tokens
*/
function SourceCodeErrorMisplaced(
    token       // Token instance beeing misplaced
) {
    SourceCodeError.call(this, token.line);
    this.token = token;
}
SourceCodeErrorMisplaced.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorMisplaced.prototype.constructor = SourceCodeErrorMisplaced;

// Token instance beeing misplaced
SourceCodeErrorMisplaced.prototype.token = null;

/**
* Generate human readable error message
* @return String        Human readable error message
*/
SourceCodeErrorMisplaced.prototype.toString = function() {
    return 'An dieser Stelle ergibt „' + this.token + '“ keinen Sinn…';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorMisplaced.prototype.clone = function() {
    return new SourceCodeErrorMisplaced(this.token);
}

/***[ parseSourceCode/SourceCodeErrorMisplacedDefinition.js ] *************************/
/**
* Class for errors resulting of nested definitions or definitions following procedure calls, loops or conditional statements
*/
function SourceCodeErrorMisplacedDefinition(
    line        // Number of line in which the problematic definition occured
) {
    SourceCodeError.call(this, line);
}
SourceCodeErrorMisplacedDefinition.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorMisplacedDefinition.prototype.constructor = SourceCodeErrorMisplacedDefinition;

/**
* Menschenlesbare Fehlermeldung generieren
* @return String        Menschenlesbare Fehlermeldung
*/
SourceCodeErrorMisplacedDefinition.prototype.toString = function() {
    return 'An dieser Stelle kann ich doch nichts Neues lernen, oder? Das geht nur ganz zu Beginn, bevor ich irgendetwas anderes tue.';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorMisplacedDefinition.prototype.clone = function() {
    return new SourceCodeErrorMisplacedDefinition(this.line);
}

/***[ parseSourceCode/SourceCodeErrorNoNumber.js ] *************************/
/**
* Class for errors resulting of non numberic declaration of iteration times
*/
function SourceCodeErrorNoNumber(
    command,    // String that was used instead of a number
    line        // Number of line in which command starts
) {
    SourceCodeError.call(this, line);
    this.noNumber = command;
}
SourceCodeErrorNoNumber.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorNoNumber.prototype.constructor = SourceCodeErrorNoNumber;

/** String that was used instead of a number */
SourceCodeErrorNoNumber.prototype.noNumber = '';

/**
* Menschenlesbare Fehlermeldung generieren
* @return String        Menschenlesbare Fehlermeldung
*/
SourceCodeErrorNoNumber.prototype.toString = function() {
    return 'Statt „' + this.noNumber + '“ muss doch die Anzahl an Wiederholungen angegeben werden, oder?';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorNoNumber.prototype.clone = function() {
    return new SourceCodeErrorNoNumber(this.noNumber, this.line);
}

/***[ parseSourceCode/SourceCodeErrorUndefined.js ] *************************/
/**
* Class for errors resulting of function calls on not defined functions
*/
function SourceCodeErrorUndefined(
    fktName,    // Not allowed name of user defined function
    line        // Number of line in which the problematic definition occured
) {
    SourceCodeError.call(this, line);
    this.fktName = fktName;
}
SourceCodeErrorUndefined.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorUndefined.prototype.constructor = SourceCodeErrorUndefined;

/** Not allowed name of user defined function */
SourceCodeErrorUndefined.prototype.fktName = '';

/**
* Menschenlesbare Fehlermeldung generieren
* @return String        Menschenlesbare Fehlermeldung
*/
SourceCodeErrorUndefined.prototype.toString = function() {
    return 'Was bedeutet „' + this.fktName + '“? Das habe ich leider nicht gelernt…';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorUndefined.prototype.clone = function() {
    return new SourceCodeErrorUndefined(this.fktName, this.line);
}

/***[ parseSourceCode/SourceCodeErrorRedefined.js ] *************************/
/**
* Class for errors resulting of definitions using previously defined function names
*/
function SourceCodeErrorRedefined(
    fktName,    // Not allowed name of user defined function
    line        // Number of line in which the problematic definition occured
) {
    SourceCodeError.call(this, line);
    this.fktName = fktName;
}
SourceCodeErrorRedefined.prototype = Object.create(SourceCodeError.prototype);
SourceCodeErrorRedefined.prototype.constructor = SourceCodeErrorRedefined;

/** Not allowed name of user defined function */
SourceCodeErrorRedefined.prototype.fktName = '';

/**
* Menschenlesbare Fehlermeldung generieren
* @return String        Menschenlesbare Fehlermeldung
*/
SourceCodeErrorRedefined.prototype.toString = function() {
    return 'Warum steht hier, dass ich erneut „' + this.fktName + '“ lernen soll? Das kann ich doch bereits…';
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeErrorRedefined.prototype.clone = function() {
    return new SourceCodeErrorRedefined(this.fktName, this.line);
}

/***[ parseSourceCode/SourceCodeCommand.js ] *************************/
/**
* A single call of a function within source code
*/
function SourceCodeCommand(
    fktName,    // Name of command to execute (or define)
    line        // Associated line number
) {
    SourceCodePart.call(this, line)
    this.fktName = fktName;
}
SourceCodeCommand.prototype = Object.create(SourceCodePart.prototype);
SourceCodeCommand.prototype.constructor = SourceCodeCommand;

// Name of command to execute (or define)
SourceCodeCommand.prototype.fktName = '';

/**
* Check if this instance represents a call of an build in function
* @return Boolean       Call of build in function?
*/
SourceCodeCommand.prototype.isBuildIn = function() {
    return SourceCodeCommand._buildIn.reduce((prev, { fktName }) => (prev || (fktName == this.fktName)), false);
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeCommand.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    if (context.reportLine) {
        ce.addCommand('reportLine');
        ce.addNumber(this.line);
    }
    if (context.runSlow)
        ce.addCommand('reportStepDone');
    const compile = SourceCodeCommand._buildIn.reduce((prev, { fktName, compile }) => (prev || ((fktName == this.fktName) ? compile : null)), null);
    if (compile)
        compile.call(this, ce, context);
    else {
        ce.addCommand('call');
        ce.addAddressFunction(this.fktName);
        if (context.reportLine) {
            ce.addCommand('reportLine');
            ce.addNumber(this.line);
        }
    }
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeCommand.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    if (!(this.fktName instanceof SourceCodeError) && (defined.indexOf(this.fktName) == -1))
        return new SourceCodeErrorUndefined(this.fktName, this.line);
    return this;
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeCommand.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeCommand))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeCommand, type: 'class' };
    if (compareWith.fktName != this.fktName)
        return { index: -1, actualValue: compareWith.fktName, plannedValue: this.fktName, type: 'name' };
    return null;
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeCommand.prototype.clone = function() {
    return new SourceCodeCommand(this.fktName, this.line);
}

/** Built in commands */
SourceCodeCommand._buildIn = [
    { fktName: 'istBegehbar', compile: ce => ce.addCommand('canForward') },
    { fktName: 'istBelegt', compile: ce => { ce.addCommand('isNot'); ce.addHalfbyte(0); } },
    { fktName: 'istPlatz', compile: ce => { ce.addCommand('isNot'); ce.addHalfbyte(15); } },
    { fktName: 'istNorden', compile: ce => ce.addCommand('isNorth') },
    { fktName: 'vor', compile: ce => ce.addCommand('forward') },
    { fktName: 'plus', compile: ce => ce.addCommand('inc') },
    { fktName: 'minus', compile: ce => ce.addCommand('dec') },
    { fktName: 'linksrum', compile: ce => ce.addCommand('turnLeft') },
    { fktName: 'abbruch', compile: (ce, context) => { ce.addCommand('returnFalse'); } },
    { fktName: 'fertig', compile: (ce, context) => { ce.addCommand('returnTrue'); } }
];

/**
* Get names of build in commands
* @return Array     All names of build in commands
*/
SourceCodeCommand.getBuildIn = function() {
    return SourceCodeCommand._buildIn.map(({ fktName }) => fktName);
}

/**
* If possible create a SourceCodeCommand instance using shifting tokens from array tokens
* @return SourceCodeCommand|SourceCodeError|null    Created instance, error message or null
*/
SourceCodeCommand.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined     // Names of already defined functions
) {
    if (!(tokens[0] instanceof TokenCommand))
        return null;
    const startToken = tokens.shift();
    if (tokens[0] instanceof TokenBracketOpen) {
        tokens.shift();
        if ((tokens[0] instanceof TokenIndent) && (tokens[1] instanceof TokenBracketClose)) {
            tokens.shift();
            tokens.shift();
        } else if (tokens[0] instanceof TokenBracketClose)
            tokens.shift();
        else if (tokens[0] instanceof TokenIndent)
            return new SourceCodeErrorMissing(TokenBracketClose, tokens[0].line);
        else
            return new SourceCodeErrorMisplaced(tokens.shift());
    }
    return new SourceCodeCommand(startToken.name, startToken.line);
}

// Add this class to possible source code parts
SourceCodePart.standalones.push(SourceCodeCommand);

/***[ parseSourceCode/SourceCodeConditional.js ] *************************/
/**
* A loop with a fixed number of repetitions within source code
*/
function SourceCodeConditional(
    condition,  // SourceCodeCommand instance beeing condition to eval
    bodyThen,   // SourceCodeSequence to execute if condition evals true
    bodyElse,   // SourceCodeSequence to execute if condition evals false (or null, if there is none)
    line,       // Line number of if statement
    lineElse    // Line number of else key word (or no Number, if there ist non)
) {
    SourceCodePart.call(this, line)
    this.condition = condition;
    this.bodyThen = bodyThen;
    this.bodyElse = bodyElse;
    if (Number.isFinite(lineElse))
        this.lineElse = lineElse;
}
SourceCodeConditional.prototype = Object.create(SourceCodePart.prototype);
SourceCodeConditional.prototype.constructor = SourceCodeConditional;

// SourceCodeCommand instance beeing condition to eval
SourceCodeConditional.prototype.condition = null;

// SourceCodeSequence to execute if condition evals true
SourceCodeConditional.prototype.bodyThen = null;

// SourceCodeSequence to execute if condition evals false (or null, if there is none)
SourceCodeConditional.prototype.bodyElse = null;

// Line number of else key word (or no Number, if there ist non)
SourceCodeConditional.prototype.lineElse = false;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodeConditional.prototype.getFirstError = function() {
    const partError = [this.condition, this.bodyThen, this.bodyElse].filter(err => (err instanceof SourceCodeError)).reduce((a, b) => (a || b), null);
    if (partError)
        return partError;
    return this.bodyThen.getFirstError() || (this.bodyElse ? this.bodyElse.getFirstError() : null);
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeConditional.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    this.condition._compile(ce, context);
    ce.addCommand('jumpIfFalse');
    let jump2End = ce.addTokenAddress();
    this.bodyThen._compile(ce, context);
    if (this.bodyElse) {
        const jump2Else = jump2End;
        ce.addCommand('jump');
        jump2End = ce.addTokenAddress();
        jump2Else.value = ce.getNextIndex();
        this.bodyElse._compile(ce, context);
    }
    jump2End.value = ce.getNextIndex();
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeConditional.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    this._ensureDefinedCalls('condition', defined);
    this._ensureDefinedCalls('bodyThen', defined);
    this._ensureDefinedCalls('bodyElse', defined);
    return this;
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodeConditional.prototype.getLineNumbers = function() {
    let range = [ this.line ];
    range = range.concat(this.bodyThen.getLineNumbers());
    if (Number.isFinite(this.lineElse))
        range.push(this.lineElse);
    if (this.bodyElse instanceof SourceCodePart)
        range = range.concat(this.bodyElse.getLineNumbers());
    return range;
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodeConditional.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    if (Number.isFinite(parentLine)) {
        if (parentLine == this.line)
            return this.bodyThen.getLineNumbers();
        if (Number.isFinite(this.lineElse) && (parentLine == this.lineElse))
            return (this.bodyElse instanceof SourceCodePart) ? this.bodyElse.getLineNumbers() : [];
        if (this.bodyElse instanceof SourceCodePart)
            return this.bodyThen.getNestedLines(parentLine) || this.bodyElse.getNestedLines(parentLine);
        return this.bodyThen.getNestedLines(parentLine);
    }
    const linesThen = this.bodyThen.getLineNumbers();
    if (this.bodyElse instanceof SourceCodePart)
        return linesThen.concat(this.bodyElse.getLineNumbers());
    return linesThen;
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeConditional.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeConditional))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeConditional, type: 'class' };
    return ['condition', 'bodyThen', 'bodyElse'].reduce((prev, key) => {
        if (prev)
            return prev;
        const conditionCmp = this[key].compare(compareWith[key]);
        if (conditionCmp)
            return { index: -1, actualValue: compareWith[key], plannedValue: this[key], type: key };
    }, null);
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeConditional.prototype.clone = function() {
    return new SourceCodeConditional(this.condition.clone(), this.bodyThen.clone(), this.bodyElse.clone(), this.line, this.lineElse);
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodeConditional.prototype.createCanonical = function() {
    return new SourceCodeConditional(this.condition.createCanonical(), this.bodyThen.createCanonical(), this.bodyElse.createCanonical(), this.line, this.lineElse);
}

/**
* If possible create a SourceCodeConditional instance using shifting tokens from array tokens
* @return SourceCodeConditional|SourceCodeError|null    Created instance, error message or null
*/
SourceCodeConditional.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined     // Names of already defined functions
) {
    if (!(tokens[0] instanceof TokenIf))
        return null;
    const line = tokens.shift().line;
    const condition = SourceCodeCommand.fromTokens(indent, tokens, defined);
    if (tokens[0] instanceof TokenThen)
        tokens.shift();
    const bodyThen = SourceCodeSequence.fromTokens(indent + 1, tokens, defined);
    let bodyElse = null;
    let lineElse = false;
    if ((tokens[0] instanceof TokenIndent) && (tokens[0].indent == indent) && (tokens[1] instanceof TokenElse)) {
        tokens.shift();
        lineElse = tokens.shift().line;
        bodyElse = SourceCodeSequence.fromTokens(indent + 1, tokens, defined);
    }
    return new SourceCodeConditional(condition, bodyThen, bodyElse, line, lineElse);
}

// Add this class to possible source code parts
SourceCodePart.standalones.push(SourceCodeConditional);

/***[ parseSourceCode/SourceCodeIteration.js ] *************************/
/**
* A loop with a fixed number of repetitions within source code
*/
function SourceCodeIteration(
    times,      // Number of loop runs
    body,       // SourceCodeSequence beeing looped
    line        // Associated line number
) {
    SourceCodePart.call(this, line)
    this.times = times;
    this.body = body;
}
SourceCodeIteration.prototype = Object.create(SourceCodePart.prototype);
SourceCodeIteration.prototype.constructor = SourceCodeIteration;

// Number of loop runs
SourceCodeIteration.prototype.times = 0;

// SourceCodeSequence beeing looped
SourceCodeIteration.prototype.body = null;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodeIteration.prototype.getFirstError = function() {
    if (this.times instanceof SourceCodeError)
        return this.times;
    if (this.body instanceof SourceCodeError)
        return this.body;
    return this.body.getFirstError();
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeIteration.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    ce.addCommand('startIteration');
    ce.addNumber(this.times);
    context.iterationLevel++;
    const indexStart = ce.getNextIndex();
    if (context.reportLine) {
        ce.addCommand('reportLine');
        ce.addNumber(this.line);
    }
    if (context.runSlow)
        ce.addCommand('reportStepDone');
    ce.addCommand('stillIterating');
    ce.addCommand('jumpIfFalse');
    const jump2End = ce.addTokenAddress();
    this.body._compile(ce, context);
    ce.addCommand('jump');
    ce.addTokenAddress(indexStart);
    jump2End.value = ce.getNextIndex();
    context.iterationLevel = Math.max(0, context.iterationLevel - 1);
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeIteration.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    this._ensureDefinedCalls('body', defined);
    return this;
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodeIteration.prototype.getLineNumbers = function() {
    return [ this.line ].concat(this.body.getLineNumbers());
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodeIteration.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    if (Number.isFinite(parentLine))
        return (parentLine == this.line) ? this.body.getLineNumbers() : this.body.getNestedLines(parentLine);
    return this.body.getLineNumbers();
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeIteration.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeIteration))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeIteration, type: 'class' };
    if (compareWith.times != this.times)
        return { index: -1, actualValue: compareWith.times, plannedValue: this.times, type: 'times' };
    const bodyCmp = this.body.compare(compareWith.body);
    if (bodyCmp)
        return { index: bodyCmp.index, actualValue: compareWith.body, plannedValue: this.body, type: 'item' };
    return null;
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeIteration.prototype.clone = function() {
    return new SourceCodeIteration(this.times, this.body.clone(), this.line);
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodeIteration.prototype.createCanonical = function() {
    if (this.times <= 0)
        return new SourceCodeSequence(this.line, this.body.indent);
    const canonical = this.body.createCanonical();
    const clone = canonical.clone();
    for (let n = 0; n < this.times; ++n)
        canonical.concat(clone.clone(), this.line);
    return canonical;
}

/**
* If possible create a SourceCodeIteration instance using shifting tokens from array tokens
* @return SourceCodeIteration|SourceCodeError|null    Created instance, error message or null
*/
SourceCodeIteration.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined     // Names of already defined functions
) {
    if (!(tokens[0] instanceof TokenRepeat))
        return null;
    const line = tokens.shift().line;
    const timesToken = SourceCodeCommand.fromTokens(indent, tokens, defined);
    let times;
    switch (true) {
        case (timesToken instanceof SourceCodeCommand):
            times = (new RegExp('^\\d+$')).exec(timesToken.fktName) ? (+timesToken.fktName) : (new SourceCodeErrorNoNumber(timesToken.fktName, line));
            break;
        case (timesToken instanceof SourceCodeError):
            times = timesToken;
            break;
        default:
            times = new SourceCodeErrorMissing(Number, line);
            break;
    }
    let body;
    if (tokens[0] instanceof TokenTimes) {
        tokens.shift();
        body = SourceCodeSequence.fromTokens(indent + 1, tokens, defined);
    } else
        body = new SourceCodeErrorMissing(TokenTimes, line);
    return new SourceCodeIteration(times, body, line);
}

// Add this class to possible source code parts
SourceCodePart.standalones.push(SourceCodeIteration);

/***[ parseSourceCode/SourceCodeSequence.js ] *************************/
/**
* Class for sequences in source code
*/
function SourceCodeSequence(
    line,       // Associated line number
    indent      // Indent (in character unit)
) {
    SourceCodePart.call(this, line)
    this.indent = indent;
    this.content = [];
}
SourceCodeSequence.prototype = Object.create(SourceCodePart.prototype);
SourceCodeSequence.prototype.constructor = SourceCodeSequence;

// SourceCodePart instances directly contained in this sequence
SourceCodeSequence.prototype.content = [];

// Indent (meassured in character unit) of this sequence
SourceCodeSequence.prototype.indent;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodeSequence.prototype.getFirstError = function() {
    return this.content.reduce((prev, curr) => (prev || ((curr instanceof SourceCodeError) ? curr : curr.getFirstError())), null);
}

/**
* Compile this sequence to executeable instructions
* @return String|SourceCodeError    String consisting of hexadecimal digits or compiling error
*/
SourceCodeSequence.prototype.compile = function() {
    let error = this.getFirstError();
    if (error)
        return error;
    const ce = new CompilingCodeEnvironment();
    const defs = this.content.filter(part => (part instanceof SourceCodeDefinition));
    defs.forEach(def => def._compileLearning(ce, { runSlow: true, iterationLevel: 0, reportLine: true }));
    this._compile(ce, { runSlow: true, iterationLevel: 0, reportLine: true });
    ce.addCommand('returnTrue');
    defs.forEach(def => def._compile(ce, { runSlow: true, iterationLevel: 0, reportLine: true }));
    return ce.getExecuteable();
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeSequence.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    this.content.filter(part => !(part instanceof SourceCodeDefinition)).forEach(part => part._compile(ce, context));
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeSequence.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    this.content = this.content.map(part => part.ensureDefinedCalls(defined));
    return this;
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodeSequence.prototype.getLineNumbers = function() {
    return this.content.reduce((prev, part) => prev.concat(part.getLineNumbers()), []);
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodeSequence.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    if (Number.isFinite(parentLine))
        return this.content.reduce((prev, part) => (prev || part.getNestedLines(parentLine)), []);
    return this.getLineNumbers();
}

/**
* Get array of directly nested objects beein instance of objClass
* @return Array     Array of objClass instances
*/
SourceCodeSequence.prototype.getContentByClass = function(
    objClass        // Parent class of wanted content
) {
    return this.content.filter(obj => (obj instanceof objClass));
}

/**
* Add a SourceCodePart instance to this sequence (if it ist a SourceCodeSequence instance, append its content instead)
* @return SourceCodeSequence    this
*/
SourceCodeSequence.prototype.concat = function(
    part,       // SourceCodePart instance to append to this sequence
    setLine     // Line number to set for very first added SourceCodePart instance
) {
    const index = this.content.length;
    if (part instanceof SourceCodeSequence)
        this.content.concat(part.content);
    else if (part instanceof SourceCodePart)
        this.content.push(part);
    if (Number.isFinite(setLine) && (index < this.content.length))
        this.content[index].line = setLine;
    return this;
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeSequence.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeSequence))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeSequence, type: 'class' };
    for (let index = 0; index < this.content.length; ++index) {
        const actualValue = compareWith.content[index];
        const plannedValue = this.content[index];
        if (!(actualValue instanceof SourceCodePart) || actualValue.compare(plannedValue))
            return { index, actualValue, plannedValue, type: 'item' };
    }
    if (compareWith.content.length > this.content.length)
        return { index: this.content.length, actualValue: compareWith.content[this.content.length], plannedValue: undefined, type: 'item' };
    return null;
}

/**
* Create an absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeSequence.prototype.clone = function() {
    const clone = new SourceCodeSequence(this.line, this.indent);
    clone.content = this.content.map(part => part.clone());
    return clone;
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodeSequence.prototype.createCanonical = function() {
    const canonical = new SourceCodeSequence(this.line, this.indent);
    this.content.forEach(part => canonical.concat(part.createCanonical()));
    return canonical;
}

/**
* Create a copy of this instance containing only parts not beeing a definition
* @return SourceCodePart    Copy without definition
*/
SourceCodeSequence.prototype.copyWithoutDefinitions = function() {
    const copy = new SourceCodeSequence(this.line, this.indent);
    copy.content = this.content.filter(part => !(part instanceof SourceCodeDefinition)).map(part => part.clone());
    return copy;
}

/**
* Get minimal indent in remaining tokens before of an indent beeing smaller than minIndent
* @return Number    Minimal indent (meassured in character units) found (or Infinity)
*/
SourceCodeSequence.getNextMinIndent = function(
    minIndent,      // Minimal indent possible for this sequence
    tokens          // Remaining tokens
) {
    let indent = Infinity;
    for (let n = 0; n < tokens.length; ++n) {
        if (!(tokens[n] instanceof TokenIndent))
            continue;
        if (tokens[n].indent < minIndent)
            break;
        if (Number.isFinite(tokens[n].indent))
            indent = Math.min(indent, tokens[n].indent);
    }
    return indent;
}

/**
* If possible create a (possibly empty) SourceCodeSequence instance using shifting tokens from array tokens
* @return SourceCodeSequence|SourceCodeError  Created instance or error message
*/
SourceCodeSequence.fromTokens = function(
    minIndent,      // Minimal indent possible for this sequence
    tokens,         // Remaining tokens
    defined         // Names of already defined functions
) {
    if (!tokens.length)
        return new SourceCodeSequence(0);
    if (!(tokens[0] instanceof TokenIndent)) {
        const line = tokens[0].line;
        while (tokens.length && !(tokens[0] instanceof TokenIndent))
            tokens.shift();
        return new SourceCodeErrorMissing(TokenIndent, line);
    }
    const thisIndent = SourceCodeSequence.getNextMinIndent(minIndent, tokens);
    if (!Number.isFinite(thisIndent))
        return new SourceCodeSequence(tokens[0].line, undefined);
    let allowDef = (minIndent < 0);
    const sequence = new SourceCodeSequence(tokens[0].line, thisIndent);
    while (tokens[0] && (tokens[0].indent >= thisIndent)) {
        if (tokens[0].indent > thisIndent) {
            sequence.content.push(SourceCodeSequence.fromTokens(thisIndent + 1, tokens, defined));
            continue;
        }
        tokens.shift();
        const nextPart = SourceCodePart.fromTokens(sequence.indent, tokens, defined, allowDef);
        sequence.content.push(nextPart);
        if (allowDef && !(nextPart instanceof SourceCodeDefinition) && !(nextPart instanceof SourceCodeError))
            allowDef = false;
        while (tokens[0] && !(tokens[0] instanceof TokenIndent))
            sequence.content.push(new SourceCodeErrorMisplaced(tokens.shift()));
    }
    return sequence;
}

/***[ parseSourceCode/SourceCodeWhile.js ] *************************/
/**
* A while-loop within source code
*/
function SourceCodeWhile(
    condition,  // SourceCodeCommand instance beeing condition
    body,       // SourceCodeSequence beeing looped
    line        // Associated line number
) {
    SourceCodePart.call(this, line)
    this.condition = condition;
    this.body = body;
}
SourceCodeWhile.prototype = Object.create(SourceCodePart.prototype);
SourceCodeWhile.prototype.constructor = SourceCodeWhile;

// SourceCodeCommand instance beeing condition
SourceCodeWhile.prototype.condition = null;

// SourceCodeSequence beeing looped
SourceCodeWhile.prototype.body = null;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodeWhile.prototype.getFirstError = function() {
    if (this.condition instanceof SourceCodeError)
        return this.condition;
    if (this.body instanceof SourceCodeError)
        return this.body;
    return this.body.getFirstError();
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeWhile.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    const indexStart = ce.getNextIndex();
    this.condition._compile(ce, context);
    ce.addCommand('jumpIfFalse');
    const jump2End = ce.addTokenAddress();
    this.body._compile(ce, context);
    ce.addCommand('jump');
    ce.addTokenAddress(indexStart);
    jump2End.value = ce.getNextIndex();
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeWhile.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    this._ensureDefinedCalls('condition', defined);
    this._ensureDefinedCalls('body', defined);
    return this;
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodeWhile.prototype.getLineNumbers = function() {
    return [ this.line ].concat(this.body.getLineNumbers());
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodeWhile.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    if (Number.isFinite(parentLine))
        return (parentLine == this.line) ? this.body.getLineNumbers() : this.body.getNestedLines(parentLine);
    return this.body.getLineNumbers();
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeWhile.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeWhile))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeWhile, type: 'class' };
    const conditionCmp = this.condition.compare(compareWith.condition);
    if (conditionCmp)
        return { index: -1, actualValue: compareWith.condition, plannedValue: this.condition, type: 'condition' };
    const bodyCmp = this.body.compare(compareWith.body);
    if (bodyCmp)
        return { index: bodyCmp.index, actualValue: compareWith.body, plannedValue: this.body, type: 'item' };
    return null;
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeWhile.prototype.clone = function() {
    return new SourceCodeWhile(this.condition.clone(), this.body.clone(), this.line);
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodeWhile.prototype.createCanonical = function() {
    return new SourceCodeWhile(this.condition.createCanonical(), this.body.createCanonical(), this.line);
}

/**
* If possible create a SourceCodeWhile instance using shifting tokens from array tokens
* @return SourceCodeWhile|SourceCodeError|null    Created instance, error message or null
*/
SourceCodeWhile.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined     // Names of already defined functions
) {
    if (!(tokens[0] instanceof TokenWhile))
        return null;
    const line = tokens.shift().line;
    const condition = SourceCodeCommand.fromTokens(indent, tokens, defined);
    const acceptCondition = ((condition instanceof SourceCodeCommand) || (condition instanceof SourceCodeError));
    const body = SourceCodeSequence.fromTokens(indent + 1, tokens, defined);
    return new SourceCodeWhile(
        acceptCondition ? condition : new SourceCodeErrorMissing(TokenCommand, line),
        body,
        line
    );
}

// Add this class to possible source code parts
SourceCodePart.standalones.push(SourceCodeWhile);

/***[ parseSourceCode/SourceCodeDefinition.js ] *************************/
/**
* A definition of an user defined procedure within source code
*/
function SourceCodeDefinition(
    name,       // Name of user defined procedure (or SourceCodeError instance)
    body,       // SourceCodeSequence beeing executed on procedure call
    isFast,     // Execute body all at once?
    line        // Associated line number
) {
    SourceCodePart.call(this, line)
    this.name = name;
    this.body = body;
    if (isFast)
        this.isSlow = false;
}
SourceCodeDefinition.prototype = Object.create(SourceCodePart.prototype);
SourceCodeDefinition.prototype.constructor = SourceCodeDefinition;

// Name of user defined procedure (or SourceCodeError instance)
SourceCodeDefinition.prototype.name = '';

// SourceCodeSequence beeing executed on procedure call
SourceCodeDefinition.prototype.body = null;

// Execute body all at once?
SourceCodeDefinition.prototype.isSlow = true;

/**
* Return first syntax error
* @return SourceCodeError       First syntax error
*/
SourceCodeDefinition.prototype.getFirstError = function() {
    if (this.name instanceof SourceCodeError)
        return this.name;
    if (this.body instanceof SourceCodeError)
        return this.body;
    return this.body.getFirstError();
}

/**
* Add this source code part to compiling environment
* @return undefined
*/
SourceCodeDefinition.prototype._compile = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    ce.registerStartOfFunction(this.name);
    if (context.reportLine) {
        ce.addCommand('reportLine');
        ce.addNumber(this.line);
    }
    const isSlow = this.isSlow && context.runSlow;
    if (isSlow)
        ce.addCommand('reportStepDone');
    this.body._compile(ce, { runSlow: isSlow, iterationLevel: 0, reportLine: context.reportLine });
    ce.addCommand('returnTrue');
}

/**
* Eventually add a reportLine- and reportStepDone to machine code to indicate learning of user defined method
* @return undefined
*/
SourceCodeDefinition.prototype._compileLearning = function(
    ce,         // Used compiling environment
    context     // Object containing values represent present compiling context, e. g. runSlow to enable use of reportStepDone after each relevant step
) {
    if (context.reportLine && context.runSlow) {
        ce.addCommand('reportLine');
        ce.addNumber(this.line);
        ce.addCommand('reportStepDone');
    }
}

/**
* Ensure all included command calls are defined
* @return this|SourceCodeErrorUndefined     Error if this is an undefined command call, otherwise this
*/
SourceCodeDefinition.prototype.ensureDefinedCalls = function(
    defined     // Names of defined functions (includeing build in commands)
) {
    this._ensureDefinedCalls('body', defined);
    return this;
}

/**
* Get number of all lines of source code belonging to this instance
* @return Array     Line numbers
*/
SourceCodeDefinition.prototype.getLineNumbers = function() {
    return [ this.line ].concat(this.body.getLineNumbers());
}

/**
* Get all subordinated lines of this very part if parentLine is undefined or tries to find all subordinated lines of line number parentLine otherwise
* @return Array     Line numbers of subordinated lines
*/
SourceCodeDefinition.prototype.getNestedLines = function(
    parentLine      // Optional: Number of line whoose subordinated lines are requested
) {
    if (Number.isFinite(parentLine))
        return (parentLine == this.line) ? this.body.getLineNumbers() : this.body.getNestedLines(parentLine);
    return this.body.getLineNumbers();
}

/**
* Find first difference between this instance and SourceCodePart instance compareWith
* @return Object    First found difference or null; like { index: "index of first sub part beeing different", actualValue: "value of this beeing different", plannedValue: "corresponding value of compareWith", type: "type of difference, e. g. 'class', 'item', …" }
*/
SourceCodeDefinition.prototype.compare = function(
    compareWith     // SourceCodePart instance representing planed structure
) {
    if (!(compareWith instanceof SourceCodeDefinition))
        return { index: -1, actualValue: (compareWith instanceof Object) ? compareWith.prototype : compareWith, plannedValue: SourceCodeDefinition, type: 'class' };
    const nameCmp = this.name.compare(compareWith.name);
    if (nameCmp)
        return { index: -1, actualValue: compareWith.name, plannedValue: this.name, type: 'name' };
    const bodyCmp = this.body.compare(compareWith.body);
    if (bodyCmp)
        return { index: bodyCmp.index, actualValue: compareWith.body, plannedValue: this.body, type: 'item' };
    if (this.isSlow != compareWith.isSlow)
        return { index: -1, actualValue: compareWith.isSlow, plannedValue: this.isSlow, type: 'isSlow' };
    return null;
}

/**
* Create a absolutely independent copy of this instance containing the very same information
* @return SourceCodePart    Clone
*/
SourceCodeDefinition.prototype.clone = function() {
    return new SourceCodeDefinition(this.name, this.body.clone(), !this.isSlow, this.line);
}

/**
* Create a canonical representation of this instance without iterations and directly nested sequences
* @return SourceCodePart    Canonical version of this instance
*/
SourceCodeDefinition.prototype.createCanonical = function() {
    return new SourceCodeDefinition(this.name, this.body.createCanonical(), !this.isSlow, this.line);
}

/**
* If possible create a SourceCodeDefinition instance using shifting tokens from array tokens
* @return SourceCodeDefinition|SourceCodeError|null    Created instance, error message or null
*/
SourceCodeDefinition.fromTokens = function(
    indent,     // Present indentation level
    tokens,     // Remaining tokens
    defined     // Names of already defined functions
) {
    if (!(tokens[0] instanceof TokenDefinition))
        return null;
    const line = tokens.shift().line;
    const nameToken = SourceCodeCommand.fromTokens(indent, tokens, defined);
    let name;
    let validName = true;
    switch (true) {
        case (nameToken instanceof SourceCodeCommand):
            name = nameToken.fktName;
            validName = (defined.indexOf(name) == -1);
            if (validName)
                defined.push(name);
            break;
        case (nameToken instanceof SourceCodeError):
            name = nameToken;
            break;
        default:
            name = new SourceCodeErrorMissing(TokenCommand, line);
            break;
    }
    const isFast = (tokens[0] instanceof TokenRunFast);
    if (isFast)
        tokens.shift();
    const body = SourceCodeSequence.fromTokens(indent + 1, tokens, defined);
    return validName ? new SourceCodeDefinition(name, body, isFast, line) : new SourceCodeErrorRedefined(name, line);
}

// Add this class to possible source code parts
SourceCodePart.standalones.push(SourceCodeDefinition);

/***[ roboterProcessor/MaschineCodeReader.js ] *************************/
/**
* Read numbers and half bytes from hex string starting with number of extra hex characters per number
*/
function MaschineCodeReader(
    code,       // String containing hex characters
    offset      // Optional: Position of char representing number length
) {
    const realOffset = Number.isFinite(offset) ? offset : 0;
    this.code = code.substring(realOffset + 1);
    const char = code.charAt(realOffset);
    this.numberLength = 1 + (char ? parseInt(char, 16) : 0);
}

/** Maschine code containing hex characters */
MaschineCodeReader.prototype.code = '0';

/** Length of line numbers and code indizes in characters */
MaschineCodeReader.prototype.numberLength = 1;

/** Next character of code to read */
MaschineCodeReader.prototype.index = 0;

/**
* Reads next character of code and steps code position
* @return Number        Read character
*/
MaschineCodeReader.prototype.read = function() {
    const char = this.code.charAt(this.index++);
    const v = char ? parseInt(char, 16) : 0;
    return Number.isFinite(v) ? v : 0;
}
MaschineCodeReader.prototype.readHalfbyte = MaschineCodeReader.prototype.read;

/**
* Read next number of code and steps code position
* @return Number        Read number
*/
MaschineCodeReader.prototype.readNumber = function() {
    const chars = this.code.substring(this.index, this.index + this.numberLength);
    const result = chars ? parseInt(chars, 16) : 0;
    this.index += this.numberLength;
    return Number.isFinite(result) ? result : 0;
}

/**
* Check if next half bytes match halfbytes
* @return Boolean       Is there a match?
*/
MaschineCodeReader.prototype.isNext = function(
    halfbytes       // Hex character string or array of 4bit-numbers to match
) {
    const substr = (halfbytes instanceof Array) ? halfbytes.reduce((s, value) => (s + value.toString(16)), '') : halfbytes;
    return (substr == this.code.substring(this.index, this.index + substr.length));
}

/***[ roboterProcessor/RobotProcessor.js ] *************************/
/**
* Class for executing a robot's programm instruction by instruction
*/
function RobotProcessor(
    memory          // Robot's world
) {
    this.memory = memory;
}

/** Robot's world */
RobotProcessor.prototype.memory = null;

/** Present x coordinate of memory cell */
RobotProcessor.prototype.x = 0;

/** Present y coordinate of memory cell */
RobotProcessor.prototype.y = 0;

/** Work direction (0 meaning increasing y, 1 meaning decreasing x, 2 meaning decreasing y and 3 meaning increasing x) */
RobotProcessor.prototype.direction = 0;

/** MaschineCodeReader instance reading code */
RobotProcessor.prototype._reader = null;

/** Present line number */
RobotProcessor.prototype.line = -1;

/** Stack of iterations */
RobotProcessor.prototype.iterations = [];

/** Stack of function calls consisting of items like ['index of command to execute after function call', 'value of this.iterations before/after function call'] */
RobotProcessor.prototype.calls = [];

/** Function call stack size */
RobotProcessor.prototype.maxCountCalls = 128;

/** Result of last instruction */
RobotProcessor.prototype.lastResult = true;

/**
* Reset code dependent properties to initial status (robot position and direction excluded)
* @return undefined
*/
RobotProcessor.prototype.reset = function() {
    this.iterations = [];
    this.calls = [];
    this.lastResult = true;
    this.line = -1;
    this._reader.index = 0;
}

/**
* Load robot's start position and program
* @return undefined
*/
RobotProcessor.prototype.load = function(
    startx,         // x coordinate of robot to start from
    starty,         // y coordinate of robot to start from
    direction,      // Robot's direction at start
    inputString,    // String containing executeable starting directly after declaration of number length at offset
    offset          // Optional: Position of char representing number length
) {
    this._reader = new MaschineCodeReader(inputString, offset);
    this.reset();
    this.x = startx;
    this.y = starty;
    this.direction = direction;
}

/** Possible results of executing a single instruction */
RobotProcessor.RESULT_OK = 0;         // Instruction successfull, no error or neccessary external action
RobotProcessor.RESULT_DONE = 1;       // Executing finished successfully
RobotProcessor.RESULT_ABORT = 7;      // Executing finished and returning false
RobotProcessor.RESULT_LOG = 2;        // Executed reportStepDone instruction
RobotProcessor.RESULT_MISPLACED = 3;  // Got illegal memory location
RobotProcessor.RESULT_UNDERFLOW = 4;  // Decreased 0
RobotProcessor.RESULT_OVERFLOW = 5;   // Increased 15
RobotProcessor.RESULT_FULLSTACK = 6;  // To much recursion

/**
* Calculate new position after move in set direction
* @return Array     Array consisting of new coordinate x and new coordinate y
*/
RobotProcessor.prototype._move = function() {
    switch (this.direction) {
        case 0:
            return [this.x, this.y + 1];
        case 1:
            return [this.x - 1, this.y];
        case 2:
            return [this.x, this.y - 1];
        case 3:
            return [this.x + 1, this.y];
        default:
            return [this.x, this.y];
    }
}

/**
* Execute next instruction
* @return Number    Result of execution (see RobotProcessor.RESULT_*)
*/
RobotProcessor.prototype.execute = function() {
    const prevResult = this.lastResult;
    if (this._reader.index < 0)
        return prevResult ? RobotProcessor.RESULT_DONE : RobotProcessor.RESULT_ABORT;
    this.lastResult = true;
    const result = [
        () => { [this._reader.index, this.iterations] = this.calls.length ? this.calls.pop() : [-1, []]; },
        () => { [this._reader.index, this.iterations] = this.calls.length ? this.calls.pop() : [-1, []]; this.lastResult = false; },
        () => { this.lastResult = this.memory.touch(...this._move()); },
        () => { this.lastResult = (this.memory.read(this.x, this.y) != this._reader.read()); },
        () => { this.lastResult = (this.direction == 0); },
        () => { if (!(this.iterations.length && this.iterations[this.iterations.length - 1]--)) { this.iterations.pop(); this.lastResult = false; } },
        () => { const moved = this._move(); if (!this.memory.touch(...moved)) return RobotProcessor.RESULT_MISPLACED; [this.x, this.y] = moved; },
        () => { const v = this.memory.read(this.x, this.y); if (v == 15) return RobotProcessor.RESULT_OVERFLOW; this.memory.write(this.x, this.y, v + 1); },
        () => { const v = this.memory.read(this.x, this.y); if (v == 0) return RobotProcessor.RESULT_UNDERFLOW; this.memory.write(this.x, this.y, v - 1); },
        () => { this.direction = (this.direction + 1) % 4; },
        () => { this._reader.index = this._reader.readNumber(); },
        () => { const target = this._reader.readNumber(); if (this.calls.length > this.maxCountCalls) return RobotProcessor.RESULT_FULLSTACK; this.calls.push([this._reader.index, this.iterations]); this.iterations = []; this._reader.index = target; },
        () => { const target = this._reader.readNumber(); if (!prevResult) this._reader.index = target; },
        () => { this.iterations.push(this._reader.readNumber()); },
        () => { this.line = this._reader.readNumber(); this.lastResult = prevResult; },
        () => RobotProcessor.RESULT_LOG
    ][this._reader.read()]();
    return result ? result : RobotProcessor.RESULT_OK;
}

/***[ roboterProcessor/RobotWorldSlot.js ] *************************/
/**
* Abstract class for a single space in the roboter's world where the robot might be positioned or might be tried to be positioned
*/
function RobotWorldSlot() {
}

/**
* Checks if the robot can be positioned on this slot
* @return Boolean       Can robot stand here?
*/
RobotWorldSlot.prototype.isAccesible = function() {
    return true;
}

/**
* Gets the number stored by this slot
* @return Number        Number stored by this slot
*/
RobotWorldSlot.prototype.read = function() {
    return 0;
}

/**
* Sets the number stored by this slot to value
* @return undefined
*/
RobotWorldSlot.prototype.write = function(
    value       // New value
) {
}

/**
* Add information on slot to compiling environment
* @return undefined
*/
RobotWorldSlot.prototype.compile = function(
    ce          // Used CompilingEnvironment instance
) {
    const type = RobotWorldSlot.slotTypes.filter(({ subclass }) => (this instanceof subclass))[0];
    if (!type)
        return;
    type.magicNumbers.forEach(value => ce.addHalfbyte(value));
}

/**
* Reconstruct own status from maschine code
* @return undefined
*/
RobotWorldSlot.prototype._fromMaschineCode = function(
    code        // MaschineCodeReader instance to read from
) {
}

/** Information on available sub classes (objects like { "subclass": "class derived from RobotWorldSlot", "magicNumbers": "array of half byte values describing this type in maschine code" } */
RobotWorldSlot.slotTypes = [];

/**
* Create slot described by maschine code
* @return RobotWorldSlot        Described slot
*/
RobotWorldSlot.fromMaschineCode = function(
    code        // MaschineCodeReader instance to read from
) {
    const type = RobotWorldSlot.slotTypes.filter(({ magicNumbers }) => code.isNext(magicNumbers))[0];
    if (!type) {
        code.readHalfbyte();
        return new RobotWorldNormalSlot();
    }
    type.magicNumbers.forEach(() => code.readHalfbyte());
    const slot = new type.subclass();
    slot._fromMaschineCode(code);
    return slot;
}

/***[ roboterProcessor/RobotWorldNormalSlot.js ] *************************/
/**
 * Class representing a single space in the robot's world that can store a number and is accessible to the robot
 */
function RobotWorldNormalSlot() {
}

// Inherit from RobotWorldSlot
RobotWorldNormalSlot.prototype = Object.create(RobotWorldSlot.prototype);
RobotWorldNormalSlot.prototype.constructor = RobotWorldNormalSlot;

/** Private attribute to store the number on this slot */
RobotWorldNormalSlot.prototype._count = 0;

/**
 * Gets the number stored by this slot
 * @return Number        Number stored by this slot
 */
RobotWorldNormalSlot.prototype.read = function() {
    return this._count;
}

/**
 * Sets the number stored by this slot to value
 * @return undefined
 */
RobotWorldNormalSlot.prototype.write = function(
    value       // New value
) {
    this._count = value;
}

/**
* Reconstruct own status from maschine code
* @return undefined
*/
RobotWorldNormalSlot.prototype._fromMaschineCode = function(
    code        // MaschineCodeReader instance to read from
) {
    this._count = code.readHalfbyte();
}

/**
* Add information on slot to compiling environment
* @return undefined
*/
RobotWorldNormalSlot.prototype.compile = function(
    ce          // Used CompilingEnvironment instance
) {
    RobotWorldSlot.prototype._compile.call(this, ce);
    ce.addHalfbyte(this._count);
}

/** Register this class */
RobotWorldSlot.slotTypes.push({ "subclass": RobotWorldNormalSlot, "magicNumbers": [0] });

/***[ roboterProcessor/RobotWorldUnaccesibleSlot.js ] *************************/
/**
 * Class representing a single space in the robot's world that is not accessible to the robot
 */
function RobotWorldUnaccesibleSlot() {
}

// Inherit from RobotWorldSlot
RobotWorldUnaccesibleSlot.prototype = Object.create(RobotWorldSlot.prototype);
RobotWorldUnaccesibleSlot.prototype.constructor = RobotWorldUnaccesibleSlot;

/**
 * Overrides the isAccessible method of RobotWorldSlot to always return false
 * @return Boolean       Can robot stand here? (always false)
 */
RobotWorldUnaccesibleSlot.prototype.isAccesible = function() {
    return false;
}

/** Register this class */
RobotWorldSlot.slotTypes.push({ "subclass": RobotWorldUnaccesibleSlot, "magicNumbers": [1, 0] });

/***[ roboterProcessor/RobotWorld.js ] *************************/
/**
* A world where the robot can live in consisting of a finite number of slots
*/
function RobotWorld() {
    this.slots = [];
}

/** Count of slots west to east */
RobotWorld.prototype.width = 1;

/** Count of slots south to west */
RobotWorld.prototype.height = 1;

/** All RobotWorldSlot instances arranged as a two demensional array */
RobotWorld.prototype.slots = [];

/**
* Set size of robot's world
* @return undefined
*/
RobotWorld.prototype.setSize = function(
    width,      // Count of slots west to east
    height      // Count of slots south to west
) {
    this.width = Math.max(1, width);
    this.height = Math.max(1, height);
    while (this.slots.length < this.height)
        this.slots.push([]);
}

/**
* Get corresponding RobotWorldSlot instance
* @return RobotWorldSlot      Instance allocated at given coordinates
*/
RobotWorld.prototype.getSlot = function(
    x,      // x coordinate of slot in question
    y       // y coordinate of slot in question
) {
    if ((x < 0) || (x >= this.width) || (y < 0) || (y >= this.height))
        return null;
    if ((this.slots[y] instanceof Array) && (this.slots[y][x] instanceof RobotWorldSlot))
        return this.slots[y][x];
    return null;
}

/**
* Check if slot is accessible to the robot
* @return Boolean       Accessible?
*/
RobotWorld.prototype.touch = function(
    x,      // x coordinate of slot in question
    y       // y coordinate of slot in question
) {
    const slot = this.getSlot(x, y);
    return Boolean(slot && slot.isAccesible());
}

/**
* Get number stored at described slot
* @return Number        Stored number
*/
RobotWorld.prototype.read = function(
    x,      // x coordinate of slot in question
    y       // y coordinate of slot in question
) {
    const slot = this.getSlot(x, y);
    return (slot instanceof RobotWorldSlot) ? slot.read() : 0;
}

/**
* Store value as number of described slot
* @return undefined
*/
RobotWorld.prototype.write = function(
    x,      // x coordinate of slot in question
    y,      // y coordinate of slot in question
    value   // Value to save
) {
    const slot = this.getSlot(x, y);
    if (slot instanceof RobotWorldSlot)
        slot.write(value);
}

/**
* Get one dimensional array (ordered by y ascending primary and x ascending secondary) of numbers stored by slots
* @return Array     Numbers stored by slots
*/
RobotWorld.prototype.getAllNumbers = function() {
    return this.slots.reduce((arr, row) => arr.concat(row), []).map(slot => slot.read());
}

/**
* Load world data from maschine code
* @return undefined
*/
RobotWorld.prototype.load = function(
    reader      // MaschineCodeReader instance to get data from
) {
    let x, y;
    for (y = 0; y < this.height; ++y) {
        this.slots[y] = [];
        for (x = 0; x < this.width; ++x)
            this.slots[y].push(RobotWorldSlot.fromMaschineCode(reader));
    }
}

/***[ roboterProcessor/RobotSimulator.js ] *************************/
/**
* Class for running an acutal simulation of the robot
*/
function RobotSimulator(
    boss        // Function to call whenever something relevant occures
) {
    this.memory = new RobotWorld(this);
    this.processor = new RobotProcessor(this.memory);
    this._boss = boss;
}

/** Is simulating? */
Object.defineProperty(RobotSimulator.prototype, 'running', {
    enumerable: true,
    get() { return (this._timeout !== false); },
    set(v) {
        if (v)
            this.start();
        else
            this.stop();
    }
});

/** Used RobotWorld instance */
RobotSimulator.prototype.memory = null;

/** Responsible RoboterProcessor instance */
RobotSimulator.prototype.processor = null;

/** Delay in milli seconds between the execution of two machine commands (except reportStepDone) */
RobotSimulator.prototype.delayNormal = 0;

/** Delay in milli seconds between the execution of reportStepDone and the next maschine command */
RobotSimulator.prototype.delayStep = 1000;

/** Timeout for next call of _nextCommand */
RobotSimulator.prototype._timeout = false;

/** Function to call whenever something relevant occures */
RobotSimulator.prototype._boss = null;

/**
* Load simulation data to execute
* @return undefined
*/
RobotSimulator.prototype.load = function(
    code        // Maschine code consisting of world data, robot position and program to execute
) {
    const reader = new MaschineCodeReader(code);
    const speed = reader.readNumber() * 100;
    this.processor.delayStep = speed;
    const width = reader.readNumber();
    const height = reader.readNumber();
    this.memory.setSize(width, height);
    this.memory.load(reader);
    const x = reader.readNumber();
    const y = reader.readNumber();
    const direction = reader.readHalfbyte();
    this.processor.load(x, y, direction, reader.code, reader.index);
}

/**
* Start or resume running programm
* @return undefined
*/
RobotSimulator.prototype.start = function() {
    if (!this.running)
        this._nextCommand();
}

/**
* Stop or pause running programm
* @return undefined
*/
RobotSimulator.prototype.stop = function() {
    if (this.running) {
        clearTimeout(this._timeout);
        this._timeout = false;
    }
}

/** Possible reasons for informing boss */
RobotSimulator.RESULT_DONE = 0;         // Executing finished returning true
RobotSimulator.RESULT_ABORT = 7;        // Executing finished returning false
RobotSimulator.RESULT_REQUESTED = 1;    // Boss requested status report
RobotSimulator.RESULT_LOG = 2;          // Executed reportStepDone instruction
RobotSimulator.RESULT_MISPLACED = 3;    // Got illegal memory location
RobotSimulator.RESULT_UNDERFLOW = 4;    // Decreased 0
RobotSimulator.RESULT_OVERFLOW = 5;     // Increased 15
RobotSimulator.RESULT_FULLSTACK = 6;    // To much recursion

/**
* Execute next command, report to boss and/or plan execution of another command
* @return undefined
*/
RobotSimulator.prototype._nextCommand = function() {
    this._timeout = false;
//console.log('Index:', this.processor._reader.index);
    const result = this.processor.execute();
    switch (result) {
        // Instruction successfull, no error or neccessary external action
        case RobotProcessor.RESULT_OK:
            this._timeout = setTimeout(() => this._nextCommand(), this.delayNormal);
            break;
        // Executed reportStepDone instruction
        case RobotProcessor.RESULT_LOG:
            this._timeout = setTimeout(() => this._nextCommand(), this.delayStep);
            // No break on purpose
        // Executing finished
        case RobotProcessor.RESULT_DONE:
        case RobotProcessor.RESULT_ABORT:
        // Got illegal memory location
        case RobotProcessor.RESULT_MISPLACED:
        // Decreased 0
        case RobotProcessor.RESULT_UNDERFLOW:
        // Increased 15
        case RobotProcessor.RESULT_OVERFLOW:
        // To much recursion
        case RobotProcessor.RESULT_FULLSTACK:
            const objTranslation = {};
            objTranslation[RobotProcessor.RESULT_OK] = RobotSimulator.RESULT_OK;
            objTranslation[RobotProcessor.RESULT_DONE] = RobotSimulator.RESULT_DONE;
            objTranslation[RobotProcessor.RESULT_ABORT] = RobotSimulator.RESULT_ABORT;
            objTranslation[RobotProcessor.RESULT_LOG] = RobotSimulator.RESULT_LOG;
            objTranslation[RobotProcessor.RESULT_MISPLACED] = RobotSimulator.RESULT_MISPLACED;
            objTranslation[RobotProcessor.RESULT_UNDERFLOW] = RobotSimulator.RESULT_UNDERFLOW;
            objTranslation[RobotProcessor.RESULT_OVERFLOW] = RobotSimulator.RESULT_OVERFLOW;
            objTranslation[RobotProcessor.RESULT_FULLSTACK] = RobotSimulator.RESULT_FULLSTACK;
            this._informBoss(objTranslation[result]);
            break;
    }
}

/**
* Initiate an extraordinary status report
* @return undefined
*/
RobotSimulator.prototype.tellStatus = function() {
    this._informBoss(RobotSimulator.RESULT_REQUESTED);
}

/**
* Notify master object of status change
* @return undefined
*/
RobotSimulator.prototype._informBoss = function(
    status      // Reason for notification (see RoboterProcessor.RESULT_*)
) {
    if ((typeof this._boss != 'function') && !(this._boss instanceof Function))
        return;
    const ce = new CompilingEnvironment();
    ce.addHalfbyte(0);                  // Message is only status
    ce.addHalfbyte(status);             // Type of status
    ce.addNumber(this.processor.line);  // Active source code line's number
    // Robot's position and direction
    ce.addNumber(this.processor.x);
    ce.addNumber(this.processor.y);
    ce.addHalfbyte(this.processor.direction);
    // Size of world and stored numbers
    ce.addNumber(this.memory.width);
    ce.addNumber(this.memory.height);
    this.memory.getAllNumbers().forEach(value => ce.addHalfbyte(value));
    // Send to boss
    this._boss.call(this, ce.getExecuteable());
}

/***[ roboterProcessor/RobotSynchronizedWorld.js ] *************************/
/**
* Class for robot's worlds' data received
*/
function RobotSynchronizedWorld() {
    this.world = new RobotWorld();
}

/** Used RobotWorld instance */
RobotSynchronizedWorld.prototype.world = null;

/** Position south to north with 0 beeing south */
RobotSynchronizedWorld.prototype.robotX = 0;

/** Position west to east with 0 beeing west */
RobotSynchronizedWorld.prototype.robotY = 0;

/** View direction of robot (0 is north, 1 is west and so on) */
RobotSynchronizedWorld.prototype.robotDirection = 0;

/**
* Load size of world, type of slots and robot start data from maschine code
* @return undefined
*/
RobotSynchronizedWorld.prototype.init = function(
    code        // Maschine code of world data to load
) {
    const reader = new MaschineCodeReader(code);
    reader.readNumber();    // Skip speed
    this.world.setSize(reader.readNumber(), reader.readNumber());
    this.world.load(reader);
    this.robotX = reader.readNumber();
    this.robotY = reader.readNumber();
    this.robotDirection = reader.readNumber();
}

/**
* Load state of world from feedback of RobotSimulator
* @return Object        { line: "Number of present line", status: "Type of feedback" }
*/
RobotSynchronizedWorld.prototype.update = function(
    code        // Maschine code generated by RobotSimulator
) {
    const reader = new MaschineCodeReader(code);
    if (reader.readHalfbyte())
        return;
    const status = reader.readHalfbyte();
    const line = reader.readNumber();
    this.robotX = reader.readNumber();
    this.robotY = reader.readNumber();
    this.robotDirection = reader.readHalfbyte();
    const width = reader.readNumber();
    const height = reader.readNumber();
    if ((width != this.world.width) || (height != this.world.height))
        this.world.setSize(width, height);
    for (y = 0; y < height; ++y) {
        for (x = 0; x < width; ++x) {
            this.world.write(x, y, reader.readHalfbyte());
        }
    }
    return { line, status };
}

/**
* Create a RobotsWorldDrawer instance to draw present state of world
* @return RobotsWorldDrawer     Created instance
*/
RobotSynchronizedWorld.prototype.draw = function() {
    return new RobotsWorldDrawer(this.world, { x: this.robotX, y: this.robotY, direction: this.robotDirection }, RobotsWorldSvgJson);
}

/**
* Check if two RobotSynchronizedWorld instances represent the same world (without checking robot's position and direction)
* @return Boolean       Same worlds?
*/
RobotSynchronizedWorld.prototype.sameWorld = function(
    otherWorld      // RobotSynchronizedWorld instance to compare with this
) {
    if (otherWorld instanceof RobotSynchronizedWorld)
        otherWorld = otherWorld.world;
    if (['width', 'height'].reduce((prev, key) => (prev || (this.world[key] != otherWorld.world[key]), false)))
        return false;
    for (y = 0; y < this.world.height; ++y) {
        for (x = 0; x < this.world.width; ++x) {
            if (this.world.read(x, y) != otherWorld.read(x, y))
                return false;
        }
    }
    return true;
}

/***[ drawWorld/template.js ] *************************/
// Data of SVG template
window.RobotsWorldSvgJson = {
    // Using isometric axonometry (true) or cavalier perspective (false)?
    isometric: true,

    // Offsets of tildes (excluding additional world height) in local units
    offsetTop: 0,
    offsetLeft: 0,
    offsetRight: .055,
    offsetBottom: .1109,

    // Conversation factor from local units to mainUnit
    factor: 4,

    // Unit of width and height of svg element
    mainUnit: 'mm',

    // Height and width of first tilde
    tildeWidth: 12,
    tildeHeight: 6,

    // Additional width and height of each further tilde row
    additionalWidthPerX:  6,
    additionalWidthPerY:  6,
    additionalHeightPerY: 3,
    additionalHeightPerX: 3,

    // Additional height foreach unit of world height
    extraHeight: 12 / 15,

    // Additional height of robot
    robotHeight: 12,

    // Maximum height of value
    maxValueHeight: 12,

    // Position of markers: Top of stack (true) or bottom (false)
    markerTop: true,

    // Id of tilde symbol
    idTilde: 'tilde',

    // Id of symbol for blocked tilde
    idBlocked: false,

    // Id of robot symbol by viewing direction (with background beeing north)
    idRobotFacingNorth: 'robot_right',
    idRobotFacingEast: 'robot_forward',
    idRobotFacingSouth: 'robot_left',
    idRobotFacingWest: 'robot_back',

    // Id of signs indicating directions
    idSignNorth: 'signNorth',
    idSignWest: 'signWest',
    idSignEast: 'signEast',
    idSignSouth: 'signSouth',

    // Ids of value symbols
    idValues: [false, 'value1', 'value2', 'value3', 'value4', 'value5', 'value6', 'value7', 'value8', 'value9', 'valueA', 'valueB', 'valueC', 'valueD', 'valueE', 'valueF'],

    // Ids of markers by value
    idMarkerValues: [false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false],

    // Actual content of SVG image
    content: '<?xml version="1.0" encoding="UTF-8"?>\n<svg width="12.055mm" height="6.1109mm" version="1.1" viewBox="0 0 12.055 6.' +
'1109" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><defs><linearGradient id="c" x1=' +
'"437.69" x2="444.67" y1="642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -168.94)" gradientUn' +
'its="userSpaceOnUse" xlink:href="#a"/><linearGradient id="a"><stop style="stop-color:#888a85" offset="0"/><stop styl' +
'e="stop-color:#555753" offset="1"/></linearGradient><linearGradient id="ag" x1="447.14" x2="453.36" y1="638.63" y2="' +
'638.63" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -166.54)" gradientUnits="userSpaceOnUse" xlink:href="#a"/' +
'><linearGradient id="af" x1="437.69" x2="444.67" y1="642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .26458' +
' -109.5 -167.74)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><linearGradient id="d" x1="437.69" x2="444.67" y1=' +
'"642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -167.74)" gradientUnits="userSpaceOnUse" xli' +
'nk:href="#a"/><linearGradient id="ai" x1="447.14" x2="453.36" y1="638.63" y2="638.63" gradientTransform="matrix(.264' +
'58 0 0 .26458 -111.9 -167.74)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><linearGradient id="ah" x1="437.69" x' +
'2="444.67" y1="642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .26458 -109.5 -168.94)" gradientUnits="userS' +
'paceOnUse" xlink:href="#a"/><linearGradient id="b" x1="437.69" x2="444.67" y1="642.87" y2="642.87" gradientTransform' +
'="matrix(.26458 0 0 .26458 -111.9 -167.68)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><linearGradient id="ae" ' +
'x1="447.14" x2="453.36" y1="638.63" y2="638.63" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -167.68)" gradien' +
'tUnits="userSpaceOnUse" xlink:href="#a"/><linearGradient id="ad" x1="437.69" x2="444.67" y1="642.87" y2="642.87" gra' +
'dientTransform="matrix(.26458 0 0 .26458 -109.5 -168.88)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><linearGra' +
'dient id="e" x1="437.69" x2="444.67" y1="642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -168' +
'.94)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><linearGradient id="ak" x1="447.14" x2="453.36" y1="638.63" y2' +
'="638.63" gradientTransform="matrix(.26458 0 0 .26458 -111.9 -166.54)" gradientUnits="userSpaceOnUse" xlink:href="#a' +
'"/><linearGradient id="aj" x1="437.69" x2="444.67" y1="642.87" y2="642.87" gradientTransform="matrix(.26458 0 0 .264' +
'58 -109.5 -167.74)" gradientUnits="userSpaceOnUse" xlink:href="#a"/><g id="robot_back"><path d="m4.2291 2.9553c0.300' +
'01 0.15 0.90001 0.15 1.2-3e-5s0.3-0.44997 0-0.6-0.9-0.15003-1.2 3e-5 -0.30001 0.45 0 0.6z" style="fill:none;stroke-l' +
'inecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m1.8291-5.5862 3.6 1.7999 4.8-2.4-3.6-1.8z' +
'" style="fill:none;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m4.2291 4.0137c' +
'0.30001 0.15 0.90001 0.15 1.2-3e-5s0.3-0.44997 0-0.6-0.9-0.15003-1.2 3e-5 -0.30001 0.45 0 0.6z" style="fill:none;str' +
'oke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.6291 2.8136c0.30001 0.15 0.90001 0' +
'.15 1.2-3e-5s0.3-0.44997 0-0.6-0.9-0.15003-1.2 3e-5 -0.30001 0.45 0 0.6z" style="fill:none;stroke-linecap:round;stro' +
'ke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m4.2291 1.0137c0.30001 0.15 0.90001 0.15 1.2-3e-5s0.3-0.449' +
'97 0-0.6-0.9-0.15003-1.2 3e-5 -0.30001 0.45 0 0.6z" style="fill:none;stroke-linecap:round;stroke-linejoin:round;stro' +
'ke-width:.2;stroke:#000"/><path d="m6.6291-0.18631c0.30001 0.15 0.90001 0.15 1.2-3e-5s0.3-0.44997 0-0.6-0.9-0.15003-' +
'1.2 3e-5 -0.30001 0.45 0 0.6z" style="fill:none;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#0' +
'00"/><path d="m3.9291 4.1636c0.45002 0.225 1.35 0.225 1.8-4e-5 0.44999-0.22505 0.45-0.67496 0-0.9-0.45-0.22505-1.35-' +
'0.22505-1.8 4e-5s-0.45001 0.675 0 0.9z" style="fill:#fce94f;stroke-linecap:round;stroke-linejoin:round;stroke-width:' +
'.2;stroke:#000"/><path d="m4.0051 0.68985v0.0176c1.4e-4 -6e-3 3e-3 -0.0117 3e-3 -0.0176zm1.6449 0c2.6e-4 3e-3 1e-3 5' +
'e-3 2e-3 8e-3v-8e-3zm-1.6449 0.03v2.9874c-1e-4 4e-3 -1e-4 8e-3 0 0.0124v0.0289h5e-3c0.014 0.1004 0.0847 0.19792 0.21' +
'911 0.2651 0.30001 0.15 0.90045 0.15003 1.2004 0 0.13427-0.0671 0.20503-0.16477 0.21911-0.2651h3e-3v-0.0196c6.4e-4 -' +
'0.0104 6.4e-4 -0.0206 0-0.031v-2.9688c-7e-3 0.10721-0.0792 0.21319-0.22221 0.28474-0.29999 0.15003-0.90043 0.15-1.20' +
'04 0-0.14723-0.0736-0.22104-0.18365-0.22376-0.29404z" style="fill:url(#b);font-variation-settings:normal;stop-color:' +
'#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.4054-0.51007v0.015c1.7e' +
'-4 -5e-3 2e-3 -0.01 3e-3 -0.015zm1.6443 0c3.6e-4 4e-3 2e-3 8e-3 2e-3 0.0114v-0.0114zm-1.6443 0.0326v2.9822c-2.1e-4 6' +
'e-3 -2.1e-4 0.0117 0 0.0176v0.0264h5e-3c0.0141 0.1003 0.0849 0.19797 0.21911 0.2651 0.30001 0.15 0.89993 0.15002 1.1' +
'999 0 0.13427-0.0671 0.20503-0.16478 0.21911-0.2651h4e-3v-0.0227c4.1e-4 -8e-3 4.1e-4 -0.0165 0-0.0248v-2.975c-5e-3 0' +
'.10828-0.0783 0.21509-0.22273 0.28732-0.29999 0.15003-0.89991 0.15052-1.1999 5.2e-4 -0.14607-0.073-0.21993-0.18194-0' +
'.22376-0.29146z" style="fill:url(#ae);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-' +
'linejoin:round;stroke-width:.2;stroke:#000"/><path d="m8.0463 1.407c0.0246 0.12774-0.0455 0.26252-0.21704 0.3483-0.2' +
'9999 0.15003-0.90043 0.15-1.2004 0-0.16667-0.0833-0.23711-0.21297-0.21859-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.010' +
'8-0.01 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.1993 0.0517 0.4142 0.32091 0.5488 0.45003 0.225 1.3499 0.22504' +
' 1.7999 0 0.30774-0.15391 0.39997-0.41213 0.2868-0.63252h6e-3l-0.37207-0.92397z" style="fill:url(#ad);font-variation' +
'-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d' +
'="m3.0291-8.5861 2.4 1.1999 3.6-1.8-2.4-1.2z" style="fill:#babdb6;stroke-linecap:round;stroke-linejoin:round;stroke-' +
'width:.2;stroke:#000"/><path d="m1.8291-5.5862 3.6 1.7999 4.8-2.4-3.6-1.8z" style="fill:#babdb6;stroke-linecap:round' +
';stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.4291-7.3861-1e-5 2.9999 3.6-1.8v-2.9999z" style="fi' +
'll:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.4291-3.7862-2e-5 5.9' +
'998 4.8-2.4v-5.9999z" style="fill:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><' +
'path d="m3.0291-8.5861 2.4 1.1999v2.9999l-2.4-1.1999z" style="fill:#eeeeec;stroke-linecap:round;stroke-linejoin:roun' +
'd;stroke-width:.2;stroke:#000"/><path d="m1.8291-5.5862 3.6 1.7999v5.9999l-3.6-1.7999z" style="fill:#eeeeec;stroke-l' +
'inecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><circle cx="4.8154" cy="-9.7749" r=".19701" style="' +
'fill:#555753;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#000"/' +
'><circle cx="7.2154" cy="-10.975" r=".19701" style="fill:#555753;font-variation-settings:normal;stop-color:#000000;s' +
'troke-linejoin:bevel;stroke-width:.124;stroke:#000"/><path d="m5.4291-8.5862-0.61361-1.1887" style="fill:none;stroke' +
'-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.6291-9.1862 0.58639-1.7887" style="fi' +
'll:none;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.6463 2.607c0.0246 0.127' +
'74-0.0455 0.26253-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16668-0.0833-0.23711-0.21297-0.2186-0.33745' +
'h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.0101 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.1993 0.0517 0.4142 0.3209' +
'1 0.5488 0.45002 0.225 1.3499 0.22504 1.7999 0 0.30774-0.1539 0.39998-0.41212 0.28681-0.63252h6e-3l-0.37207-0.92397z' +
'" style="fill:url(#b);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;s' +
'troke-width:.2;stroke:#000"/></g><g id="robot_left"><path d="m4.0037-0.56838v0.0176c1.4e-4 -6e-3 3e-3 -0.0117 3e-3 -' +
'0.0176zm1.6449 0c2.6e-4 3e-3 1e-3 5e-3 2e-3 8e-3v-8e-3zm-1.6449 0.03v2.9874c-1e-4 4e-3 -1e-4 8e-3 0 0.0124v0.0289h5e' +
'-3c0.014 0.1004 0.0847 0.19792 0.21911 0.2651 0.30001 0.15 0.90045 0.15003 1.2004 0 0.13427-0.0671 0.20503-0.16477 0' +
'.21911-0.2651h3e-3v-0.0196c6.4e-4 -0.0104 6.4e-4 -0.0206 0-0.031v-2.9688c-7e-3 0.10721-0.0792 0.21319-0.22221 0.2847' +
'4-0.29999 0.15003-0.90043 0.15-1.2004 0-0.14723-0.0736-0.22104-0.18365-0.22376-0.29404z" style="fill:url(#e);font-va' +
'riation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/>' +
'<path d="m5.6449 1.3488c0.0246 0.12774-0.0455 0.26253-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16668-0' +
'.0833-0.23711-0.21297-0.2186-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.0101 0.0211-0.0134 0.032l-6e-3 0.016h4e-3' +
'c-0.0529 0.1993 0.0517 0.4142 0.32091 0.5488 0.45002 0.225 1.3499 0.22504 1.7999 0 0.30774-0.1539 0.39998-0.41212 0.' +
'28681-0.63252h6e-3l-0.37207-0.92397z" style="fill:url(#e);font-variation-settings:normal;stop-color:#000000;stroke-l' +
'inecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-9.2443 3.6 1.7999 2.4-1.2-3.6-1.8z' +
'" style="fill:#babdb6;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m1.8277-6.24' +
'44 4.8 2.3999 3.6-1.8-4.8-2.4z" style="fill:#babdb6;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;strok' +
'e:#000"/><path d="m6.6277-7.4443-1e-5 2.9999 2.4-1.2v-2.9999z" style="fill:#d3d7cf;stroke-linecap:round;stroke-linej' +
'oin:round;stroke-width:.2;stroke:#000"/><path d="m6.404 0.63172v0.015c1.7e-4 -5e-3 2e-3 -0.01 3e-3 -0.015zm1.6443 0c' +
'3.6e-4 4e-3 2e-3 8e-3 2e-3 0.0114v-0.0114zm-1.6443 0.0326v2.9822c-2.1e-4 6e-3 -2.1e-4 0.0117 0 0.0176v0.0264h5e-3c0.' +
'0141 0.1003 0.0849 0.19797 0.21911 0.2651 0.30001 0.15 0.89993 0.15002 1.1999 0 0.13427-0.0671 0.20503-0.16478 0.219' +
'11-0.2651h4e-3v-0.0227c4.1e-4 -8e-3 4.1e-4 -0.0165 0-0.0248v-2.975c-5e-3 0.10828-0.0783 0.21509-0.22273 0.28732-0.29' +
'999 0.15003-0.89991 0.15052-1.1999 5.2e-4 -0.14607-0.073-0.21993-0.18194-0.22376-0.29146z" style="fill:url(#ak);font' +
'-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000' +
'"/><path d="m8.0449 2.5488c0.0246 0.12774-0.0455 0.26252-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.1666' +
'7-0.0833-0.23711-0.21297-0.21859-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.01 0.0211-0.0134 0.032l-6e-3 0.016h4e' +
'-3c-0.0529 0.1993 0.0517 0.4142 0.32091 0.5488 0.45003 0.225 1.3499 0.22504 1.7999 0 0.30774-0.15391 0.39997-0.41213' +
' 0.2868-0.63252h6e-3l-0.37207-0.92397z" style="fill:url(#aj);font-variation-settings:normal;stop-color:#000000;strok' +
'e-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.6277-3.8444-2e-5 5.9998 3.6-1.8v-5.9' +
'999z" style="fill:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-' +
'9.2443 3.6 1.7999v2.9999l-3.6-1.7999z" style="fill:#eeeeec;stroke-linecap:round;stroke-linejoin:round;stroke-width:.' +
'2;stroke:#000"/><path d="m1.8277-6.2444 4.8 2.3999v5.9999l-4.8-2.3999z" style="fill:#eeeeec;stroke-linecap:round;str' +
'oke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.8623-5.9506v0.77247l-2.0692-1.0346v-0.77248z" style="fi' +
'll:#fff;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.1597-7.8096 0.88023 0.4' +
'6514v1.1629l-0.88023-0.46515z" style="fill:#fff;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#0' +
'00"/><path d="m3.6154-8.6068 0.88023 0.46514v1.1629l-0.88023-0.46515z" style="fill:#fff;stroke-linecap:round;stroke-' +
'linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.1597-7.0406 0.29812 0.15753v0.39386l-0.29812-0.15754z" styl' +
'e="stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m4.1975-7.5301 0.29812 0.15753v' +
'0.39386l-0.29812-0.15754z" style="stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><circle c' +
'x="4.8277" cy="-11.044" r=".19701" style="fill:#555753;font-variation-settings:normal;stop-color:#000000;stroke-line' +
'join:bevel;stroke-width:.124;stroke:#000"/><circle cx="7.2277" cy="-9.8444" r=".19701" style="fill:#555753;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#000"/><path d="m5.4277-9.24' +
'44-0.59999-1.7999" style="fill:none;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d' +
'="m6.6277-8.6444 0.6-1.2" style="fill:none;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/>' +
'</g><g id="robot_forward"><path d="m1.8277-5.6445 3.6 1.7999 4.8-2.4-3.6-1.8z" style="fill:none;stroke-linecap:round' +
';stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.9277 4.1053c0.45002 0.225 1.35 0.225 1.8-4e-5 0.449' +
'99-0.22505 0.45-0.67496 0-0.9-0.45-0.22505-1.35-0.22505-1.8 4e-5s-0.45001 0.675 0 0.9z" style="fill:#fce94f;stroke-l' +
'inecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m4.0037 0.63153v0.0176c1.4e-4 -6e-3 3e-3 -' +
'0.0117 3e-3 -0.0176zm1.6449 0c2.6e-4 3e-3 1e-3 5e-3 2e-3 8e-3v-8e-3zm-1.6449 0.03v2.9874c-1e-4 4e-3 -1e-4 8e-3 0 0.0' +
'124v0.0289h5e-3c0.014 0.1004 0.0847 0.19792 0.21911 0.2651 0.30001 0.15 0.90045 0.15003 1.2004 0 0.13427-0.0671 0.20' +
'503-0.16477 0.21911-0.2651h3e-3v-0.0196c6.4e-4 -0.0104 6.4e-4 -0.0206 0-0.031v-2.9688c-7e-3 0.10721-0.0792 0.21319-0' +
'.22221 0.28474-0.29999 0.15003-0.90043 0.15-1.2004 0-0.14723-0.0736-0.22104-0.18365-0.22376-0.29404z" style="fill:ur' +
'l(#d);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;s' +
'troke:#000"/><path d="m6.4041-0.56839v0.015c1.7e-4 -5e-3 2e-3 -0.01 3e-3 -0.015zm1.6443 0c3.6e-4 4e-3 2e-3 8e-3 2e-3' +
' 0.0114v-0.0114zm-1.6443 0.0326v2.9822c-2.1e-4 6e-3 -2.1e-4 0.0117 0 0.0176v0.0264h5e-3c0.0141 0.1003 0.0849 0.19797' +
' 0.21911 0.2651 0.30001 0.15 0.89993 0.15002 1.1999 0 0.13427-0.0671 0.20503-0.16478 0.21911-0.2651h4e-3v-0.0227c4.1' +
'e-4 -8e-3 4.1e-4 -0.0165 0-0.0248v-2.975c-5e-3 0.10828-0.0783 0.21509-0.22273 0.28732-0.29999 0.15003-0.89991 0.1505' +
'2-1.1999 5.2e-4 -0.14607-0.073-0.21993-0.18194-0.22376-0.29146z" style="fill:url(#ai);font-variation-settings:normal' +
';stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m8.0449 1.3487' +
'c0.0246 0.12774-0.0455 0.26252-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16667-0.0833-0.23711-0.21297-0' +
'.21859-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.01 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.1993 0.0517 0' +
'.4142 0.32091 0.5488 0.45003 0.225 1.3499 0.22504 1.7999 0 0.30774-0.15391 0.39997-0.41213 0.2868-0.63252h6e-3l-0.37' +
'207-0.92397z" style="fill:url(#ah);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-lin' +
'ejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-8.6444 2.4 1.1999 3.6-1.8-2.4-1.2z" style="fill:#babdb6;s' +
'troke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m1.8277-5.6445 3.6 1.7999 4.8-2.4-3' +
'.6-1.8z" style="fill:#babdb6;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.42' +
'77-7.4444-1e-5 2.9999 3.6-1.8v-2.9999z" style="fill:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke-width:' +
'.2;stroke:#000"/><path d="m5.4277-3.8445-2e-5 5.9998 4.8-2.4v-5.9999z" style="fill:#d3d7cf;stroke-linecap:round;stro' +
'ke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-8.6444 2.4 1.1999v2.9999l-2.4-1.1999z" style="fill:' +
'#eeeeec;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m1.8277-5.6445 3.6 1.7999v' +
'5.9999l-3.6-1.7999z" style="fill:#eeeeec;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><p' +
'ath d="m6.295-5.9796v0.77247l2.0692-1.0346v-0.77248z" style="fill:#fff;stroke-linecap:round;stroke-linejoin:round;st' +
'roke-width:.2;stroke:#000"/><path d="m6.8698-7.836-0.88023 0.46514v1.1629l0.88023-0.46515z" style="fill:#fff;stroke-' +
'linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m8.4141-8.6332-0.88023 0.46514v1.1629l0.8' +
'8023-0.46515z" style="fill:#fff;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6' +
'.8698-7.0669-0.29812 0.15753v0.39386l0.29812-0.15754z" style="stroke-linecap:round;stroke-linejoin:round;stroke-widt' +
'h:.2;stroke:#000"/><path d="m7.832-7.5565-0.29812 0.15753v0.39386l0.29812-0.15754z" style="stroke-linecap:round;stro' +
'ke-linejoin:round;stroke-width:.2;stroke:#000"/><circle cx="4.8141" cy="-9.8332" r=".19701" style="fill:#555753;font' +
'-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#000"/><circle cx="7.21' +
'41" cy="-11.033" r=".19701" style="fill:#555753;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:be' +
'vel;stroke-width:.124;stroke:#000"/><path d="m5.4277-8.6445-0.61361-1.1887" style="fill:none;stroke-linecap:round;st' +
'roke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.6277-9.2445 0.58639-1.7887" style="fill:none;stroke-li' +
'necap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.6449 2.5487c0.0246 0.12774-0.0455 0.2625' +
'3-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16668-0.0833-0.23711-0.21297-0.2186-0.33745h-4e-3l-0.38241 ' +
'0.94878c-4e-3 0.0108-0.0101 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.1993 0.0517 0.4142 0.32091 0.5488 0.45002' +
' 0.225 1.3499 0.22504 1.7999 0 0.30774-0.1539 0.39998-0.41212 0.28681-0.63252h6e-3l-0.37207-0.92397z" style="fill:ur' +
'l(#d);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;s' +
'troke:#000"/></g><g id="robot_right"><path d="m4.0037-0.56849v0.0176c1.4e-4 -6e-3 3e-3 -0.0117 3e-3 -0.0176zm1.6449 ' +
'0c2.6e-4 3e-3 1e-3 5e-3 2e-3 8e-3v-8e-3zm-1.6449 0.03v2.9874c-1e-4 4e-3 -1e-4 8e-3 0 0.0124v0.0289h5e-3c0.014 0.1004' +
' 0.0847 0.19792 0.21911 0.2651 0.30001 0.15 0.90045 0.15003 1.2004 0 0.13427-0.0671 0.20503-0.16477 0.21911-0.2651h3' +
'e-3v-0.0196c6.4e-4 -0.0104 6.4e-4 -0.0206 0-0.031v-2.9688c-7e-3 0.10721-0.0792 0.21319-0.22221 0.28474-0.29999 0.150' +
'03-0.90043 0.15-1.2004 0-0.14723-0.0736-0.22104-0.18365-0.22376-0.29404z" style="fill:url(#c);font-variation-setting' +
's:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m5.645' +
' 1.3487c0.0246 0.12774-0.0455 0.26253-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16668-0.0833-0.23711-0.' +
'21297-0.2186-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.0101 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.1993 ' +
'0.0517 0.4142 0.32091 0.5488 0.45002 0.225 1.3499 0.22504 1.7999 0 0.30774-0.1539 0.39998-0.41212 0.28681-0.63252h6e' +
'-3l-0.37207-0.92397z" style="fill:url(#c);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;str' +
'oke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-9.2444 3.6 1.7999 2.4-1.2-3.6-1.8z" style="fill:#b' +
'abdb6;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m1.8278-6.2445 4.8 2.3999 3.' +
'6-1.8-4.8-2.4z" style="fill:#babdb6;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d' +
'="m6.6277-7.4444-1e-5 2.9999 2.4-1.2v-2.9999z" style="fill:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke' +
'-width:.2;stroke:#000"/><path d="m6.4041 0.63161v0.015c1.7e-4 -5e-3 2e-3 -0.01 3e-3 -0.015zm1.6443 0c3.6e-4 4e-3 2e-' +
'3 8e-3 2e-3 0.0114v-0.0114zm-1.6443 0.0326v2.9822c-2.1e-4 6e-3 -2.1e-4 0.0117 0 0.0176v0.0264h5e-3c0.0141 0.1003 0.0' +
'849 0.19797 0.21911 0.2651 0.30001 0.15 0.89993 0.15002 1.1999 0 0.13427-0.0671 0.20503-0.16478 0.21911-0.2651h4e-3v' +
'-0.0227c4.1e-4 -8e-3 4.1e-4 -0.0165 0-0.0248v-2.975c-5e-3 0.10828-0.0783 0.21509-0.22273 0.28732-0.29999 0.15003-0.8' +
'9991 0.15052-1.1999 5.2e-4 -0.14607-0.073-0.21993-0.18194-0.22376-0.29146z" style="fill:url(#ag);font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m8.' +
'045 2.5487c0.0246 0.12774-0.0455 0.26252-0.21704 0.3483-0.29999 0.15003-0.90043 0.15-1.2004 0-0.16667-0.0833-0.23711' +
'-0.21297-0.21859-0.33745h-4e-3l-0.38241 0.94878c-4e-3 0.0108-0.01 0.0211-0.0134 0.032l-6e-3 0.016h4e-3c-0.0529 0.199' +
'3 0.0517 0.4142 0.32091 0.5488 0.45003 0.225 1.3499 0.22504 1.7999 0 0.30774-0.15391 0.39997-0.41213 0.2868-0.63252h' +
'6e-3l-0.37207-0.92397z" style="fill:url(#af);font-variation-settings:normal;stop-color:#000000;stroke-linecap:round;' +
'stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m6.6278-3.8445-2e-5 5.9998 3.6-1.8v-5.9999z" style="fil' +
'l:#d3d7cf;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/><path d="m3.0277-9.2444 3.6 1.799' +
'9v2.9999l-3.6-1.7999z" style="fill:#eeeeec;stroke-linecap:round;stroke-linejoin:round;stroke-width:.2;stroke:#000"/>' +
'<path d="m1.8278-6.2445 4.8 2.3999v5.9999l-4.8-2.3999z" style="fill:#eeeeec;stroke-linecap:round;stroke-linejoin:rou' +
'nd;stroke-width:.2;stroke:#000"/><circle cx="4.8277" cy="-11.044" r=".19701" style="fill:#555753;font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#000"/><circle cx="7.2277" cy="-9.8445' +
'" r=".19701" style="fill:#555753;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#000"/><path d="m5.4277-9.2445-0.59999-1.7999" style="fill:none;stroke-linecap:round;stroke-linejoin:r' +
'ound;stroke-width:.2;stroke:#000"/><path d="m6.6277-8.6445 0.6-1.2" style="fill:none;stroke-linecap:round;stroke-lin' +
'ejoin:round;stroke-width:.2;stroke:#000"/></g><path id="tilde" d="m6.0277 0.055454 6 3-6 3-6-3z" style="fill:#eeeeec' +
';stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#888a85"/><text id="signSouth" transform="matrix(' +
'-1 -.5 1 -.5 0 0)" x="-6.069376" y="-3.6416562" style="font-family:sans-serif;font-size:1.7639px;line-height:1.25;st' +
'roke-width:.26458;text-align:center;text-anchor:middle" xml:space="preserve"><tspan x="-6.069376" y="-3.6416562" sty' +
'le="font-family:sans-serif;font-size:1.7639px;font-weight:bold;stroke-width:.26458;text-align:center;text-anchor:mid' +
'dle">S</tspan></text><text id="signEast" transform="matrix(-1 .5 -1 -.5 0 0)" x="0.041604515" y="-9.6693048" style="' +
'font-family:sans-serif;font-size:1.7639px;line-height:1.25;stroke-width:.26458;text-align:center;text-anchor:middle"' +
' xml:space="preserve"><tspan x="0.041604515" y="-9.6693048" style="font-family:sans-serif;font-size:1.7639px;font-we' +
'ight:bold;stroke-width:.26458;text-align:center;text-anchor:middle">O</tspan></text><text id="signNorth" transform="' +
'matrix(1 .5 -1 .5 0 0)" x="6.069376" y="-3.5583482" style="fill:#a40000;font-family:sans-serif;font-size:1.7639px;li' +
'ne-height:1.25;stroke-width:.26458;text-align:center;text-anchor:middle" xml:space="preserve"><tspan x="6.069376" y=' +
'"-3.5583482" style="fill:#a40000;font-family:sans-serif;font-size:1.7639px;font-weight:bold;stroke-width:.26458;text' +
'-align:center;text-anchor:middle">N</tspan></text><text id="signWest" transform="matrix(1 -.5 1 .5 0 0)" x="-0.04159' +
'6886" y="2.4693089" style="font-family:sans-serif;font-size:1.7639px;line-height:1.25;stroke-width:.26458;text-align' +
':center;text-anchor:middle" xml:space="preserve"><tspan x="-0.041596886" y="2.4693089" style="font-family:sans-serif' +
';font-size:1.7639px;font-weight:bold;stroke-width:.26458;text-align:center;text-anchor:middle">W</tspan></text><g id' +
'="valueF"><g transform="translate(-235.42 -161.59)"><path d="m235.44 163.85v0.79998l6 3v-0.79998z" style="-inkscape-' +
'stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;s' +
'troke:#a40000"/><path d="m235.44 163.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variat' +
'ion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 162' +
'.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#0000' +
'00;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 161.45v0.79998l6 3v-0.79998z" style="-i' +
'nkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#a40000"/><path d="m235.44 160.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;fon' +
't-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m23' +
'5.44 159.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-col' +
'or:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v-0.79998z" s' +
'tyle="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;str' +
'oke-width:.124;stroke:#a40000"/><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef' +
'6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><pat' +
'h d="m235.44 157.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;' +
'stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.7' +
'9998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:b' +
'evel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;' +
'fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a4000' +
'0"/><path d="m235.44 155.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings' +
':normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l' +
'6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-li' +
'nejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stro' +
'ke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;strok' +
'e:#a40000"/><path d="m235.44 152.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-' +
'settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="transl' +
'ate(-235.78 -158.08)"><path d="m247.81 160.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-varia' +
'tion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 15' +
'9.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;s' +
'troke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 158.74-6 3v0.79998l6-3z" style="-inkscape-s' +
'troke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;st' +
'roke:#a40000"/><path d="m247.81 157.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-se' +
'ttings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 157.14-6 ' +
'3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-l' +
'inejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 156.34-6 3v0.79998l6-3z" style="-inkscape-stroke:n' +
'one;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a' +
'40000"/><path d="m247.81 155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:' +
'normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 154.74-6 3v0.799' +
'98l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin' +
':bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fil' +
'l:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/' +
'><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;' +
'stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z' +
'" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;' +
'stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef29' +
'29;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path ' +
'd="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-co' +
'lor:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style' +
'="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-' +
'width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font' +
'-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transf' +
'orm="matrix(1 -.57735 0 1 -177.01 -181.38)" style="stroke-width:.124;stroke:#a40000" aria-label="15"><path d="m185.2' +
'2 289.11h-0.74577v-2.0398q0-0.12842 5e-3 -0.33585 0.01-0.20743 0.0148-0.36548-0.0247 0.0296-0.10866 0.10866-0.079 0.' +
'0741-0.14817 0.13335l-0.40499 0.32597-0.36053-0.44944 1.1359-0.90382h0.61242z" style="fill:#fff"/><path d="m187.49 2' +
'86.86q0.32102 0 0.57291 0.12348 0.25188 0.12347 0.39511 0.36053 0.14816 0.23707 0.14816 0.58773 0 0.57291-0.3556 0.9' +
'0382-0.3556 0.32597-1.052 0.32597-0.27658 0-0.52352-0.0494-0.24201-0.0494-0.42475-0.14323v-0.64206q0.18274 0.0938 0.' +
'43957 0.16299 0.25682 0.0642 0.48401 0.0642 0.3309 0 0.50376-0.13335 0.1778-0.13829 0.1778-0.42474 0-0.5334-0.70626-' +
'0.5334-0.13829 0-0.28645 0.0296-0.14817 0.0247-0.24695 0.0494l-0.29633-0.15804 0.13335-1.7978h1.9114v0.63218h-1.2594' +
'l-0.0642 0.69144q0.084-0.0148 0.1778-0.0296 0.0988-0.0198 0.27164-0.0198z" style="fill:#fff"/></g><g transform="matr' +
'ix(1 .57735 0 1 -177.01 -181.38)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="' +
'15"><path d="m179.22 77.724h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.0247 0.02963-0.10' +
'865 0.10866-0.079 0.07408-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="stop-color' +
':#000000"/><path d="m181.49 75.472q0.32103 0 0.57291 0.12347 0.25189 0.12347 0.39511 0.36054 0.14817 0.23707 0.14817' +
' 0.58773 0 0.57291-0.3556 0.90382-0.3556 0.32597-1.052 0.32597-0.27658 0-0.52353-0.04939-0.242-0.04939-0.42474-0.143' +
'23v-0.64206q0.18274 0.09384 0.43956 0.16298 0.25682 0.06421 0.48401 0.06421 0.33091 0 0.50377-0.13335 0.1778-0.13829' +
' 0.1778-0.42474 0-0.5334-0.70626-0.5334-0.13829 0-0.28646 0.02963-0.14817 0.02469-0.24694 0.04939l-0.29634-0.15804 0' +
'.13335-1.7978h1.9114v0.63218h-1.2594l-0.0642 0.69144q0.084-0.01482 0.1778-0.02963 0.0988-0.01976 0.27164-0.01976z" s' +
'tyle="stop-color:#000000"/></g><path d="m6.0277-11.945 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-line' +
'join:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="valueE"><g transform="translate(-235.42 -160.79)"><path d="' +
'm235.44 163.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-' +
'color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 162.25v0.79998l6 3v-0.79998z' +
'" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;' +
'stroke-width:.124;stroke:#a40000"/><path d="m235.44 161.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:' +
'#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><' +
'path d="m235.44 160.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:norm' +
'al;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.85v0.79998l6 3v-' +
'0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoi' +
'n:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:no' +
'ne;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a4' +
'0000"/><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-setti' +
'ngs:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.799' +
'98l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke' +
'-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-s' +
'troke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;st' +
'roke:#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variati' +
'on-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.' +
'05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#00000' +
'0;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-in' +
'kscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width' +
':.124;stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font' +
'-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235' +
'.44 152.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-colo' +
'r:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -157.28)"><pa' +
'th d="m247.81 159.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop' +
'-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 158.74-6 3v0.79998l6-3z" st' +
'yle="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stro' +
'ke-width:.124;stroke:#a40000"/><path d="m247.81 157.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;f' +
'ont-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m' +
'247.81 157.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:' +
'#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 156.34-6 3v0.79998l6-3z" style="-i' +
'nkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#a40000"/><path d="m247.81 155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-var' +
'iation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 ' +
'154.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000' +
';stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape' +
'-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-' +
'settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-' +
'6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke' +
'-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke' +
':none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:' +
'#a40000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-setting' +
's:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.7' +
'9998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejo' +
'in:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;f' +
'ill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000' +
'"/></g><g transform="matrix(1 -.57735 0 1 -177.01 -180.58)" style="stroke-width:.124;stroke:#a40000" aria-label="14"' +
'><path d="m185.22 288.72h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20744 0.0148-0.36548-0.0247 0.0296-0.10866 ' +
'0.10865-0.079 0.0741-0.14817 0.13335l-0.40499 0.32597-0.36053-0.44944 1.1359-0.90381h0.61242z" style="fill:#fff"/><p' +
'ath d="m188.75 287.99h-0.42474v0.73095h-0.72602v-0.73095h-1.5064v-0.51859l1.5459-2.2768h0.68651v2.2176h0.42474zm-1.1' +
'508-1.1755q0-0.11359 5e-3 -0.27658 5e-3 -0.16298 0.01-0.30127 0.01-0.14323 0.0148-0.18768h-0.0198q-0.0444 0.0988-0.0' +
'938 0.19262-0.0494 0.0938-0.11853 0.19262l-0.647 0.9779h0.84949z" style="fill:#fff"/></g><g transform="matrix(1 .577' +
'35 0 1 -177.01 -180.58)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="14"><path' +
' d="m179.22 77.337h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.0247 0.02963-0.10865 0.108' +
'66-0.079 0.07408-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="stop-color:#000000"' +
'/><path d="m182.75 76.606h-0.42475v0.73096h-0.72601v-0.73096h-1.5064v-0.51858l1.5459-2.2768h0.6865v2.2176h0.42475zm-' +
'1.1508-1.1755q0-0.11359 5e-3 -0.27658 5e-3 -0.16298 0.01-0.30127 0.01-0.14323 0.0148-0.18768h-0.0198q-0.0445 0.09878' +
'-0.0938 0.19262-0.0494 0.09384-0.11854 0.19262l-0.64699 0.9779h0.84949z" style="stop-color:#000000"/></g><path d="m6' +
'.0277-11.144 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a4000' +
'0"/></g><g id="valueD"><g transform="translate(-235.42 -159.99)"><path d="m235.44 162.25v0.79998l6 3v-0.79998z" styl' +
'e="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke' +
'-width:.124;stroke:#a40000"/><path d="m235.44 161.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef646' +
'4;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d' +
'="m235.44 160.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;sto' +
'p-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.85v0.79998l6 3v-0.7999' +
'8z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:beve' +
'l;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fil' +
'l:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/' +
'><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:no' +
'rmal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.79998l6 3' +
'v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linej' +
'oin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:' +
'none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#' +
'a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-set' +
'tings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.7' +
'9998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stro' +
'ke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape' +
'-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-varia' +
'tion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 15' +
'2.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -156.48)"><path d="' +
'm247.81 158.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color' +
':#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 157.94-6 3v0.79998l6-3z" style="-' +
'inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-wid' +
'th:.124;stroke:#a40000"/><path d="m247.81 157.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-va' +
'riation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81' +
' 156.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#00000' +
'0;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 155.54-6 3v0.79998l6-3z" style="-inkscap' +
'e-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124' +
';stroke:#a40000"/><path d="m247.81 154.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation' +
'-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.94' +
'-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;strok' +
'e-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-strok' +
'e:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke' +
':#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settin' +
'gs:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.' +
'79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linej' +
'oin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;' +
'fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a4000' +
'0"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:norm' +
'al;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6' +
'-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bev' +
'el;stroke-width:.124;stroke:#a40000"/></g><g transform="matrix(1 -.57735 0 1 -177.01 -179.78)" style="stroke-width:.' +
'124;stroke:#a40000" aria-label="13"><path d="m185.22 288.32h-0.74577v-2.0398q0-0.12842 5e-3 -0.33585 0.01-0.20743 0.' +
'0148-0.36548-0.0247 0.0296-0.10866 0.10866-0.079 0.0741-0.14817 0.13335l-0.40499 0.32597-0.36053-0.44944 1.1359-0.90' +
'382h0.61242z" style="fill:#fff"/><path d="m188.53 285.58q0 0.36548-0.22225 0.58279-0.21731 0.21731-0.53834 0.29633v0' +
'.0148q0.42475 0.0494 0.64206 0.25682 0.22225 0.20744 0.22225 0.5581 0 0.30621-0.15311 0.55315-0.14816 0.24201-0.4642' +
'5 0.38524-0.31115 0.13829-0.80504 0.13829-0.57291 0-1.0174-0.19262v-0.63218q0.22719 0.1136 0.47413 0.17286 0.25188 0' +
'.0593 0.46426 0.0593 0.40005 0 0.55809-0.13829 0.16298-0.13829 0.16298-0.39017 0-0.14817-0.0741-0.24695-0.0741-0.103' +
'71-0.26176-0.1531-0.18274-0.0543-0.51365-0.0543h-0.2667v-0.57291h0.27164q0.32597 0 0.49389-0.0593 0.17286-0.0642 0.2' +
'3213-0.16792 0.0642-0.10866 0.0642-0.24694 0-0.18768-0.11854-0.2914-0.11359-0.10865-0.38523-0.10865-0.25188 0-0.4395' +
'6 0.0889-0.18274 0.084-0.31115 0.16792l-0.34572-0.51365q0.20743-0.14816 0.48401-0.24694 0.28151-0.0988 0.66675-0.098' +
'8 0.54327 0 0.85936 0.22225 0.32103 0.21731 0.32103 0.61736z" style="fill:#fff"/></g><g transform="matrix(1 .57735 0' +
' 1 -177.01 -179.78)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="13"><path d="' +
'm179.22 76.937h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.0247 0.02963-0.10865 0.10866-0' +
'.079 0.07408-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="stop-color:#000000"/><p' +
'ath d="m182.53 74.201q0 0.36548-0.22225 0.58279-0.21731 0.21731-0.53834 0.29633v0.01482q0.42474 0.04939 0.64205 0.25' +
'682 0.22225 0.20743 0.22225 0.5581 0 0.30621-0.1531 0.55316-0.14817 0.242-0.46426 0.38523-0.31115 0.13829-0.80504 0.' +
'13829-0.57291 0-1.0174-0.19262v-0.63218q0.22719 0.1136 0.47413 0.17286 0.25189 0.05927 0.46426 0.05927 0.40005 0 0.5' +
'581-0.13829 0.16298-0.13829 0.16298-0.39017 0-0.14817-0.0741-0.24694-0.0741-0.10372-0.26177-0.15311-0.18274-0.05433-' +
'0.51364-0.05433h-0.2667v-0.57291h0.27164q0.32596 0 0.49389-0.05927 0.17286-0.06421 0.23213-0.16792 0.0642-0.10866 0.' +
'0642-0.24694 0-0.18768-0.11853-0.29139-0.1136-0.10866-0.38524-0.10866-0.25188 0-0.43956 0.0889-0.18274 0.08396-0.311' +
'15 0.16792l-0.34572-0.51364q0.20743-0.14817 0.48401-0.24694 0.28152-0.09878 0.66675-0.09878 0.54328 0 0.85937 0.2222' +
'5 0.32103 0.21731 0.32103 0.61736z" style="stop-color:#000000"/></g><path d="m6.0277-10.344 6 3-6 3-6-3z" style="fil' +
'l:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="valueC"><g transform="' +
'translate(-235.42 -159.19)"><path d="m235.44 161.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464' +
';font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d=' +
'"m235.44 160.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop' +
'-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.85v0.79998l6 3v-0.79998' +
'z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel' +
';stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill' +
':#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/>' +
'<path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:nor' +
'mal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.79998l6 3v' +
'-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejo' +
'in:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:n' +
'one;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a' +
'40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.79' +
'998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;strok' +
'e-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape-' +
'stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;s' +
'troke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variat' +
'ion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 152' +
'.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#0000' +
'00;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -155.68)"><path d="m' +
'247.81 157.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:' +
'#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 157.14-6 3v0.79998l6-3z" style="-i' +
'nkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#a40000"/><path d="m247.81 156.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-var' +
'iation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 ' +
'155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000' +
';stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 154.74-6 3v0.79998l6-3z" style="-inkscape' +
'-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-' +
'settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.14-' +
'6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke' +
'-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z" style="-inkscape-stroke' +
':none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:' +
'#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-setting' +
's:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 150.74-6 3v0.7' +
'9998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejo' +
'in:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;f' +
'ill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000' +
'"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:norma' +
'l;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="matrix(1 -.57735 0 1' +
' -177.01 -178.98)" style="stroke-width:.124;stroke:#a40000" aria-label="12"><path d="m185.22 287.94h-0.74577v-2.0398' +
'q0-0.12841 5e-3 -0.33584 0.01-0.20744 0.0148-0.36548-0.0247 0.0296-0.10866 0.10865-0.079 0.0741-0.14817 0.13335l-0.4' +
'0499 0.32597-0.36053-0.44944 1.1359-0.90381h0.61242z" style="fill:#fff"/><path d="m188.67 287.94h-2.4645v-0.51858l0.' +
'88407-0.89394q0.2667-0.27658 0.42968-0.45932 0.16298-0.18767 0.23707-0.33584 0.0741-0.15311 0.0741-0.32597 0-0.21237' +
'-0.11853-0.31609-0.1136-0.10371-0.31115-0.10371-0.2025 0-0.39512 0.0938-0.19261 0.0938-0.40498 0.2667l-0.40499-0.479' +
'08q0.1531-0.13335 0.32102-0.24694 0.17287-0.1136 0.39512-0.18274 0.22718-0.0741 0.54327-0.0741 0.34573 0 0.59267 0.1' +
'2841 0.25188 0.12347 0.38523 0.34078 0.13829 0.21237 0.13829 0.48401 0 0.2914-0.11853 0.5334-0.1136 0.24201-0.33585 ' +
'0.47907-0.21731 0.23707-0.52846 0.52353l-0.45437 0.42474v0.0346h1.536z" style="fill:#fff"/></g><g transform="matrix(' +
'1 .57735 0 1 -177.01 -178.98)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="12"' +
'><path d="m179.22 76.561h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.0247 0.02963-0.10865' +
' 0.10866-0.079 0.07408-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="stop-color:#0' +
'00000"/><path d="m182.67 76.561h-2.4645v-0.51858l0.88406-0.89394q0.2667-0.27658 0.42968-0.45932 0.16299-0.18768 0.23' +
'707-0.33584 0.0741-0.15311 0.0741-0.32597 0-0.21237-0.11853-0.31609-0.11359-0.10372-0.31115-0.10372-0.20249 0-0.3951' +
'1 0.09384t-0.40499 0.2667l-0.40499-0.47907q0.15311-0.13335 0.32103-0.24694 0.17286-0.1136 0.39511-0.18274 0.22719-0.' +
'07408 0.54328-0.07408 0.34572 0 0.59267 0.12841 0.25188 0.12347 0.38523 0.34078 0.13829 0.21237 0.13829 0.48401 0 0.' +
'2914-0.11854 0.5334-0.11359 0.24201-0.33584 0.47907-0.21731 0.23707-0.52846 0.52352l-0.45438 0.42474v0.03457h1.536z"' +
' style="stop-color:#000000"/></g><path d="m6.0277-9.5445 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-li' +
'nejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="valueB"><g transform="translate(-235.42 -158.39)"><path d' +
'="m235.44 160.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;sto' +
'p-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.85v0.79998l6 3v-0.7999' +
'8z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:beve' +
'l;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fil' +
'l:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/' +
'><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:no' +
'rmal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.79998l6 3' +
'v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linej' +
'oin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:' +
'none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#' +
'a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-set' +
'tings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.7' +
'9998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stro' +
'ke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape' +
'-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-varia' +
'tion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 15' +
'2.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -154.88)"><path d="' +
'm247.81 157.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color' +
':#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 156.34-6 3v0.79998l6-3z" style="-' +
'inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-wid' +
'th:.124;stroke:#a40000"/><path d="m247.81 155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-va' +
'riation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81' +
' 154.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#00000' +
'0;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscap' +
'e-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124' +
';stroke:#a40000"/><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation' +
'-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34' +
'-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;strok' +
'e-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-strok' +
'e:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke' +
':#a40000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settin' +
'gs:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.' +
'79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linej' +
'oin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;' +
'fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a4000' +
'0"/></g><g transform="matrix(1 -.57735 0 1 -177.01 -178.18)" style="stroke-width:.124;stroke:#a40000" aria-label="11' +
'"><path d="m185.22 287.52h-0.74577v-2.0398q0-0.12841 5e-3 -0.33585 0.01-0.20743 0.0148-0.36547-0.0247 0.0296-0.10866' +
' 0.10865-0.079 0.0741-0.14817 0.13335l-0.40499 0.32597-0.36053-0.44944 1.1359-0.90382h0.61242z" style="fill:#fff"/><' +
'path d="m188.05 287.52h-0.74577v-2.0398q0-0.12841 5e-3 -0.33585 0.01-0.20743 0.0148-0.36547-0.0247 0.0296-0.10865 0.' +
'10865-0.079 0.0741-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="fill:#fff"/></g><' +
'g transform="matrix(1 .57735 0 1 -177.01 -178.18)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40' +
'000" aria-label="11"><path d="m179.22 76.137h-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.' +
'0247 0.02963-0.10865 0.10866-0.079 0.07408-0.14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" ' +
'style="stop-color:#000000"/><path d="m182.05 76.137h-0.74578v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.3' +
'6548-0.0247 0.02963-0.10866 0.10866-0.079 0.07408-0.14816 0.13335l-0.40499 0.32597-0.36054-0.44944 1.1359-0.90382h0.' +
'61243z" style="stop-color:#000000"/></g><path d="m6.0277-8.7445 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;st' +
'roke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="valueA"><g transform="translate(-235.42 -157.59)">' +
'<path d="m235.44 159.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:nor' +
'mal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 159.05v0.79998l6 3v' +
'-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejo' +
'in:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:n' +
'one;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a' +
'40000"/><path d="m235.44 157.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79' +
'998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;strok' +
'e-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-' +
'stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;s' +
'troke:#a40000"/><path d="m235.44 155.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variat' +
'ion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154' +
'.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#0000' +
'00;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-i' +
'nkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#a40000"/><path d="m235.44 152.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;fon' +
't-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g trans' +
'form="translate(-235.78 -154.08)"><path d="m247.81 156.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef292' +
'9;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d' +
'="m247.81 155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-col' +
'or:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 154.74-6 3v0.79998l6-3z" style=' +
'"-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-w' +
'idth:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-' +
'variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.' +
'81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z" style="-inksc' +
'ape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.1' +
'24;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variati' +
'on-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 150.' +
'74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;str' +
'oke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style="-inkscape-str' +
'oke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stro' +
'ke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="matrix(1 -' +
'.57735 0 1 -177.01 -177.38)" style="stroke-width:.124;stroke:#a40000" aria-label="10"><path d="m185.22 287.12h-0.745' +
'77v-2.0398q0-0.12841 5e-3 -0.33585 0.01-0.20743 0.0148-0.36548-0.0247 0.0296-0.10866 0.10866-0.079 0.0741-0.14817 0.' +
'13335l-0.40499 0.32597-0.36053-0.44944 1.1359-0.90382h0.61242z" style="fill:#fff"/><path d="m188.65 285.36q0 0.56798' +
'-0.12347 0.97297-0.11853 0.40498-0.39017 0.6223-0.2667 0.21731-0.72108 0.21731-0.63711 0-0.93345-0.47908-0.29633-0.4' +
'8401-0.29633-1.3335 0-0.57291 0.11853-0.9779 0.11854-0.40498 0.39017-0.6223 0.27164-0.21731 0.72108-0.21731 0.63218 ' +
'0 0.93345 0.47908 0.30127 0.47907 0.30127 1.3384zm-1.7237 0q0 0.60255 0.10372 0.90876 0.10372 0.30127 0.38523 0.3012' +
'7 0.27658 0 0.38524-0.30127 0.10865-0.30127 0.10865-0.90876 0-0.60254-0.10865-0.90875-0.10866-0.30621-0.38524-0.3062' +
'1-0.28151 0-0.38523 0.30621t-0.10372 0.90875z" style="fill:#fff"/></g><g transform="matrix(1 .57735 0 1 -177.01 -177' +
'.38)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="10"><path d="m179.22 75.739h' +
'-0.74577v-2.0398q0-0.12841 5e-3 -0.33584 0.01-0.20743 0.0148-0.36548-0.0247 0.02963-0.10865 0.10866-0.079 0.07408-0.' +
'14817 0.13335l-0.40499 0.32597-0.36054-0.44944 1.136-0.90382h0.61242z" style="stop-color:#000000"/><path d="m182.65 ' +
'73.976q0 0.56797-0.12347 0.97296-0.11854 0.40499-0.39017 0.6223-0.2667 0.21731-0.72108 0.21731-0.63712 0-0.93345-0.4' +
'7907-0.29634-0.48401-0.29634-1.3335 0-0.57291 0.11854-0.9779 0.11853-0.40499 0.39017-0.6223t0.72108-0.21731q0.63218 ' +
'0 0.93345 0.47907t0.30127 1.3384zm-1.7237 0q0 0.60254 0.10371 0.90876 0.10372 0.30127 0.38524 0.30127 0.27658 0 0.38' +
'523-0.30127 0.10866-0.30127 0.10866-0.90876 0-0.60254-0.10866-0.90876-0.10865-0.30621-0.38523-0.30621-0.28152 0-0.38' +
'524 0.30621-0.10371 0.30621-0.10371 0.90876z" style="stop-color:#000000"/></g><path d="m6.0277-7.9445 6 3-6 3-6-3z" ' +
'style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value9"><g t' +
'ransform="translate(-235.42 -156.79)"><path d="m235.44 159.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fi' +
'll:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"' +
'/><path d="m235.44 158.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:n' +
'ormal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.79998l6 ' +
'3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-line' +
'join:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke' +
':none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:' +
'#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-se' +
'ttings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.' +
'79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;str' +
'oke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscap' +
'e-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124' +
';stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 1' +
'52.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#00' +
'0000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -153.28)"><path d=' +
'"m247.81 155.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-colo' +
'r:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 154.74-6 3v0.79998l6-3z" style="' +
'-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-wi' +
'dth:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-v' +
'ariation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.8' +
'1 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#0000' +
'00;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z" style="-inksca' +
'pe-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.12' +
'4;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variatio' +
'n-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 150.7' +
'4-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stro' +
'ke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style="-inkscape-stro' +
'ke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;strok' +
'e:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-setti' +
'ngs:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="matrix(1 -.' +
'57735 0 1 -177.01 -176.58)" style="stroke-width:.124;stroke:#a40000" aria-label="9"><path d="m187.24 284.69q0 0.3012' +
'7-0.0445 0.60255-0.0444 0.29633-0.15804 0.56303-0.10866 0.26176-0.31115 0.4692-0.19756 0.20249-0.50871 0.32102-0.306' +
'21 0.1136-0.74577 0.1136-0.10372 0-0.24201-0.01-0.13829-5e-3 -0.23213-0.0247v-0.59761q0.0988 0.0198 0.20744 0.0346 0' +
'.10865 0.0148 0.21731 0.0148 0.43956 0 0.67663-0.13829 0.242-0.14323 0.34078-0.39017 0.0988-0.24695 0.11359-0.57291h' +
'-0.0296q-0.10372 0.17286-0.27658 0.29139-0.16792 0.11853-0.50376 0.11853-0.44944 0-0.72108-0.29139-0.27164-0.2914-0.' +
'27164-0.8248 0-0.57291 0.32103-0.89887 0.32596-0.32597 0.87912-0.32597 0.36054 0 0.65193 0.16792 0.2914 0.16299 0.46' +
'426 0.50871 0.17286 0.34078 0.17286 0.86924zm-1.2742-0.94826q-0.21731 0-0.3556 0.14816-0.13829 0.14817-0.13829 0.469' +
'2 0 0.25682 0.11853 0.40499 0.11853 0.14816 0.36054 0.14816 0.16792 0 0.28646-0.0741 0.12347-0.0741 0.19261-0.18274 ' +
'0.0692-0.11359 0.0692-0.23213 0-0.16298-0.0593-0.32103-0.0593-0.15804-0.1778-0.25682-0.11853-0.10371-0.29633-0.10371' +
'z" style="fill:#fff"/></g><g transform="matrix(1 .57735 0 1 -177.01 -176.58)" style="fill:#fff;stop-color:#000000;st' +
'roke-width:.124;stroke:#a40000" aria-label="9"><path d="m181.24 73.312q0 0.30127-0.0444 0.60254-0.0445 0.29633-0.158' +
'04 0.56303-0.10866 0.26176-0.31115 0.46919-0.19756 0.2025-0.50871 0.32103-0.30621 0.11359-0.74577 0.11359-0.10371 0-' +
'0.242-0.0099-0.13829-0.0049-0.23213-0.02469v-0.5976q0.0988 0.01975 0.20743 0.03457 0.10866 0.01482 0.21731 0.01482 0' +
'.43956 0 0.67663-0.13829 0.24201-0.14323 0.34078-0.39017 0.0988-0.24694 0.1136-0.57291h-0.0296q-0.10372 0.17286-0.27' +
'658 0.2914-0.16792 0.11853-0.50377 0.11853-0.44944 0-0.72108-0.2914-0.27164-0.29139-0.27164-0.82479 0-0.57291 0.3210' +
'3-0.89888 0.32597-0.32597 0.87912-0.32597 0.36054 0 0.65194 0.16792 0.29139 0.16298 0.46425 0.50871 0.17286 0.34078 ' +
'0.17286 0.86924zm-1.2742-0.94827q-0.21731 0-0.3556 0.14817-0.13829 0.14817-0.13829 0.4692 0 0.25682 0.11854 0.40499 ' +
'0.11853 0.14817 0.36053 0.14817 0.16793 0 0.28646-0.07408 0.12347-0.07408 0.19262-0.18274 0.0691-0.11359 0.0691-0.23' +
'213 0-0.16298-0.0593-0.32103-0.0593-0.15804-0.1778-0.25682-0.11853-0.10372-0.29633-0.10372z" style="stop-color:#0000' +
'00"/></g><path d="m6.0277-7.1444 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-widt' +
'h:.124;stroke:#a40000"/></g><g id="value8"><g transform="translate(-235.42 -155.99)"><path d="m235.44 158.25v0.79998' +
'l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-l' +
'inejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 157.45v0.79998l6 3v-0.79998z" style="-inkscape-str' +
'oke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stro' +
'ke:#a40000"/><path d="m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation' +
'-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.85' +
'v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;' +
'stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.79998l6 3v-0.79998z" style="-inks' +
'cape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.' +
'124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-v' +
'ariation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.4' +
'4 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:' +
'#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 152.65v0.79998l6 3v-0.79998z" styl' +
'e="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke' +
'-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -152.48)"><path d="m247.81 154.74-6 3v0.79998l6-3z"' +
' style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;s' +
'troke-width:.124;stroke:#a40000"/><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef292' +
'9;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d' +
'="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-col' +
'or:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v0.79998l6-3z" style=' +
'"-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-w' +
'idth:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-' +
'variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.' +
'81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998l6-3z" style="-inksc' +
'ape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.1' +
'24;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variati' +
'on-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="mat' +
'rix(1 -.57735 0 1 -177.01 -175.78)" style="stroke-width:.124;stroke:#a40000" aria-label="8"><path d="m186.01 282.75q' +
'0.30621 0 0.55809 0.0938 0.25682 0.0938 0.40993 0.28151 0.15804 0.18768 0.15804 0.47414 0 0.31609-0.18273 0.52352-0.' +
'1778 0.2025-0.44944 0.33585 0.18767 0.0988 0.3556 0.23212 0.17286 0.12841 0.27657 0.30621 0.10866 0.17287 0.10866 0.' +
'40993 0 0.2914-0.15804 0.50871-0.15311 0.21731-0.43463 0.33584-0.27657 0.11854-0.64205 0.11854-0.59267 0-0.91864-0.2' +
'5189-0.32102-0.25188-0.32102-0.69144 0-0.36548 0.19755-0.58773 0.19756-0.22225 0.47907-0.36054-0.23706-0.14817-0.404' +
'98-0.3556-0.16299-0.21237-0.16299-0.52846 0-0.28152 0.15805-0.4692 0.15804-0.18767 0.41486-0.28151 0.26176-0.0938 0.' +
'5581-0.0938zm-5e-3 0.54328q-0.18274 0-0.30621 0.0938-0.11854 0.0938-0.11854 0.2667 0 0.18273 0.12842 0.29633 0.12841' +
' 0.11359 0.30127 0.19756 0.16792-0.079 0.29633-0.18768 0.12841-0.1136 0.12841-0.30621 0-0.17287-0.12347-0.2667-0.118' +
'53-0.0938-0.30621-0.0938zm-0.54328 2.0892q0 0.19261 0.13335 0.31609 0.13829 0.12347 0.40499 0.12347 0.27658 0 0.4148' +
'7-0.11853 0.13829-0.11854 0.13829-0.31609 0-0.13335-0.079-0.23213-0.079-0.10372-0.19755-0.18274-0.1136-0.084-0.23707' +
'-0.15311l-0.0642-0.0346q-0.22719 0.10866-0.37042 0.25189-0.14323 0.14322-0.14323 0.34572z" style="fill:#fff"/></g><g' +
' transform="matrix(1 .57735 0 1 -177.01 -175.78)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a400' +
'00" aria-label="8"><path d="m180.01 71.363q0.30621 0 0.5581 0.09384 0.25682 0.09384 0.40992 0.28152 0.15805 0.18768 ' +
'0.15805 0.47413 0 0.31609-0.18274 0.52352-0.1778 0.20249-0.44944 0.33584 0.18768 0.09878 0.3556 0.23213 0.17286 0.12' +
'841 0.27658 0.30621 0.10865 0.17286 0.10865 0.40993 0 0.29139-0.15804 0.50871-0.15311 0.21731-0.43462 0.33584-0.2765' +
'8 0.11853-0.64206 0.11853-0.59266 0-0.91863-0.25188-0.32103-0.25188-0.32103-0.69144 0-0.36548 0.19756-0.58773 0.1975' +
'5-0.22225 0.47907-0.36054-0.23707-0.14817-0.40499-0.3556-0.16298-0.21237-0.16298-0.52846 0-0.28152 0.15804-0.46919 0' +
'.15805-0.18768 0.41487-0.28152 0.26176-0.09384 0.55809-0.09384zm-5e-3 0.54328q-0.18274 0-0.30621 0.09384-0.11853 0.0' +
'9384-0.11853 0.2667 0 0.18274 0.12841 0.29633 0.12841 0.1136 0.30127 0.19756 0.16792-0.07902 0.29634-0.18768 0.12841' +
'-0.11359 0.12841-0.30621 0-0.17286-0.12348-0.2667-0.11853-0.09384-0.30621-0.09384zm-0.54327 2.0892q0 0.19262 0.13335' +
' 0.31609 0.13828 0.12347 0.40498 0.12347 0.27658 0 0.41487-0.11853 0.13829-0.11853 0.13829-0.31609 0-0.13335-0.079-0' +
'.23213-0.079-0.10372-0.19756-0.18274-0.11359-0.08396-0.23706-0.15311l-0.0642-0.03457q-0.22719 0.10866-0.37042 0.2518' +
'8-0.14322 0.14323-0.14322 0.34572z" style="stop-color:#000000"/></g><path d="m6.0277-6.3445 6 3-6 3-6-3z" style="fil' +
'l:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value7"><g transform="' +
'translate(-235.42 -155.19)"><path d="m235.44 157.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464' +
';font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d=' +
'"m235.44 156.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop' +
'-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998' +
'z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel' +
';stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill' +
':#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/>' +
'<path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:nor' +
'mal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v' +
'-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejo' +
'in:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 152.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:n' +
'one;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a' +
'40000"/></g><g transform="translate(-235.78 -151.68)"><path d="m247.81 153.94-6 3v0.79998l6-3z" style="-inkscape-str' +
'oke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stro' +
'ke:#a40000"/><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-sett' +
'ings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 152.34-6 3v' +
'0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-lin' +
'ejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inkscape-stroke:non' +
'e;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40' +
'000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:no' +
'rmal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.94-6 3v0.79998' +
'l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:b' +
'evel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:' +
'#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><' +
'/g><g transform="matrix(1 -.57735 0 1 -177.01 -174.98)" style="stroke-width:.124;stroke:#a40000" aria-label="7"><pat' +
'h d="m185.15 285.92 1.3236-2.8991h-1.7385v-0.62724h2.5287v0.46919l-1.3335 3.0572z" style="fill:#fff"/></g><g transfo' +
'rm="matrix(1 .57735 0 1 -177.01 -174.98)" style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria' +
'-label="7"><path d="m179.15 74.537 1.3236-2.8991h-1.7385v-0.62724h2.5287v0.4692l-1.3335 3.0572z" style="stop-color:#' +
'000000"/></g><path d="m6.0277-5.5444 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;stroke-' +
'width:.124;stroke:#a40000"/></g><g id="value6"><g transform="translate(-235.42 -154.39)"><path d="m235.44 156.65v0.7' +
'9998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stro' +
'ke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape' +
'-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m235.44 155.05v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-varia' +
'tion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 15' +
'4.25v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-' +
'inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-wid' +
'th:.124;stroke:#a40000"/><path d="m235.44 152.65v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;fo' +
'nt-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g tran' +
'sform="translate(-235.78 -150.88)"><path d="m247.81 153.14-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef29' +
'29;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path ' +
'd="m247.81 152.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-co' +
'lor:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style' +
'="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-' +
'width:.124;stroke:#a40000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font' +
'-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247' +
'.81 149.94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#00' +
'0000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inks' +
'cape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.' +
'124;stroke:#a40000"/></g><g transform="matrix(1 -.57735 0 1 -177.01 -174.18)" style="stroke-width:.124;stroke:#a4000' +
'0" aria-label="6"><path d="m184.77 284.02q0-0.30621 0.0444-0.60254 0.0445-0.29634 0.1531-0.5581 0.1136-0.2667 0.3111' +
'5-0.46919 0.2025-0.20743 0.50871-0.32103 0.31115-0.11853 0.75071-0.11853 0.10372 0 0.242 0.01 0.13829 5e-3 0.23213 0' +
'.0247v0.59761q-0.0938-0.0247-0.20743-0.0346-0.10866-0.0148-0.21731-0.0148-0.43956 0-0.68157 0.13829-0.23707 0.13828-' +
'0.33584 0.39017-0.0988 0.24694-0.1136 0.57291h0.0296q0.0988-0.17286 0.28151-0.2914 0.18768-0.11853 0.48401-0.11853 0' +
'.46426 0 0.7359 0.2914 0.27164 0.29139 0.27164 0.82479 0 0.57291-0.32597 0.89888-0.32103 0.32596-0.87418 0.32596-0.3' +
'6054 0-0.65194-0.16298-0.29139-0.16792-0.46425-0.5087-0.17286-0.34573-0.17286-0.87419zm1.2742 0.94827q0.21731 0 0.35' +
'56-0.14817 0.13829-0.1531 0.13829-0.46919 0-0.25682-0.11853-0.40499-0.11854-0.14817-0.36054-0.14817-0.16299 0-0.2864' +
'6 0.0741-0.12347 0.0691-0.19262 0.18273-0.0691 0.1136-0.0691 0.23213 0 0.16299 0.0593 0.32103 0.0593 0.15311 0.1778 ' +
'0.25682 0.12347 0.10372 0.29633 0.10372z" style="fill:#fff"/></g><g transform="matrix(1 .57735 0 1 -177.01 -174.18)"' +
' style="fill:#fff;stop-color:#000000;stroke-width:.124;stroke:#a40000" aria-label="6"><path d="m178.77 72.635q0-0.30' +
'621 0.0445-0.60254 0.0444-0.29633 0.15311-0.5581 0.11359-0.2667 0.31115-0.46919 0.20249-0.20743 0.5087-0.32103 0.311' +
'15-0.11853 0.75071-0.11853 0.10372 0 0.24201 0.0099 0.13829 0.0049 0.23213 0.02469v0.59761q-0.0938-0.0247-0.20744-0.' +
'03457-0.10865-0.01482-0.21731-0.01482-0.43956 0-0.68156 0.13829-0.23707 0.13829-0.33585 0.39017-0.0988 0.24694-0.113' +
'59 0.57291h0.0296q0.0988-0.17286 0.28152-0.29139 0.18767-0.11853 0.48401-0.11853 0.46425 0 0.73589 0.2914 0.27164 0.' +
'29139 0.27164 0.82479 0 0.57291-0.32597 0.89888-0.32102 0.32597-0.87418 0.32597-0.36054 0-0.65193-0.16298-0.2914-0.1' +
'6792-0.46426-0.50871-0.17286-0.34572-0.17286-0.87418zm1.2742 0.94827q0.21732 0 0.3556-0.14817 0.13829-0.15311 0.1382' +
'9-0.4692 0-0.25682-0.11853-0.40499-0.11853-0.14817-0.36054-0.14817-0.16298 0-0.28645 0.07408-0.12348 0.06914-0.19262' +
' 0.18274-0.0691 0.11359-0.0691 0.23213 0 0.16298 0.0593 0.32103 0.0593 0.15311 0.1778 0.25682 0.12347 0.10372 0.2963' +
'3 0.10372z" style="stop-color:#000000"/></g><path d="m6.0277-4.7444 6 3-6 3-6-3z" style="fill:#c00;stop-color:#00000' +
'0;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value5"><g transform="translate(-235.42 -153.5' +
'9)"><path d="m235.44 155.85v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings' +
':normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 155.05v0.79998l' +
'6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-li' +
'nejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 154.25v0.79998l6 3v-0.79998z" style="-inkscape-stro' +
'ke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;strok' +
'e:#a40000"/><path d="m235.44 153.45v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-' +
'settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m235.44 152.65v' +
'0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;s' +
'troke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g transform="translate(-235.78 -150.08)"><path d="m247.' +
'81 152.34-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000' +
'000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 151.54-6 3v0.79998l6-3z" style="-inksc' +
'ape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.1' +
'24;stroke:#a40000"/><path d="m247.81 150.74-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variati' +
'on-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.' +
'94-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;str' +
'oke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m247.81 149.14-6 3v0.79998l6-3z" style="-inkscape-str' +
'oke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stro' +
'ke:#a40000"/></g><g transform="matrix(1 -.57735 0 1 -177.01 -173.38)" style="stroke-width:.124;stroke:#a40000" aria-' +
'label="5"><path d="m186.08 282.84q0.32103 0 0.57291 0.12348 0.25189 0.12347 0.39511 0.36053 0.14817 0.23707 0.14817 ' +
'0.58773 0 0.57291-0.3556 0.90382-0.3556 0.32597-1.052 0.32597-0.27658 0-0.52352-0.0494-0.24201-0.0494-0.42475-0.1432' +
'3v-0.64206q0.18274 0.0938 0.43956 0.16299 0.25682 0.0642 0.48401 0.0642 0.33091 0 0.50377-0.13335 0.1778-0.13829 0.1' +
'778-0.42474 0-0.5334-0.70626-0.5334-0.13829 0-0.28646 0.0296-0.14816 0.0247-0.24694 0.0494l-0.29633-0.15804 0.13335-' +
'1.7978h1.9114v0.63218h-1.2594l-0.0642 0.69144q0.084-0.0148 0.1778-0.0296 0.0988-0.0198 0.27164-0.0198z" style="fill:' +
'#fff"/></g><g transform="matrix(1 .57735 0 1 -177.01 -173.38)" style="fill:#fff;stop-color:#000000;stroke-width:.124' +
';stroke:#a40000" aria-label="5"><path d="m180.08 71.46q0.32102 0 0.57291 0.12347 0.25188 0.12347 0.39511 0.36054 0.1' +
'4817 0.23707 0.14817 0.58773 0 0.57291-0.3556 0.90382-0.3556 0.32597-1.052 0.32597-0.27658 0-0.52352-0.04939-0.24201' +
'-0.04939-0.42474-0.14323v-0.64206q0.18273 0.09384 0.43956 0.16298 0.25682 0.06421 0.48401 0.06421 0.3309 0 0.50376-0' +
'.13335 0.1778-0.13829 0.1778-0.42474 0-0.5334-0.70626-0.5334-0.13829 0-0.28645 0.02963-0.14817 0.02469-0.24695 0.049' +
'39l-0.29633-0.15804 0.13335-1.7978h1.9114v0.63218h-1.2594l-0.0642 0.69144q0.084-0.01482 0.1778-0.02963 0.0988-0.0197' +
'6 0.27164-0.01976z" style="stop-color:#000000"/></g><path d="m6.0277-3.9444 6 3-6 3-6-3z" style="fill:#c00;stop-colo' +
'r:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value4"><path d="m0.02772 2.2555v0.799' +
'98l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke' +
'-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m0.02772 1.4555v0.79998l6 3v-0.79998z" style="-inkscape-' +
'stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;s' +
'troke:#a40000"/><path d="m0.02772 0.65551v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m0.02772-' +
'0.14447v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#' +
'000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 2.2555-6 3v0.79998l6-3z" style="-in' +
'kscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width' +
':.124;stroke:#a40000"/><path d="m12.028 1.4555-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 0' +
'.65551-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000' +
';stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028-0.14447-6 3v0.79998l6-3z" style="-inkscap' +
'e-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124' +
';stroke:#a40000"/><path d="m6.0277-3.1445 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;st' +
'roke-width:.124;stroke:#a40000"/></g><g id="value3"><path d="m0.02772 2.2555v0.79998l6 3v-0.79998z" style="-inkscape' +
'-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;' +
'stroke:#a40000"/><path d="m0.02772 1.4555v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m0.02772 ' +
'0.65551v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#' +
'000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 2.2555-6 3v0.79998l6-3z" style="-in' +
'kscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width' +
':.124;stroke:#a40000"/><path d="m12.028 1.4555-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-vari' +
'ation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 0' +
'.65551-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000' +
';stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m6.0277-2.3445 6 3-6 3-6-3z" style="fill:#c00;sto' +
'p-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value2"><path d="m0.02772 2.2555' +
'v0.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;' +
'stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m0.02772 1.4555v0.79998l6 3v-0.79998z" style="-ink' +
'scape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:' +
'.124;stroke:#a40000"/><path d="m12.028 2.2555-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-varia' +
'tion-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 1.' +
'4555-6 3v0.79998l6-3z" style="-inkscape-stroke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;s' +
'troke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m6.0277-1.5445 6 3-6 3-6-3z" style="fill:#c00;stop-' +
'color:#000000;stroke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/></g><g id="value1"><path d="m0.02772 2.2555v0' +
'.79998l6 3v-0.79998z" style="-inkscape-stroke:none;fill:#ef6464;font-variation-settings:normal;stop-color:#000000;st' +
'roke-linejoin:bevel;stroke-width:.124;stroke:#a40000"/><path d="m12.028 2.2555-6 3v0.79998l6-3z" style="-inkscape-st' +
'roke:none;fill:#ef2929;font-variation-settings:normal;stop-color:#000000;stroke-linejoin:bevel;stroke-width:.124;str' +
'oke:#a40000"/><path d="m6.0277-0.74453 6 3-6 3-6-3z" style="fill:#c00;stop-color:#000000;stroke-linejoin:bevel;strok' +
'e-width:.124;stroke:#a40000"/></g></defs></svg>'
}

/***[ drawWorld/RobotsWorldSvg.js ] *************************/
/**
* Class for building a concrete image of robot's world
*/
function RobotsWorldSvg(
    json        // JSON describing template
) {
    Object.defineProperty(this, 'template', { 'enumerable': true, 'value': json });
}

// SVG image
RobotsWorldSvg.prototype.svg = null;

// Used template
RobotsWorldSvg.prototype.template = null;

// Function to call after loading template
RobotsWorldSvg.prototype.onloaded = null;

/**
* Start loading of template
* @return undefined
*/
RobotsWorldSvg.prototype.load = function() {
    var self = this;
    var oReq = new XMLHttpRequest();
    oReq.addEventListener("load", function() {
        self.svg = oReq.responseXML;
        if (typeof self.onloaded == 'function')
            self.onloaded.call(self);
    });
    oReq.open("GET", this.getUrl());
    oReq.send();
}

/**
* Create blob URI of svg image
* @return String    Blob URI (or undefined)
*/
RobotsWorldSvg.prototype.getBlobUrl = function() {
    if (!this.svg)
        return;
    this.revokeBlobUrl();
    this._lastBlobUrl = window.URL.createObjectURL(new Blob([new XMLSerializer().serializeToString(this.svg)], {type: 'image/svg+xml'}));
    return this._lastBlobUrl;
}

/**
* Release the latest blob URI (if any)
* @return undefined
*/
RobotsWorldSvg.prototype.revokeBlobUrl = function() {
    if (this._lastBlobUrl)
        window.URL.revokeObjectURL(this._lastBlobUrl);
    this._lastBlobUrl = undefined;
}

/**
* Constructs data URI of SVG template
* @return String    Data URI
*/
RobotsWorldSvg.prototype.getUrl = function() {
    return 'data:image/svg+xml;charset=UTF-8,' + encodeURIComponent(this.template.content);
}

/**
* Set image dimension to given width and height with respect to the given offsets
* @return undefined
*/
RobotsWorldSvg.prototype.setDimension = function(
    offsetLeft,     // Horizontal offset
    offsetTop,      // Vertical offset
    width,          // New image width
    height          // New image height
) {
    if (!this.svg)
        return;
    var root = this.svg.documentElement;
    var factor = this.template.factor ? this.template.factor : 1;
    root.setAttribute('viewBox', [-offsetLeft, -offsetTop, width, height].join(' '));
    root.setAttribute('width', factor * width + this.template.mainUnit);
    root.setAttribute('height', factor * height + this.template.mainUnit);
}

/**
* Add another group at end of image
* @return SVGGroupElement   Created element
*/
RobotsWorldSvg.prototype.addGroup = function() {
    if (!this.svg)
        return null;
    var g = this.svg.createElementNS(this.svg.documentElement.namespaceURI, 'g');
    this.svg.documentElement.appendChild(g);
    return g;
}

// Namespace of xlink
RobotsWorldSvg.NAMESPACE_XLINK = 'http://www.w3.org/1999/xlink';

/**
* Add a use SVG element
* @return SVGUseElement     Created element
*/
RobotsWorldSvg.prototype.addSymbol = function(
    group,      // Optional: Group element that should be parent node (if undefined use documentElement)
    what,       // Property name of template that has the symbol's id as value (without prefix id, e. g. 'Tilde' or 'RobotFacingNorth')
    x,          // Horizontal position on image in local units
    y,          // Vertical position on image in local units
    variant     // Optional: Index if this.template['id' + what] is array
) {
    const id = ((variant !== undefined) && (this.template['id' + what] instanceof Array)) ? this.template['id' + what][variant] : this.template['id' + what];
    if (!this.svg || !id)
        return null;
    var use = this.svg.createElementNS(this.svg.documentElement.namespaceURI, 'use');
    (group ? group : this.svg.documentElement).appendChild(use);
    use.setAttribute('x', x);
    use.setAttribute('y', y);
    use.setAttributeNS(RobotsWorldSvg.NAMESPACE_XLINK, 'href', '#' + id);
    return use;
}

/***[ drawWorld/RobotsWorldDrawer.js ] *************************/
/**
* Create a SVG image showing a concrete RobotsWorldData instance
*/
function RobotsWorldDrawer(
    world,          // RobotsWorld instance to draw
    robot,          // Data of robot: { x: 'x coordinate', y: 'y coordinate', direction: '0 meaning north, 1 meaning west, …' }
    template        // JSON describing template
) {
    Object.defineProperty(this, 'world', { 'enumerable': true, 'value': world });
    Object.defineProperty(this, 'template', { 'enumerable': true, 'value': template });
    Object.defineProperty(this, 'svg', { 'enumerable': true, 'value': new RobotsWorldSvg(template) });
    this._robotX = robot.x;
    this._robotY = robot.y;
    this._robotDirection = robot.direction;
}

/**
* Start to draw
* @return undefined
*/
RobotsWorldDrawer.prototype.draw = function() {
    var self = this;
    this.svg.onloaded = function() { self._draw(); };
    this.svg.load();
}

// RobotsWorldData instance to draw
RobotsWorldDrawer.prototype.world = null;

// JSON describing template
RobotsWorldDrawer.prototype.template = null;

// Used RobotsWorldSvg instance
RobotsWorldDrawer.prototype.svg = null;

// Function to call after creating image
RobotsWorldDrawer.prototype.onloaded = null;

/**
* Draw SVG
* @return undefined
*/
RobotsWorldDrawer.prototype._draw = function() {
    this.svg.setDimension(this.template.offsetLeft, this.template.offsetTop, this.calcImageWidth(), this.calcImageHeight());
    /*!*/this._drawIsometric();
    if (typeof this.onloaded)
        this.onloaded.call(this, this.svg);
}

/**
* Get x dimension of world in tilde units
* @return Number    X dimension
*/
RobotsWorldDrawer.prototype._countXTildes = function() {
    return this.world.height;
}

/**
* Get y dimension of world in tilde units
* @return Number    Y dimension
*/
RobotsWorldDrawer.prototype._countYTildes = function() {
    return this.world.width;
}

/**
* Calculate vertical shift of ground
* @return Number    Shift in local units
*/
RobotsWorldDrawer.prototype.calcBaselineShift = function() {
    return this.template.robotHeight + this.template.maxValueHeight;
}

/**
* Calculate width of image
* @return Number    Image width in local units
*/
RobotsWorldDrawer.prototype.calcImageWidth = function() {
    return this.template.offsetLeft + this.template.offsetRight + this.template.tildeWidth + this.template.additionalWidthPerX * (this._countXTildes() - 1) + this.template.additionalWidthPerY * (this._countYTildes() - 1);
}

/**
* Calculate height of image
* @return Number    Image height in local units
*/
RobotsWorldDrawer.prototype.calcImageHeight = function() {
    return this.template.offsetTop + this.template.offsetBottom + this.template.tildeHeight + this.template.additionalHeightPerY * (this._countYTildes() - 1) + this.template.additionalHeightPerX * (this._countXTildes() - 1) + this.calcBaselineShift();
}

/**
* Is robot on that tilde?
* @return boolean   Robot's position?
*/
RobotsWorldDrawer.prototype.isRobotsPosition = function(
    x,          // Horizontal position in tilde units
    y           // Vertical position in tilde units
) {
    return ((this._robotY == this.world.height - x - 1) && (this._robotX == y));
}

/**
* Convert perspective value to string
* @return String    Perspective name starting with capital letter
*/
RobotsWorldDrawer.perspectiveToString = function(
    perspective     // Perspective to convert
) {
    switch (perspective % 4) {
        case 3:
            return 'East';
        case 2:
            return 'South';
        case 1:
            return 'West';
        case 0:
        default:
            return 'North';
    }
}

/**
* Draw a tilde and all on that tilde
* @return SVGGroupElement   Group with all symbols
*/
RobotsWorldDrawer.prototype._drawTilde = function(
    tilde,  // RobotsTildeData instance to draw
    x,      // Tilde's horizontal position in local units
    y,      // Tilde's vertical position in local units
    tildeX, // Tilde's horizontal position in tilde units
    tildeY  // Tilde's vertical position in tilde units
) {
    var g = this.svg.addGroup();
    if (!tilde.isAccesible()) {
        this.svg.addSymbol(g, 'Blocked', x, y);
        const maxX = this._countXTildes() - 1;
        const maxY = this._countYTildes() - 1;
        if ((tildeX == 0) && (tildeY == Math.floor(maxY / 2)))
            this.svg.addSymbol(g, 'SignNorth', x, y);
        if ((tildeY == 0) && (tildeX == Math.floor(maxX / 2)))
            this.svg.addSymbol(g, 'SignWest', x, y);
        if ((tildeX == maxX) && (tildeY == Math.floor(maxY / 2)))
            this.svg.addSymbol(g, 'SignSouth', x, y);
        if ((tildeY == maxY) && (tildeX == Math.floor(maxX / 2)))
            this.svg.addSymbol(g, 'SignEast', x, y);
    } else {
        this.svg.addSymbol(g, 'Tilde', x, y);
        const value = tilde.read();

        const maxX = this._countXTildes() - 1;
        const maxY = this._countYTildes() - 1;
        if ((tildeX == 0) && (tildeY == Math.floor(maxY / 2)))
            this.svg.addSymbol(g, 'SignNorth', x, y);
        if ((tildeY == 0) && (tildeX == Math.floor(maxX / 2)))
            this.svg.addSymbol(g, 'SignWest', x, y);
        if ((tildeX == maxX) && (tildeY == Math.floor(maxY / 2)))
            this.svg.addSymbol(g, 'SignSouth', x, y);
        if ((tildeY == maxY) && (tildeX == Math.floor(maxX / 2)))
            this.svg.addSymbol(g, 'SignEast', x, y);

        if (!this.template.markerTop)
            this.svg.addSymbol(g, 'MarkerValues', x, y, value);
        this.svg.addSymbol(g, 'Values', x, y, value);
        /*for (var n = 0; n < tilde.brickCount; n++) {
            this.svg.addSymbol(g, 'Brick', x, y);
            y -= this.template.extraHeight;
        }*/
        if (this.template.markerTop)
            this.svg.addSymbol(g, 'MarkerValues', x, y - value * this.template.extraHeight, value);
        if (this.isRobotsPosition(tildeX, tildeY))
            this.svg.addSymbol(g, 'RobotFacing' + RobotsWorldDrawer.perspectiveToString(this._robotDirection), x, y - value * this.template.extraHeight);
    }
    return g;
}

/**
* Find corresponding RobotsTildeData instance and draw it at specified local position
* @return undefined
*/
RobotsWorldDrawer.prototype._drawTildeAt = function(
    tildeX,     // Tilde's horizontal position in tilde units
    tildeY,     // Tilde's vertical position in tilde units
    localX,     // Tilde's horizontal position in local units
    localY      // Tilde's vertical position in local units
) {
    var tilde = this.world.getSlot(tildeY, this.world.height - tildeX - 1);
    if (tilde)
        this._drawTilde(tilde, localX, localY, tildeX, tildeY);
}

/**
* Draw whole isometric world
* @return undefined
*/
RobotsWorldDrawer.prototype._drawIsometric = function() {
    var localX = (this._countXTildes() - 1) * this.template.additionalWidthPerX;
    var localY = this.calcBaselineShift();

    // Show directions
    /*!*///this.svg.addSymbol(g, 'SignWest', localX + this.template.tildeWidth * tildeY, localY);

    var tildeY, x, y;
    for (x = 0; x < this._countXTildes(); ++x) {
        for (tildeY = 0; tildeY <= x; ++tildeY)
            this._drawTildeAt(x - tildeY, tildeY, localX + this.template.tildeWidth * tildeY, localY);
        localX -= this.template.additionalWidthPerX;
        localY += this.template.additionalHeightPerX;
    }
    localX += this.template.additionalWidthPerX + this.template.additionalWidthPerY;
    --x;
    for (y = 1; y < this._countYTildes(); ++y) {
        for (tildeY = y; tildeY < this._countYTildes(); ++tildeY)
            this._drawTildeAt(x - tildeY + y, tildeY, localX + this.template.tildeWidth * (tildeY - y), localY);
        /*for (tildeY = y; tildeY <= x; ++tildeY)
            this._drawTildeAt(x - tildeY + 0, tildeY, localX + this.template.tildeWidth * (tildeY - y), localY);*/
        localX += this.template.additionalWidthPerY;
        localY += this.template.additionalHeightPerY;
    }
}

/***[ editSourceCode/RobotCodeEditor.js ] *************************/
/**
* Class for an editor for robot's program
*/
function RobotCodeEditor() {
    Object.defineProperty(this, 'node', { enumerable: true, value: createHtml('section', undefined, { contentEditable: 'true', spellcheck: 'false', 'class': 'robotcode' }) });
    createHtml('br', createHtml('span', createHtml('div', this.node)));
    this.node.setAttributeNS("http://www.w3.org/XML/1998/namespace", "xml:space", "preserve");
    this.node.addEventListener('input', ev => this._procInput(ev));
    this.node.addEventListener('keypress', ev => this._procKeypress(ev));
    this.node.addEventListener('keydown', ev => this._procKeydown(ev));
    document.addEventListener('selectionchange', ev => this._procSelectionChange(ev));
}

/** Function to call whenever text content changes */
RobotCodeEditor.prototype.onchange = function(text) {};

/** Can text not be edited by user? */
Object.defineProperty(RobotCodeEditor.prototype, 'readonly', {
    enumerable: true,
    get() { return !(this.node.contentEditable == 'true') },
    set(v) { this.node.contentEditable = v ? 'false' : 'true'; }
});

/** Plain code */
Object.defineProperty(RobotCodeEditor.prototype, 'text', {
    enumerable: true,
    get() { return RobotCodeEditor.domToText(this.node, true); },
    set(s) { this.reformat(s); }
});

/** Number of line (starting with 1) containing caret (undefined if there ist no caret) */
Object.defineProperty(RobotCodeEditor.prototype, 'caretLine', {
    enumerable: true,
    get() {
        const selection = window.getSelection();
        return this.nodeToLineNumber(selection.focusNode, selection.focusOffset);
    },
    set(line) {
        this.selectRange(line, 0, line, Infinity);
    }
});

/** Number of line (starting with 1) containing anchor of selection (undefined if there ist no anchor) */
Object.defineProperty(RobotCodeEditor.prototype, 'anchorLine', {
    enumerable: true,
    get() {
        const selection = window.getSelection();
        return this.nodeToLineNumber(selection.anchorNode, selection.anchorOffset);
    }
});

/** Number of characters in same line before of caret (undefined if there ist no caret) */
Object.defineProperty(RobotCodeEditor.prototype, 'caretChar', {
    enumerable: true,
    get() {
        const selection = window.getSelection();
        return this.nodeToCharNumber(selection.focusNode, selection.focusOffset);
    }
});

/** Number of characters in same line before of selection's anchor (undefined if there ist no anchor) */
Object.defineProperty(RobotCodeEditor.prototype, 'anchorChar', {
    enumerable: true,
    get() {
        const selection = window.getSelection();
        return this.nodeToCharNumber(selection.anchorNode, selection.anchorOffset);
    }
});

/** Number of line to highlight */
RobotCodeEditor.prototype._highlightLine = undefined;
Object.defineProperty(RobotCodeEditor.prototype, 'highlightLine', {
    enumerable: true,
    get() { return this._highlightLine; },
    set(nr) {
        if (nr === this._highlightLine)
            return;
        const oldLine = this.node.getElementsByTagName('div')[this._highlightLine - 1];
        if (Number.isFinite(this._highlightLine) && oldLine)
            oldLine.className = '';
        this._highlightLine = nr;
        const newLine = this.node.getElementsByTagName('div')[this._highlightLine - 1];
        if (Number.isFinite(this._highlightLine) && newLine)
            newLine.className = 'active';
    }
});

/**
* Get line number of position specified by node and offset within node
* @return Number        Line number of position (starting with 1)
*/
RobotCodeEditor.prototype.nodeToLineNumber = function(
    node,       // Node containing specified position
    offset      // Offset of position within node
) {
    if (!this.containsNode(node))
        return undefined;
    return RobotCodeEditor.domToTextTill(this.node, node, offset).toString().split('\n').length;
}

/**
* Get number of previous chars in this line corresponding to position specified by node and offset within node
* @return Number        Previous chars left of specified position
*/
RobotCodeEditor.prototype.nodeToCharNumber = function(
    node,       // Node containing specified position
    offset      // Offset of position within node
) {
    if (!this.containsNode(node))
        return undefined;
    const lines = RobotCodeEditor.domToTextTill(this.node, node, offset).toString().split('\n');
    const lastLine = lines.length ? lines[lines.length - 1] : '';
    return [...lastLine].length;
}

/**
* Find position specified by a node and an offset within the node corresponding to line number and character position fitting as arguments for Range.setStart and Range.setEnd
* @return Array     [node, offset]
*/
RobotCodeEditor.prototype.getRangePoint = function(
    line,       // Number of line in which point is (starting with 1 for first line)
    char        // Offset in chars
) {
    const nodeLines = this.node.getElementsByTagName('div');
    const nodeLineContainer = nodeLines[line - 1];
    if (!nodeLineContainer) {
        const lastLineContainer = nodeLines[nodeLines.length - 1];
        const lastLine = lastLineContainer.firstChild ? lastLineContainer.firstChild : lastLineContainer;
        return [lastLine, Math.max(0, lastLine.childNodes.length - 1)];
    }
    const nodeLine = nodeLineContainer.firstChild ? nodeLineContainer.firstChild : nodeLineContainer;
    if (char == 0)
        return [nodeLine, 0];
    let offset = 0;
    let node = nodeLine.firstChild;
    while (node) {
        const contentLength = [...node.textContent].length;
        if (offset + contentLength >= char)
            return [(node.nodeType == Node.ELEMENT_NODE) ? node.firstChild : node, Math.max(0, char - offset)];
        offset += contentLength;
        node = node.nextSibling;
    }
    let lastPart = nodeLine;
    while (lastPart.lastChild)
        lastPart = lastPart.lastChild;
    return [lastPart, (lastPart.nodeType == Node.ELEMENT_NODE) ? Math.max(0, nodeLine.childNodes.length - 1) : lastPart.nodeValue.length];
}

/**
* Checks if node is or is (directly or not directly) contained by the editor
* @return Boolean       Does this.node contain node?
*/
RobotCodeEditor.prototype.containsNode = function(
    node        // DOM node in question
) {
    for (node; node; node = node.parentNode) {
        if (node == this.node)
            return true;
    }
    return false;
}

/**
* Ensure correct syntax highlighting and one p Element per line
* @return undefined
*/
RobotCodeEditor.prototype.reformat = function(
    s           // Optional: New text content
) {
    const lastUpdating = this._updating;
    this._updating = true;
    const selectionInfo = (s === undefined) ? [ this.anchorLine, this.anchorChar, this.caretLine, this.caretChar ] : [undefined, undefined, undefined, undefined];
    const tokens = Token.tokenize((s !== undefined) ? s : this.text);
    const df = this.node.ownerDocument.createDocumentFragment();
    tokens.forEach(token => token.toHtml(df));
    for (let node = df.firstChild; node; node = node.nextSibling) {
        node.appendChild(node.ownerDocument.createElement('br'));
    }
    const highlightLine = df.childNodes[this.highlightLine - 1];
    if (Number.isFinite(this.highlightLine) && highlightLine)
        highlightLine.className = 'active';
    this.node.innerHTML = '';
    this.node.appendChild(df);
    this.selectRange(...selectionInfo);
    this._updating = lastUpdating;
}

/**
* Set selection and caret
* @return undefined
*/
RobotCodeEditor.prototype.selectRange = function(
    startLine,      // Start line number of selection (first line is 1)
    startChar,      // Number of chars in first line before of selection
    endLine,        // End line number of selection (first line is 1)
    endChar         // Number of chars in last line before of selection's end
) {
    const lastUpdating = this._updating;
    this._updating = true;
    if (startLine > endLine) {
        [startLine, endLine] = [endLine, startLine];
        [startChar, endChar] = [endChar, startChar];
    }
    if ((startLine == endLine) && (startChar > endChar))
        [startChar, endChar] = [endChar, startChar];
    const range = this.node.ownerDocument.createRange();
    const start = (Number.isFinite(startLine) && Number.isFinite(startChar)) ? this.getRangePoint(startLine, startChar) : null;
    const end = (Number.isFinite(endLine) && Number.isFinite(endChar)) ? this.getRangePoint(endLine, endChar) : null;
    if (start)
        range.setStart(...start);
    if (end)
        range.setEnd(...end);
    const selection = window.getSelection();
    selection.removeAllRanges();
    if (start || end)
        selection.addRange(range);
    this._updating = lastUpdating;
    this.highlightLine = endLine;
}

/**
* React on input
* @return undefined
*/
RobotCodeEditor.prototype._procInput = function(
    ev          // Event object of event
) {
    if (this._updating)
        return;
    if (ev.isComposing)
        return;
    this.reformat();
    if (this.onchange instanceof Function)
        this.onchange.call(this, this.text);
}

/**
* Change text with respect to caret's position
* @return undefined
*/
RobotCodeEditor.prototype.edit = function(
    strInsert,      // Text to add at caret pos
    delBefore,      // Number of characters, new lines or indents to remove left of caret (only if caret is collapsed)
    delAfter        // Number of characters, new lines or indents to remove right of caret (only if caret is collapsed)
) {
    const arrText = [...this.node.getElementsByTagName('div')].map(node => [...node.firstChild.textContent]);
    // Get start end end of selection in reading direction
    let endLine = this.caretLine;
    let endChar = this.caretChar;
    let startLine = this.anchorLine;
    let startChar = this.anchorChar;
    if (!Number.isFinite(endLine))
        endLine = arrText.length;
    if (!Number.isFinite(endChar))
        endChar = arrText[endLine - 1].length;
    if (!Number.isFinite(startLine) || !Number.isFinite(startChar))
        [startLine, startChar] = [endLine, endChar];
    if ((startLine > endLine) || ((startLine == endLine) && (startChar > endChar)))
        [startLine, startChar, endLine, endChar] = [endLine, endChar, startLine, startChar];

    // Do not delete more than selected text unless caret is collapsed
    if ((startLine != endLine) || (startChar != endChar)) {
        delBefore = 0;
        delAfter = 0;
    }

    // Delete all between start and end of selection
    if (startLine == endLine) {
        if (startChar != endChar)
            arrText[startLine -1].splice(startChar, endChar - startChar);
    } else {
        arrText[startLine - 1].splice(startChar);
        arrText[startLine - 1] = arrText[startLine - 1].concat(arrText[endLine - 1].slice(endChar));
        arrText.splice(startLine, endLine - startLine);
    }

    // Delete after caret
    while (delAfter) {
        const delCount = Math.min(delAfter, arrText[startLine - 1].length - startChar);
        if (delCount) {
            arrText[startLine - 1].splice(startChar, delCount);
            delAfter -= delCount;
        }
        if (delAfter) {
            if (startLine < arrText.length) {
                arrText[startLine - 1] = arrText[startLine - 1].concat(arrText[startLine]);
                arrText.splice(startLine, 1);
                --delAfter;
            } else
                delAfter = 0;
        }
    }

    // Delete before caret
    while (delBefore) {
        // Start of line?
        if (!startChar) {
            if (startLine == 1)
                break;
            startChar = arrText[startLine - 2].length;
            arrText[startLine - 2] = arrText[startLine - 2].concat(arrText[startLine - 1]);
            --startLine;
            arrText.splice(startLine, 1);
            --delBefore;
            continue;
        }
        // Only whitespace between start of line and caret?
        const startLineTokens = Token.tokenize(arrText[startLine - 1].slice(0, startChar).join(''));
        if (startLineTokens.length == 1) {
            const relevantIndents = Token.getIndentLevels(arrText.slice(0, startLine - 1).map(arr => arr.join('')).join('\n'));
            const startIndent = startLineTokens[0].indent;
            const newIndent = relevantIndents.filter(indent => (indent < startIndent)).reduce((prev, curr) => Math.max(prev, curr), 0);
            arrText[startLine - 1].splice(0, startChar, ...((' ').repeat(newIndent).split('')));
            startChar = newIndent;
            --delBefore;
            continue;
        }
        // Just delete char
        --startChar;
        arrText[startLine - 1].splice(startChar, 1);
        --delBefore;
    }

    // Insert new text
    const arrInsert = [...strInsert];
    while (arrInsert.length) {
        const charInsert = arrInsert.shift();
        switch (charInsert) {
            case '\n':
                arrText.splice(startLine, 0, arrText[startLine - 1].slice(startChar));
                arrText[startLine - 1].splice(startChar);
                ++startLine;
                startChar = 0;
                const tokensBefore = Token.tokenize(arrText[startLine - 2].join(''));
                let indent = tokensBefore[0].indent;
                if ((tokensBefore[1] instanceof Token) && tokensBefore[1].isKeyword())
                    indent = Math.floor(indent / 4 + 1) * 4;
                while (indent > startChar) {
                    arrText[startLine - 1].unshift(' ');
                    ++startChar;
                }
                break;
            case '\t':
                /*!*/
                break;
            default:
                arrText[startLine - 1].splice(startChar, 0, charInsert);
                ++startChar;
                break;
        }
    }
    // Set text and caret position
    const lastUpdating = this._updating;
    this._updating = true;
    this.text = arrText.map(line => line.join('')).join('\n');
    this.selectRange(startLine, startChar, startLine, startChar);
    this._updating = lastUpdating;
}


/** A switch to activate processing of events (set true while updateing) */
RobotCodeEditor.prototype._updating = false;

/**
* React on keydown event
* @return undefined
*/
RobotCodeEditor.prototype._procKeydown = function(
    ev          // Event object of event
) {
    if (ev.key == 'Delete') {
        ev.preventDefault();
        ev.stopPropagation();
        this.edit('', 0, 1);
        return;
    }
    if (ev.key == 'Backspace') {
        ev.preventDefault();
        ev.stopPropagation();
        this.edit('', 1, 0);
        return;
    }
    /*!*/
}

/**
* React on keypress event
* @return undefined
*/
RobotCodeEditor.prototype._procKeypress = function(
    ev          // Event object of event
) {
    if (ev.key == 'Enter') {
        ev.preventDefault();
        ev.stopPropagation();
        this.edit('\n');
        return;
    }
    /*!*/
}

/**
* React on change of selection in document
* @return undefined
*/
RobotCodeEditor.prototype._procSelectionChange = function(
    ev          // Event object of event
) {
    if (this._updating)
        return;
    this.highlightLine = this.caretLine;
}

/**
* Convert DOM node to string to represent contained text and inserts new line characters with respect to div, p and br nodes
* @return String    Text representation
*/
RobotCodeEditor.domToText = function(
    node,           // Content of which node to convert
    trimNewline     // Optional: Trim last new line character?
) {
    if ([Node.PROCESSING_INSTRUCTION_NODE, Node.COMMENT_NODE, Node.DOCUMENT_NODE, Node.DOCUMENT_TYPE_NODE].indexOf(node.nodeType) > -1)
        return '';
    if (node.nodeType != Node.ELEMENT_NODE)
        return node.textContent;
    let addNewlineIfNeeded = false;
    switch (node.tagName) {
        case 'BR':
            return '\n';
        case 'DIV':
        case 'P':
            addNewlineIfNeeded = true;
    }
    let text = [...node.childNodes].reduce((prev, curr) => (prev + RobotCodeEditor.domToText(curr, false)), '');
    if (trimNewline)
        return (text.charAt(text.length - 1) == '\n') ? text.substring(0, text.length - 1) : text;
    if (addNewlineIfNeeded)
        return (text.charAt(text.length - 1) == '\n') ? text : (text + '\n');
    return text;
}

/**
* Works like RobotCodeEditor.domToText, but stops on offset-th child of endNode
* @return Object    { str: "Text representation of node until offset-th child of endNode", found: "Found endNode?", toString() { return this.str; } }
*/
RobotCodeEditor.domToTextTill = function(
    node,           // Content of which node to convert
    endNode,        // Last text node or last nodes' parent node to consider
    offset          // Number of child node or char to stop at
) {
    const result = {
        str: '',
        found: (node == endNode),
        toString(){ return this.str; }
    };
    if ([Node.PROCESSING_INSTRUCTION_NODE, Node.COMMENT_NODE, Node.DOCUMENT_NODE, Node.DOCUMENT_TYPE_NODE].indexOf(node.nodeType) > -1)
        return result;
    if (node.nodeType != Node.ELEMENT_NODE) {
        result.str = result.found ? node.textContent.substring(0, offset) : node.textContent;
        return result;
    }
    let addNewlineIfNeeded = false;
    switch (node.tagName) {
        case 'BR':
            result.str = '\n';
            return result;
        case 'DIV':
        case 'P':
            addNewlineIfNeeded = true;
    }
    const relevantChildren = result.found ? [...node.childNodes].slice(0, offset) : [...node.childNodes];
    for (let i = 0; i < relevantChildren.length; ++i) {
        const childsResult = RobotCodeEditor.domToTextTill(relevantChildren[i], endNode, offset);
        result.str += childsResult.str;
        if (childsResult.found) {
            result.found = true;
            break;
        }
    }
    if (addNewlineIfNeeded && (result.str.charAt(result.str.length - 1) != '\n') && !result.found)
        result.str += '\n';
    return result;
}

/***[ RobotTerminal.js ] *************************/
/**
* Class for a full featured access to robot's simulation consisting of editor, controls and view of robot's world
*/
function RobotTerminal(
    world       // Optional: Maschine code of world includeing robot's speed and position
) {
    if (world)
        this._worldCode = world;
    this._synchronizedWorld = new RobotSynchronizedWorld();
    this._synchronizedWorld.init(this._worldCode);
    this._simulator = new RobotSimulator(code => this._updateState(code));
    this._editor = new RobotCodeEditor(document);
    this._editor.onchange = ((text) => this._reportcodechange(text));
    this._msgPlace = new ScoTools.MessagePlace();
    this._createDom();
}

/** DOM representation of terminal */
RobotTerminal.prototype.node = null;

/** Can code be edited (provided its not executed right now)? */
RobotTerminal.prototype._editingEnabled = true;
Object.defineProperty(RobotTerminal.prototype, 'editingEnabled', {
    enumerable: true,
    get() { return this._editingEnabled; },
    set(v) {
        this._editingEnabled = Boolean(v);
        this._editor.readonly = (!this._editingEnabled || this.running);
    }
});

/** Image element displaying present state */
RobotTerminal.prototype._img = null;

/** Node of start/stop button */
RobotTerminal.prototype._runBtn = null;

/** Maschine code of present world of robot includeing robot's speed and position */
RobotTerminal.prototype._worldCode = '0A1100000';
Object.defineProperty(RobotTerminal.prototype, 'world', {
    enumerable: true,
    set(s) {
        this.running = false;
        this._worldCode = s;
        this._synchronizedWorld.init(this._worldCode);
        this.updateImage();
    },
    get() { return this._worldCode; }
});

/** Source code */
Object.defineProperty(RobotTerminal.prototype, 'code', {
    enumerable: true,
    get() { return this._editor.text; },
    set(s) { this._editor.text = s; this._reportcodechange(s); }
});

/** Still executing program? */
Object.defineProperty(RobotTerminal.prototype, 'running', {
    enumerable: true,
    get() { return this._simulator.running; },
    set(v) {
        if (v)
            this.startExecution();
        else
            this.stopExecution();
    }
});

/** Function to ask, whether or not to run a code; has to return either true for yes and false for no or an object like { accept: "Execution allowed?", line: "Number of line containing problem (optional)", msg: "Error message to show (optional)" } */
RobotTerminal.prototype.onbeforeExecution = function(
    code,       // Value of this.code
    tokens,     // Array of all tokens representing code
    compiled,   // SourceCodePart parent instance representing parsed code
    worldDef    // Maschine code representing robot's world
) {
    return true;
}

/** Function to call whenever program execution terminates; if function returns either false for a general missmatch or an object like { accept: "Was execution a success?", line: "Number of line to highlight (optional)", msg: "Message to display" } */
RobotTerminal.prototype.onafterExecution = function(
    code,       // Value of this.code executed
    line,       // Number of last line executed
    result,     // Reason for termination: undefined if aborted by user, otherwise RobotSimulator.RESULT_*
    worldDef,   // Maschine code representing robot's world before any execution
    worldStat,  // Present robot's world after execution (RobotWorld instance)
    robotStat   // Location and viewing direction of robot after execution: { x, y, direction }
) {
    return true;
}

/** Function to call whenever code input value changes */
RobotTerminal.prototype.oncodeChange = function(
    code        // Value this.code
) {
}

/** Used RobotCodeEditor instance */
RobotTerminal.prototype._editor = null;

/** RobotSynchronizedWorld instance used to represent present state */
RobotTerminal.prototype._synchronizedWorld = null;

/** RobotSimulator instance used to run program */
RobotTerminal.prototype._simulator = null;

/** MessagePlace instance used to display messages above of source code */
RobotTerminal.prototype._msgPlace = null;

/** Function to call in order to revoke present image */
RobotTerminal.prototype._revokeImg = (() => {});

/**
* Create DOM representation of terminal
* @return Element       New value of this.node;
*/
RobotTerminal.prototype._createDom = function() {
    const nodeMain = ScoTools.createHtml('div', null, { 'class': 'robotterminal' });
    const nodeEdit = ScoTools.createHtml('div', nodeMain);
    const nodeView = ScoTools.createHtml('figure', nodeMain);

    this._msgPlace.appendAsChild(nodeEdit);
    nodeEdit.appendChild(this._editor.node);
    this._runBtn = ScoTools.createHtml('button', nodeEdit, null, 'Roboter starten');
    this._runBtn.addEventListener('click', ev => {
        ev.preventDefault();
        if (this.running)
            this._msgPlace.add('Ausführung abgebrochen', ScoTools.Message.TYPE_INFO);
        this.toggleExecution();
    });

    this._img = ScoTools.createHtml('img', nodeView, { src: '', alt: 'Roboter in seiner Welt' });
    this.updateImage();
    /*!*/
    this.node = nodeMain;
    return nodeMain;
}

/**
* Update image of robot's world
* @return undefined
*/
RobotTerminal.prototype.updateImage = function() {
    const drawer = this._synchronizedWorld.draw();
    drawer.onloaded = (svg => {
        this._img.src = svg.getBlobUrl();
        this._revokeImg();
        this._revokeImg = (() => svg.revokeBlobUrl());
    });
    drawer.draw();
}

/**
* Update synchronized robot's world to state returned by RobotSimulator instance
* @return undefined
*/
RobotTerminal.prototype._updateState = function(
    code            // Maschine code created by RobotSimulator representing current world's state
) {
    const state = (code !== undefined) ? this._synchronizedWorld.update(code) : { status: undefined, line: this._editor.caretLine };
    this._editor.caretLine = state.line;
    this.updateImage();
    switch (state.status) {
        case ((state.status === undefined) ? state.status : !state.status):
            this._msgPlace.add('Ausführung abgebrochen', ScoTools.Message.TYPE_INFO);
            break;
        case RobotSimulator.RESULT_REQUESTED:
        case RobotSimulator.RESULT_LOG:
            return;
        case RobotSimulator.RESULT_DONE:
            this._msgPlace.add('Erfolgreich erledigt', ScoTools.Message.TYPE_SUCCESS);
            break;
        case RobotSimulator.RESULT_ABORT:
            this._msgPlace.add('Algorithmus veranlasste Notabschaltung', ScoTools.Message.TYPE_INFO);
            break;
        case RobotSimulator.RESULT_MISPLACED:
            this._msgPlace.add('Notabschaltung beim Betreten eines nicht-begehbaren Feldes');
            break;
        case RobotSimulator.RESULT_UNDERFLOW:
            this._msgPlace.add('Notabschaltung: minus auf leerem Feld nicht möglich');
            break;
        case RobotSimulator.RESULT_OVERFLOW:
            this._msgPlace.add('Notabschaltung: plus auf vollem Feld nicht möglich');
            break;
        case RobotSimulator.RESULT_FULLSTACK:
            this._msgPlace.add('Notabschaltung wegen zu viel Rekursion');
            break;
        default:
            this._msgPlace.add('Notabschaltung wegen unbekanntem Fehler');
            break;
    }
    this._editor.readonly = !this._editingEnabled;
    this._runBtn.textContent = 'Roboter starten';
    const externalCheck = this.onafterExecution(this.code, state.line, state.status, this.world, this._synchronizedWorld.world, { x: this._synchronizedWorld.robotX, y: this._synchronizedWorld.robotY, direction: this._synchronizedWorld.robotDirection });
    if ((externalCheck === false) && ((state.status == RobotSimulator.RESULT_DONE) || (state.status == RobotSimulator.RESULT_ABORT))) {
        this._msgPlace.add('Die Aufgabe wurde nicht erfolgreich erfüllt');
        return;
    }
    if ((typeof externalCheck != 'object') || !externalCheck)
        return;
    if (Number.isFinite(externalCheck.line))
        this._editor.caretLine = externalCheck.line;
    if (externalCheck.accept && !externalCheck.msg)
        return;
    if ([RobotSimulator.RESULT_DONE, RobotSimulator.RESULT_LOG, RobotSimulator.RESULT_REQUESTED].reduce((prev, status) => (prev || (status === state.status)), false))
        this._msgPlace.add(externalCheck.msg ? externalCheck.msg : 'Die Aufgabe wurde nicht erfolgreich erfüllt', externalCheck.accept ? ScoTools.Message.TYPE_SUCCESS : ScoTools.TYPE_ERROR);
}

/**
* Start robot if not running and stops him otherwise
* @return undefined
*/
RobotTerminal.prototype.toggleExecution = function() {
    this.running = !this.running;
}

/**
* Stop execution of source code
* @return undefined
*/
RobotTerminal.prototype.stopExecution = function() {
    if (!this.running)
        return;
    this._simulator.tellStatus();
    this._simulator.stop();
    this._editor.readonly = !this._editingEnabled;
    this._runBtn.textContent = 'Roboter starten';
    this._updateState(undefined);
}

/**
* Start execution of inputed source code
* @return undefined
*/
RobotTerminal.prototype.startExecution = function() {
    this.stopExecution();
    this._msgPlace.clear();
    // Compile
    const tokens = Token.tokenize(this.code);
    const program = SourceCodePart.tokensToSequenz(tokens);
    const compiled = program.compile();
    if (compiled instanceof SourceCodeError) {
        this._editor.caretLine = compiled.line;
        this._msgPlace.add(compiled.toString());
        return;
    }
    // Perform external check if this code is allowed
    const externalCheck = this.onbeforeExecution(this.code, tokens, program, this.world);
    if ((externalCheck === false) || ((typeof externalCheck == 'object') && !externalCheck.accept)) {
        this._msgPlace.add(((typeof externalCheck == 'object') && externalCheck.msg) ? externalCheck.msg : 'Dieser Quelltext entspricht nicht den Anforderungen');
        if ((typeof externalCheck == 'object') && Number.isFinite(externalCheck.line))
            this._editor.caretLine = externalCheck.line;
        return;
    }
    // Update ui
    this._editor.readonly = true;
    this._runBtn.textContent = 'Algorithmus abbrechen';
    // Start execution
    this._simulator.load(this._worldCode + compiled);
    this._simulator.start();
}

/**
*/
RobotTerminal.prototype._reportcodechange = function(
    code        // Present value of this.code
) {
    if (this.oncodeChange instanceof Function)
        this.oncodeChange.call(this, code);
}

/**
* Remove all former messages from message output
* @return undefined
*/
RobotTerminal.prototype.clearMsg = function() {
    this._msgPlace.clear();
}

/**
* Display message
* @return undefined
*/
RobotTerminal.prototype.addMsg = function(
    text,       // Text of message
    type        // Type of message (optional, default value Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}
